﻿/** @file cchar.cpp

	@author Sasaji
	@date   2014.12.01

	@brief 文字列ポインタリスト
*/

#include "cchar.h"
#include "common.h"
#ifdef _WIN32
#include <windows.h>
#endif

//
//
//
//
CNchar::CNchar() : CCharTemp<char>()
{
	wstr = NULL;
}
CNchar::CNchar(const CNchar &src) : CCharTemp<char>(src)
{
	wstr = NULL;
}
CNchar::CNchar(const char *src_str) : CCharTemp<char>(src_str, src_str ? (int)strlen(src_str) : 0)
{
	wstr = NULL;
}
CNchar::CNchar(const char *src_str, int src_len) : CCharTemp<char>(src_str, src_len)
{
	wstr = NULL;
}
CNchar::CNchar(const wchar_t *src_str) : CCharTemp<char>()
{
	wstr = NULL;
	this->SetW(src_str); 
}
CNchar::CNchar(const wchar_t *src_str, int src_len) : CCharTemp<char>()
{
	wstr = NULL;
	this->SetW(src_str, src_len); 
}
CNchar::~CNchar() {
	delete [] wstr;
	wstr = NULL;
}

/// @brief 文字列をセット
/// @param[in] src_str 値
void CNchar::Set(const char *src_str)
{
	CCharTemp<char>::Set(src_str, src_str ? (int)strlen(src_str) : 0);
}
/// @brief 文字列をセット
/// @param[in] src_str 値
void CNchar::SetN(const char *src_str)
{
	Set(src_str);
}
/// @brief 文字列をセット
/// @param[in] src_str 値
/// @param[in] src_len 長さ
void CNchar::SetN(const char *src_str, int src_len)
{
	CCharTemp<char>::Set(src_str, src_len);
}
/// @brief 文字列が一致するか
/// @param[in] value 値
/// @return true / false
bool CNchar::MatchString(const char *value)
{
	if (str != NULL && value != NULL && strcmp(str, value) == 0) {
		return true;
	}
	return false;
}
/// @brief 文字列の一部が一致するか
/// @param[in] value 値
/// @return true / false
bool CNchar::MatchSubString(const char *value)
{
	if (str != NULL && value != NULL && (strstr(str, value) != NULL || strstr(value, str) != NULL)) {
		return true;
	}
	return false;
}
/// @brief 文字列を代入
/// @param[in] src 値
/// @return 文字列
CNchar &CNchar::operator=(const CNchar &src)
{
	CCharTemp<char>::Set(src);
	return *this;
}
/// @brief 文字列をセット
/// @param[in] src_str 値
/// @return 文字列
CNchar &CNchar::operator=(const char *src_str)
{
	CCharTemp<char>::Set(src_str, src_str ? (int)strlen(src_str) : 0);
	return *this;
}
/// @brief 文字列を返す
/// @return 文字列
const char *CNchar::GetN() const
{
	return this->Get();
}
/// @brief 文字列をコピー
/// @param[out] dst_str コピー先
/// @param[in]  dst_len コピー先バッファサイズ
void CNchar::GetN(char *dst_str, int dst_len)
{
	if (GetN()) {
		memset(dst_str, 0, sizeof(char) * dst_len);
		strncpy(dst_str, str, dst_len);
	}
}
/// @brief 文字列をセット(変換あり)
/// @param[in] src_str ワイド文字
/// @return true / false
void CNchar::SetW(const wchar_t *src_str)
{
	SetW(src_str, -1);
}
/// @brief 文字列をセット(変換あり)
/// @param[in]  src_str ワイド文字
/// @param[in]  src_len ワイド文字長さ
/// @attention Windows only
void CNchar::SetW(const wchar_t *src_str, int src_len)
{
#ifdef _WIN32
	UINT acp = GetACP();
	delete [] str;
	str = NULL;
	// calc length converting to narrow char from wide char
	len = WideCharToMultiByte(acp, 0, src_str, src_len, NULL, 0, NULL, NULL);
	if (len == 0) return;
	// copy data
	str = new char[len + 1];
	WideCharToMultiByte(acp, 0, src_str, src_len, str, len, NULL, NULL);
	str[len] = 0;
#endif
}
/// @brief ワイド文字に変換して返す
/// @return 文字列へのポインタ
/// @attention Windows only
const wchar_t *CNchar::GetW()
{
#ifdef _WIN32
	delete [] wstr;
	// MBS -> WideChar
	// 長さを取得
	int wlen = MultiByteToWideChar(CP_ACP, 0, str, -1, wstr, 0);
	wstr = new wchar_t[wlen + 1];
	MultiByteToWideChar(CP_ACP, 0, str, -1, wstr, wlen);
	return wstr;
#else
	return NULL;
#endif
}
/// @brief ワイド文字に変換してコピー
/// @param[out] dst_str コピー先
/// @param[in]  dst_len コピー先バッファサイズ
void CNchar::GetW(wchar_t *dst_str, int dst_len)
{
	if (GetW()) {
		memset(dst_str, 0, sizeof(wchar_t) * dst_len);
		wcsncpy(dst_str, wstr, dst_len);
	}
}

#if (defined(UNICODE) || defined(_UNICODE))
//
//
//
//
CWchar::CWchar() : CCharTemp<wchar_t>()
{
	nstr = NULL;
}
CWchar::CWchar(const CWchar &src) : CCharTemp<wchar_t>(src)
{
	nstr = NULL;
}
CWchar::CWchar(const wchar_t *src_str) : CCharTemp<wchar_t>(src_str, src_str ? (int)wcslen(src_str) : 0)
{
	nstr = NULL;
}
CWchar::CWchar(const wchar_t *src_str, int src_len) : CCharTemp<wchar_t>(src_str, src_len)
{
	nstr = NULL;
}
CWchar::~CWchar()
{
	delete [] nstr;
	nstr = NULL;
}
/// @brief 文字列をセット(変換あり)
/// @param[in] src_str 値
/// @return true / false
CWchar::CWchar(const char *src_str)
{
	nstr = NULL;
	SetN(src_str);
}
/// @brief 文字列をセット(変換あり)
/// @param[in] src_str 値
/// @param[in] src_len 長さ
/// @return true / false
CWchar::CWchar(const char *src_str, int src_len)
{
	nstr = NULL;
	SetN(src_str, src_len);
}
/// @brief 文字列をセット
/// @param[in] src_str 値
/// @return true / false
void CWchar::Set(const wchar_t *src_str)
{
	CCharTemp<wchar_t>::Set(src_str, src_str ? (int)wcslen(src_str) : 0);
}
/// @brief 文字列を取得
/// @return 文字列
const wchar_t *CWchar::GetW() const
{
	return this->Get();
}
/// @brief 文字列をコピー
/// @param[out] dst_str コピー先
/// @param[in]  dst_len コピー先バッファサイズ
void CWchar::GetW(wchar_t *dst_str, int dst_len)
{
	if (GetW()) {
		memset(dst_str, 0, sizeof(wchar_t) * dst_len);
		wcsncpy(dst_str, str, dst_len);
	}
}

/// @brief 文字列をセット(変換あり)
/// @param[in] src_str 値
/// @return true / false
void CWchar::SetN(const char *src_str)
{
	SetN(src_str, -1);
}
/// @brief 文字列をセット(変換あり)
/// @param[in] src_str 値
/// @param[in] src_len 長さ
/// @return true / false
void CWchar::SetN(const char *src_str, int src_len)
{
#ifdef _WIN32
	UINT acp = GetACP();
	// calc length converting to wide char from narrow char
	int wlen = MultiByteToWideChar(acp, 0, src_str, src_len, NULL, 0);
	if (wlen == 0) return;
	// allocate buffer
	CopyData(NULL, wlen);
	// copy data
	MultiByteToWideChar(acp, 0, src_str, src_len, str, wlen);
#endif
}
/// @brief 文字列を取得(変換あり)
/// @param[in] codepage コードページ
/// @return 文字列
char *CWchar::GetN(int codepage)
{
#ifdef _WIN32
	delete [] nstr;
	UINT acp = (codepage < 0 ? GetACP() : codepage);
	// calc length converting to narrow char from wide char
	int nlen = WideCharToMultiByte(acp, 0, str, -1, NULL, 0, NULL, NULL);
	if (nlen == 0) return NULL;
	nstr = new char[nlen + 1];
	// copy data
	WideCharToMultiByte(acp, 0, str, -1, nstr, nlen, NULL, NULL);
	return nstr;
#else
	return NULL;
#endif
}
/// @brief 文字列をコピー(変換あり)
/// @param[in] codepage コードページ
void CWchar::GetN(char *dst_str, int dst_len, int codepage)
{
	if (GetN(codepage)) {
		memset(dst_str, 0, sizeof(char) * dst_len);
		strncpy(dst_str, nstr, dst_len);
	}
}
/// @brief 文字列が一致するか
/// @param[in] value 値
/// @return true / false
bool CWchar::MatchString(const wchar_t *value)
{
	if (str != NULL && wcscmp(str, value) == 0) {
		return true;
	}
	return false;
}
/// @brief 文字列の一部が一致するか
/// @param[in] value 値
/// @return true / false
bool CWchar::MatchSubString(const wchar_t *value)
{
	if (str != NULL && (wcsstr(str, value) != NULL || wcsstr(value, str) != NULL)) {
		return true;
	}
	return false;
}
/// @brief 文字列を代入
/// @param[in] src 値
/// @return 文字列
CWchar &CWchar::operator=(const CWchar &src)
{
	CCharTemp<wchar_t>::Set(src);
	return *this;
}
/// @brief 文字列をセット
/// @param[in] src_str 値
/// @return 文字列
CWchar &CWchar::operator=(const wchar_t *src_str)
{
	CCharTemp<wchar_t>::Set(src_str, src_str ? (int)wcslen(src_str) : 0);
	return *this;
}

#endif /* defined(UNICODE) || defined(_UNICODE) */
