/** @file clocale.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.12.01

	@brief [i18n with gettext]
*/

#include "clocale.h"
#include "utility.h"

#if defined(_WIN32) && !defined(__MINGW32__)
#pragma warning( disable : 4996 )
#endif

#ifdef _WIN32
#include <shlwapi.h>

static const char *locale_for_search;
static LCID decide_lcid;
#endif

static const _TCHAR *sub_dir[] = {
	_T(""),
#ifdef _WIN32
	_T("..\\"),
	_T("..\\..\\"),
	_T("..\\..\\..\\"),
	_T("..\\..\\..\\..\\"),
#else
	_T("../"),
	_T("../../"),
	_T("../../../"),
	_T("../../../../"),
#endif
	NULL
};


CLocale::CLocale()
{
	Clear();
}

CLocale::CLocale(const _TCHAR *new_app_path, const _TCHAR *new_package_name, const char *new_locale)
{
	Clear();

	SetLocale(new_app_path, new_package_name, new_locale);
}

CLocale::~CLocale()
{
	if (package_name != NULL) {
#ifdef USE_GETTEXT
#if defined(_WIN32) && !defined(__MINGW32__)
		libintl_tfreeres();
#endif
#endif
		delete [] package_name;
	}
	if (locale_name != NULL) {
		delete [] locale_name;
	}
}

void CLocale::Clear()
{
	package_name = NULL;
	locale_name = NULL;
#ifdef _WIN32
	lcid = 0;
#endif

#ifdef _WIN32
	locale_for_search = NULL;
	decide_lcid = 0;
#endif
}

/// @brief Initialize and set a locale
bool CLocale::SetLocale(const _TCHAR *new_app_path, const _TCHAR *new_package_name, const char *new_locale) {
	if (new_package_name == NULL) {
		return false;
	}

	const char *locale = setlocale(LC_ALL, new_locale);
	if (locale == NULL) {
        return false;
    }
#ifdef _WIN32
	if (new_locale != NULL && new_locale[0] != '\0') {
		// change language
		LCID new_lcid = get_lcid(locale);
		if (new_lcid > 0) {
			lcid = new_lcid;
			SetThreadLocale();
		}
	} else {
		lcid = ::GetThreadLocale();
	}
#endif

#ifdef USE_GETTEXT
	if (_ttextdomain(new_package_name) == NULL) {
        return false;
    }

	delete [] package_name;
	package_name = new _TCHAR[_tcslen(new_package_name)+1];
	_tcscpy(package_name, new_package_name);

	_TCHAR locale_dir[_MAX_PATH];

	bool exist = false;
	for(int i=0; sub_dir[i] != NULL; i++) {
		_tcscpy(locale_dir, new_app_path);
		_tcscat(locale_dir, sub_dir[i]);
		_tcscat(locale_dir, _T("locale"));

		struct stat st;
		if (stat(UTILITY::slim_path(locale_dir), &st) == 0) {
			exist = true;
			break;
		}
	}
	if (!exist) {
		return false;
	}

    if (_tbindtextdomain(new_package_name, locale_dir) == NULL) {
        return false;
    }
#if defined(UNICODE) || defined(_UNICODE)
    if (wbind_textdomain_codeset(new_package_name, _T("UTF-16LE")) == NULL) {
        return false;
    }
#else
#if defined(USE_SDL) || defined(USE_SDL2)
	if (bind_textdomain_codeset(new_package_name, _T("UTF-8")) == NULL) {
		return false;
	}
#endif
#endif

#endif	// USE_GETTEXT

	delete [] locale_name;
	locale_name = new char[strlen(locale)+1];
	strcpy(locale_name, locale);

	return true;
}

/// @brief set dummy locale. So, no translate.
bool CLocale::UnsetLocale()
{
	const _TCHAR package[] = _T("dummy_locale");
#ifdef USE_GETTEXT
	if (_ttextdomain(package) == NULL) {
        return false;
    }
#endif
	delete [] package_name;
	package_name = new _TCHAR[_tcslen(package)+1];
	_tcscpy(package_name, package);

	return true;
}

/// @brief set locale
/// @note SetThreadLocale WinAPI set a locale on current thread only. So, you need call this on every threads.
bool CLocale::SetThreadLocale()
{
#ifdef _WIN32
	if (lcid == 0) {
		lcid = ::GetThreadLocale();
	}
	return (::SetThreadLocale(lcid) == TRUE);
#else
	return true;
#endif
}

#ifdef _WIN32

/// @brief Compare with specified locale name and supported locale name
BOOL CALLBACK CLocale::enum_lcids_proc(LPSTR lcid_string)
{
	if (locale_for_search == NULL) {
		return FALSE;
	}

	char language_name[128], country_name[128];
	char *endp;
	LCID lcid = strtoul(lcid_string, &endp, 16);
	GetLocaleInfoA(lcid, LOCALE_SENGLANGUAGE, language_name, sizeof(language_name));
	char *p = strrchr(language_name, '(');
	if (p != NULL) *p = '\0';
	GetLocaleInfoA(lcid, LOCALE_SENGCOUNTRY,  country_name, sizeof(country_name));
	if (strstr(locale_for_search, language_name) != NULL && strstr(locale_for_search, country_name) != NULL) {
		decide_lcid = lcid;
		return FALSE;
	}
	return TRUE;
}

/// @brief Get lcid from a locale name
LCID CLocale::get_lcid(const char *locale)
{
	locale_for_search = locale;
	decide_lcid = 0;
	EnumSystemLocalesA(enum_lcids_proc, LCID_SUPPORTED);
	locale_for_search = NULL;
	return decide_lcid;
}

#endif
