/** @file gtk_input.cpp

	Skelton for retropc emulator
	GTK+ + SDL edition

	@author Sasaji
	@date   2017.01.21

	@brief [ gtk input ]

	@note
	This code is based on win32_input.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "fifo.h"
#include "fileio.h"
#include "gtk_input.h"

#define KEY_KEEP_FRAMES 3

void EMU::EMU_INPUT()
{
	vm_key_status = NULL;
	vm_key_status_size = 0;

	// initialize status
	memset(key_status, 0, sizeof(key_status));
	memset(mouse_status, 0, sizeof(mouse_status));
#ifdef USE_MOUSE_ABSOLUTE
	mouse_posx = mouse_posy = 0;
#endif

#ifdef USE_JOYSTICK
	memset(joy_status, 0, sizeof(joy_status));
	for(int i=0; i<2; i++) {
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
#endif

	pressed_global_key = false;

#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = NULL;
	autokey_phase = 0;
	autokey_shift = 0;
	autokey_enabled = false;
#endif
}

void EMU::initialize_input()
{
	out_debug(_T("EMU::initialize_input"));

#ifdef USE_JOYSTICK
	use_joystick = FLG_USEJOYSTICK_ALL ? true : false;
	// initialize joysticks
	if (SDL_InitSubSystem(SDL_INIT_JOYSTICK)) {
		out_logf(LOG_WARN, _T("SDL_InitSubSystem(SDL_INIT_JOYSTICK): %s."), SDL_GetError());
	}
#endif
	initialize_joystick();

	// mouse emulation is disabled
	mouse_disabled = 1;
	initialize_mouse(FLG_USEMOUSE ? true : false);

#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = new FIFO(65536);
	autokey_buffer->clear();
	autokey_phase = 31;	// wait a few sec.
	autokey_shift = 0;
	autokey_enabled = false;
#endif
	lost_focus = false;
	key_mod = 0;
}

void EMU::initialize_joystick()
{
#ifdef USE_JOYSTICK
	for(int i = 0; i < 2; i++) {
		joy_xmin[i] = -16384;
		joy_xmax[i] = 16384;
		joy_ymin[i] = -16384;
		joy_ymax[i] = 16384;
	}
	SDL_JoystickEventState(SDL_IGNORE);
	reset_joystick();
#endif
}

void EMU::reset_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	joy_num = SDL_NumJoysticks();
	for(int i = 0; i < joy_num && i < 2; i++) {
#ifndef USE_SDL2
		if (!SDL_JoystickOpened(i)) {
#else
		if (!SDL_JoystickGetAttached(joy[i])) {
#endif
			if ((joy[i] = SDL_JoystickOpen(i)) != NULL) {
				joy_enabled[i] = true;
			}
		}
	}
#endif
}

void EMU::release_input()
{
	// release mouse
	if(!mouse_disabled) {
		disable_mouse(0);
	}
	// release joystick
	release_joystick();

#ifdef USE_AUTO_KEY
	// release autokey buffer
	if(autokey_buffer) {
		autokey_buffer->release();
		delete autokey_buffer;
	}
#endif
}

void EMU::release_joystick()
{
#ifdef USE_JOYSTICK
	// release joystick
	for(int i = 0; i < 2; i++) {
		if (joy[i] != NULL) {
			SDL_JoystickClose(joy[i]);
		}
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
#endif
}

void EMU::update_input()
{
//	out_debug("EMU::update_input");

	// release keys
#ifdef USE_AUTO_KEY
	if(lost_focus && !autokey_enabled) {
#else
	if(lost_focus) {
#endif
		// we lost key focus so release all pressed keys
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x80) {
				key_status[i] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	else {
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x7f) {
				key_status[i] = (key_status[i] & 0x80) | ((key_status[i] & 0x7f) - 1);
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	lost_focus = false;

#ifdef USE_JOYSTICK
	update_joystick();
#endif
#ifdef USE_KEY_TO_JOY
	// emulate joystick #1 with keyboard
	if(key_status[0x26]) joy_status[0] |= 0x01;	// up
	if(key_status[0x28]) joy_status[0] |= 0x02;	// down
	if(key_status[0x25]) joy_status[0] |= 0x04;	// left
	if(key_status[0x27]) joy_status[0] |= 0x08;	// right
#endif

	// update mouse status
	update_mouse();

#ifdef USE_LIGHTPEN
	if(FLG_USELIGHTPEN) {
		// get current status
		VmPoint pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));
		pt.x = mixed_width  * (pt.x - stretched_dest_real_x) / stretched_width;
		pt.y = mixed_height * (pt.y - stretched_dest_real_y) / stretched_height;

		mouse_status[0] = pt.x;
		mouse_status[1] = pt.y;
		mouse_status[2] = (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0) | (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
	}
#endif
}

void EMU::update_mouse()
{
	if (!initialized) return;

	// update mouse status
	memset(mouse_status, 0, sizeof(mouse_status));
	if(!mouse_disabled) {
		// get current status
		VmPoint pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));

#ifndef USE_SDL2
		mouse_status[0]  = pt.x - display_width / 2;
		mouse_status[1]  = pt.y - display_height / 2;
#else
		mouse_status[0]  = pt.x - screen_width / 2;
		mouse_status[1]  = pt.y - screen_height / 2;
#endif
		mouse_status[2] =  (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_MIDDLE) ? 4 : 0);
		// move mouse cursor to the center of window
		if(!(mouse_status[0] == 0 && mouse_status[1] == 0)) {
#ifndef USE_SDL2
			pt.x = display_width / 2;
			pt.y = display_height / 2;
			SDL_WarpMouse(pt.x, pt.y);
#else
			pt.x = screen_width / 2;
			pt.y = screen_height / 2;
			SDL_WarpMouseInWindow(window, pt.x, pt.y);
#endif
		}
	}
}

void EMU::update_autokey()
{
#ifdef USE_AUTO_KEY
	// auto key
	switch(autokey_phase) {
	case 1:
		if(autokey_buffer && !autokey_buffer->empty()) {
#ifdef USE_EMU_INHERENT_SPEC
			// update graph key status
			int graph = autokey_buffer->read_not_remove(0) & AUTO_KEY_GRAPH_MASK;
			if(graph && !autokey_shift) {
				vm_key_down(AUTO_KEY_GRAPH);
				out_debugf(_T("autokey phase:% 2d: graph down"), autokey_phase);
				autokey_shift = graph;
				autokey_phase++;
				break;
			}
#endif
			// update shift key status
			int shift = autokey_buffer->read_not_remove(0) & AUTO_KEY_SHIFT_MASK;
			if(shift && !autokey_shift) {
				vm_key_down(AUTO_KEY_SHIFT);
				out_debugf(_T("autokey phase:% 2d: shift down"), autokey_phase);
				autokey_shift = shift;
				autokey_phase++;
				break;
			}
			autokey_shift = 0;
		}
		autokey_phase++;
		break;
	case 2:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_code = autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK;
			vm_key_down(autokey_code);
			out_debugf(_T("autokey phase:% 2d: key down %x"), autokey_phase, autokey_code);
		}
		autokey_phase++;
		break;
	case USE_AUTO_KEY:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debugf(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 2:
		if(autokey_shift & AUTO_KEY_SHIFT_MASK) {
			vm_key_up(AUTO_KEY_SHIFT);
			out_debugf(_T("autokey phase:% 2d: shift up:"), autokey_phase);
		}
#ifdef USE_EMU_INHERENT_SPEC
		if(autokey_shift & AUTO_KEY_GRAPH_MASK) {
			vm_key_up(AUTO_KEY_GRAPH);
			out_debugf(_T("autokey phase:% 2d: graph up"), autokey_phase);
		}
#endif
		autokey_shift = 0;
		if(autokey_buffer && !autokey_buffer->empty()) {
			vm_key_up(autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK);
		}
		out_debugf(_T("autokey phase:% 2d: key up %x"), autokey_phase, autokey_code);
		autokey_code = 0;
		autokey_phase++;
		break;
	case USE_AUTO_KEY + 3:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debugf(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 4:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// wait enough while vm analyzes one line
			int code = autokey_buffer->read();
			out_debugf(_T("autokey phase:% 2d: code %x"), autokey_phase, code);
			if(code == AUTO_KEY_RETURN) {
				autokey_phase++;	// wait 30 frames
				break;
			}
		}
		// through
	case 30:
	case 150:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_phase = 1;
		}
		else {
			stop_auto_key();
			autokey_phase = 0;
		}
		break;
	default:
		if(autokey_phase) {
			autokey_phase++;
		}
		break;
	}
#endif
}

void EMU::recognized_key(uint16_t key_code)
{
#ifdef USE_AUTO_KEY
	if(autokey_phase == 0) return;
	if(autokey_phase == USE_AUTO_KEY && autokey_code) {
		out_debugf(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		if (autokey_code == key_code) {
			autokey_phase = USE_AUTO_KEY + 1;
		}
	}
	if(autokey_phase == USE_AUTO_KEY + 3 && key_code == AUTO_KEY_NONE) {
		out_debugf(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		autokey_phase =  USE_AUTO_KEY + 4;
	}
#endif
}

void EMU::update_joystick()
{
#ifdef USE_JOYSTICK
	memset(joy_status, 0, sizeof(joy_status));

	if (!use_joystick) return;

	// update joystick status
	SDL_JoystickUpdate();
	for(int i = 0; i < 2; i++) {
		if(joy_enabled[i] == true) {
#ifndef USE_SDL2
			if (SDL_JoystickOpened(i)) {
#else
			if (!SDL_JoystickGetAttached(joy[i])) {
#endif
				int x = SDL_JoystickGetAxis(joy[i], 0);
				int y = SDL_JoystickGetAxis(joy[i], 1);
				if(y < joy_ymin[i]) joy_status[i] |= 0x01;
				if(y > joy_ymax[i]) joy_status[i] |= 0x02;
				if(x < joy_xmin[i]) joy_status[i] |= 0x04;
				if(x > joy_xmax[i]) joy_status[i] |= 0x08;
				int nbuttons = SDL_JoystickNumButtons(joy[i]);
				for(int n=0; n<nbuttons && n<28; n++) {
					if (SDL_JoystickGetButton(joy[i], n)) {
						joy_status[i] |= (0x10 << n);
					}
				}
			} else {
				joy_enabled[i] = false;
			}
		}
	}
#endif
}

int EMU::key_down_up(uint8_t type, int code, short scan_code)
{
	bool keep_frames = false;

	// translate keycode
	// type change UP to DOWN when capslock key in macosx
	if (!translate_keysym(type, code, scan_code, &code, &keep_frames)) {
#ifdef LOG_MEASURE
		emu->out_debugf(_T("KEYDOWN: code:%02x scancode:%02x")
			, code, scan_code);
#endif
		if (key_mod & KEY_MOD_ALT_KEY) {
			// notify key down
			system_key_down(code);
			// execute for pressed global key
			execute_global_keys(code, 0);
		} else {
			key_down(code, keep_frames);
		}
	} else {
#ifdef LOG_MEASURE
		out_debugf(_T("KEYUP: code:%02x scancode:%02x")
			, code, scan_code);
#endif
		if (key_mod & KEY_MOD_ALT_KEY) {
			// notify key down
			system_key_down(code);
			// release global key
			if (release_global_keys(code, 0)) return 0;
		}
		key_up(code, keep_frames);
	}
	return 0;
}

int EMU::key_down(int code, bool keep_frames)
{
	key_status[code] = keep_frames ? KEY_KEEP_FRAMES * FRAME_SPLIT_NUM : 0x80;

#ifdef NOTIFY_KEY_DOWN_TO_GUI
	gui->KeyDown(code, mod);
#endif
#ifdef NOTIFY_KEY_DOWN
	vm->key_down(code);
#endif
	return code;
}

void EMU::key_up(int code, bool keep_frames)
{
	if(key_status[code]) {
		key_status[code] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN_TO_GUI
		gui->KeyUp(code);
#endif
#ifdef NOTIFY_KEY_DOWN
		if(!key_status[code]) {
			vm->key_up(code);
		}
#endif
	}
}

int EMU::vm_key_down(int code)
{
#ifdef USE_EMU_INHERENT_SPEC
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] |= 1;
	}
	return code;
}
void EMU::vm_key_up(int code)
{
#ifdef USE_EMU_INHERENT_SPEC
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] &= ~1;
	}
}

#ifdef USE_BUTTON
void EMU::press_button(int num)
{
	int code = buttons[num].code;

	if(code) {
		key_down(code, false);
		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
	}
	else {
		// code=0: reset virtual machine
		vm->reset();
	}
}
#endif

void EMU::initialize_mouse(bool enable)
{
	if (enable) enable_mouse(0);
}

void EMU::enable_mouse(int mode)
{
	// enable mouse emulation
	int pmouse_disabled = mouse_disabled;
	mouse_disabled &= (mode ? ~2 : ~1);
	if(pmouse_disabled && !mouse_disabled) {
		// hide mouse cursor
		SDL_ShowCursor(SDL_DISABLE);
		// move mouse cursor to the center of window
		VmPoint pt;
#ifndef USE_SDL2
		pt.x = display_width / 2;
		pt.y = display_height / 2;
		SDL_WarpMouse(pt.x, pt.y);
#else
		pt.x = screen_width / 2;
		pt.y = screen_height / 2;
		SDL_WarpMouseInWindow(window, pt.x, pt.y);
#endif
	}
}

void EMU::disable_mouse(int mode)
{
	// disable mouse emulation
	if(!mouse_disabled) {
		SDL_ShowCursor(SDL_ENABLE);
	}
	mouse_disabled |= (mode ? 2 : 1);
}

void EMU::toggle_mouse()
{
	// toggle mouse enable / disable
	if(mouse_disabled & 1) {
		enable_mouse(0);
	} else {
		disable_mouse(0);
	}
}

bool EMU::get_mouse_enabled()
{
	return ((mouse_disabled & 1) == 0);
}



#ifdef USE_AUTO_KEY
bool EMU::start_auto_key(const char *str)
{
	bool rc = true;

	stop_auto_key();

	autokey_code = 0;
	autokey_buffer->clear();

	int size = (int)strlen(str);
	int prev_code = 0;
	int prev_kana = 0;
	parse_auto_key(str, size, prev_code, prev_kana);
	if(prev_kana) {
		autokey_buffer->write(AUTO_KEY_KANA);
#ifdef USE_EMU_INHERENT_SPEC
		if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
	}

	autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
	autokey_shift = 0;
	autokey_enabled = true;

	update_config();

	return rc;
}

bool EMU::open_auto_key(const _TCHAR *file_path)
{
	FILEIO *fp;
	char buf[1025];
	bool rc = true;

	stop_auto_key();

	autokey_code = 0;
	fp = new FILEIO();
	int prev_kana = 0;
	if (fp->Fopen(file_path, FILEIO_READ_BINARY)) {
		autokey_buffer->clear();
		int prev_code = 0;
		int size = 0;
		while((size = (int)fp->Fread(buf, sizeof(char), 1024)) > 0) {
			parse_auto_key(buf, size, prev_code, prev_kana);
		}
		fp->Fclose();

		if(prev_kana) {
			autokey_buffer->write(AUTO_KEY_KANA);
#ifdef USE_EMU_INHERENT_SPEC
			if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
		}

		autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
		autokey_shift = 0;
		autokey_enabled = true;
		config.opened_autokey_path.Set(file_path, 0);
	} else {
		out_log(LOG_ERROR, _T("Auto key file couldn't be opened."));
		rc = false;
	}
	delete fp;

	config.initial_autokey_path.SetFromPath(file_path);

	update_config();

	return rc;
}

void EMU::parse_auto_key(const char *buf, int size, int &prev_code, int &prev_kana)
{
	for(int i = 0; i < size; i++) {
		int code = buf[i] & 0xff;
#ifdef USE_EMU_INHERENT_SPEC
		if(prev_code == 0xd && code == 0xa) {
			prev_code = code;
			continue;	// cr-lf
		}
#else
		if((0x81 <= code && code <= 0x9f) || 0xe0 <= code) {
			i++;	// kanji ?
			continue;
		}
		else if(code == 0xa) {
			continue;	// cr-lf
		}
#endif
		prev_code = code;

		if((code = autokey_table[code]) != 0) {
#ifdef USE_EMU_INHERENT_SPEC
			int kana = code & 0x3000;
			if(kana - prev_kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(kana - prev_kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
			if(prev_kana - kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(prev_kana - kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#else
			int kana = code & 0x200;
			if(prev_kana != kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#endif
			prev_kana = kana;
#if defined(USE_AUTO_KEY_NO_CAPS)
			if((code & AUTO_KEY_SHIFT_MASK) && !(code & (AUTO_KEY_UPPER_MASK | AUTO_KEY_LOWER_MASK))) {
#elif defined(USE_AUTO_KEY_CAPS)
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_LOWER_MASK)) {
#else
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_UPPER_MASK)) {
#endif
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_SHIFT_MASK);
			}
#ifdef USE_EMU_INHERENT_SPEC
			else if (code & AUTO_KEY_GRAPH_MASK) {
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_GRAPH_MASK);
			}
#endif
			else {
				autokey_buffer->write(code & AUTO_KEY_MASK);
			}
		}
	}
}

void EMU::stop_auto_key()
{
#ifndef USE_EMU_INHERENT_SPEC
	if(autokey_shift) {
		key_up(AUTO_KEY_SHIFT);
	}
#else
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_SHIFT);
	}
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_GRAPH);
	}
	if (vm_key_status) {
		for (int i=0; i<vm_key_status_size; i++) {
			vm_key_status[i] &= ~1;
		}
	}
#endif
	autokey_phase = (autokey_phase > 30 ? autokey_phase : 0);
	autokey_shift = 0;
	autokey_enabled = false;

	config.opened_autokey_path.Clear();

	update_config();
}
#endif /* USE_AUTO_KEY */
