/** @file win_gui.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2014.03.24 -

	@brief [ gui for windows ]
*/

#include <Windows.h>
#include <SDL_syswm.h>
#include "win_gui.h"
#include "../../config.h"
#include "winfont.h"
#include "../../emu.h"
#include "../../emumsg.h"
#include "win_filebox.h"
#include "win_seldrvbox.h"
#include "../../utility.h"
#include "../../sdl_main.h"
#ifdef USE_VKEYBOARD
#include "win_vkeyboard.h"
#endif

#ifdef GUI_TYPE_WINDOWS

using namespace GUI_WIN;

#define TRIM_STRING_SIZE	72

void GUI::update_recent_menu(HMENU hMenu, int menu_id, CRecentPathList &list)
{
	bool flag = false;
	HMENU hSubMenu = get_sub_menu(hMenu, menu_id);
	_TCHAR str[_MAX_PATH];
	if (hSubMenu) {
		while(DeleteMenu(hSubMenu, 0, MF_BYPOSITION) != 0) {}
		for(int i = 0; i < MAX_HISTORY && i < list.Count(); i++) {
			if (!GetRecentFileStr(list[i]->path, list[i]->num, str, TRIM_STRING_SIZE)) break;
			AppendMenu(hSubMenu, MF_STRING, menu_id + i, str);
			flag = true;
		}
		if(!flag) {
			AppendMenu(hSubMenu, MF_GRAYED | MF_STRING, menu_id, _("None"));
		}
	}
}

void GUI::update_multivolume_menu(HMENU hMenu, int drv, int menu_id)
{
	bool flag = false;
	HMENU hSubMenu = get_sub_menu(hMenu, menu_id);
	_TCHAR str[32];
	if (hSubMenu) {
		while(DeleteMenu(hSubMenu, 0, MF_BYPOSITION) != 0) {}
		d88_file_t *d88_file = GetD88File(drv);
		if(d88_file->bank_nums >= 1) {
			for(int i = 0; i < d88_file->bank_nums; i++) {
				GetMultiVolumeStr(i, d88_file->bank[i].name, str);
				AppendMenu(hSubMenu
					, MF_STRING | (d88_file->bank_nums <= 1 ? MF_GRAYED : 0) | (i == d88_file->cur_bank ? MF_CHECKED : 0)
					, menu_id + i, str);
			}
			flag = true;
		}
		if(!flag) {
			AppendMenu(hSubMenu, MF_GRAYED | MF_STRING, menu_id, _("None"));
		}
	}
}

HMENU GUI::get_sub_menu(HMENU hMenu, int id)
{
	HMENU hSubMenu = NULL;
	int count = GetMenuItemCount(hMenu);
	for(int n=count-1; n>=0; n--) {
		hSubMenu = GetSubMenu(hMenu,n);
		if (!hSubMenu) continue;

		MENUITEMINFO mii;
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = sizeof(mii);
		mii.fMask = MIIM_ID;
		if (!GetMenuItemInfo(hSubMenu, id, FALSE, &mii)) {
			hSubMenu = NULL;
			continue;
		}
		break;
	}
	return hSubMenu;
}

/// for locale
#define MSG_PTR_MAX 32
static char *msg_buf_ptr[MSG_PTR_MAX];
static int cur_msg_buf_ptr;

/****************************************

 GUI for windows class

 ****************************************/

GUI::GUI(int argc, char **argv, EMU *new_emu) : GUI_BASE(argc, argv, new_emu)
{
	hInstance = ::GetModuleHandle(NULL);

	font = new CFont();
	font->SetDefaultFont();

	configbox = new ConfigBox(hInstance, font, emu, this);
	keybindbox = new KeybindBox(hInstance, font, emu, this);
	volumebox = new VolumeBox(hInstance, font, emu, this);
	recvidbox = new RecVideoBox(hInstance, font, emu, this);
	recaudbox = new RecAudioBox(hInstance, font, emu, this);
	aboutbox = new AboutBox(hInstance, font, emu, this);

	KeybindControl::RegisterClass(hInstance);

	now_menu = false;
	now_menuloop = false;

	memset(msg_buf_ptr, 0, sizeof(msg_buf_ptr));
	cur_msg_buf_ptr = 0;

	OrigProcPtr = NULL;
}

GUI::~GUI()
{
	if (OrigProcPtr) {
		SetWindowLongPtr(hWindow, GWLP_WNDPROC, (LONG_PTR)OrigProcPtr);
	}

	for(int i=0; i<MSG_PTR_MAX; i++) {
		if (msg_buf_ptr[i]) delete [] msg_buf_ptr[i];
	}

	delete aboutbox;
	delete recaudbox;
	delete recvidbox;
	delete volumebox;
	delete keybindbox;
	delete configbox;
	delete font;

	KeybindControl::UnregisterClass(hInstance);
}

int GUI::CreateWidget(SDL_Surface *screen, int width, int height)
{
	GUI_BASE::CreateWidget(screen, width, height);
#ifndef USE_SDL2
	// enable to accept drag and drops
	::DragAcceptFiles(hWindow, TRUE);
#endif
	return 0;
}

int GUI::CreateMenu()
{
	GUI_BASE::CreateMenu();

	// Hook default window message procedure
	if (OrigProcPtr == NULL) {
		OrigProcPtr = (WNDPROC)GetWindowLongPtr(hWindow, GWLP_WNDPROC);
		SetWindowLongPtr(hWindow, GWLP_USERDATA, (LONG_PTR)this);
		SetWindowLongPtr(hWindow, GWLP_WNDPROC, (LONG_PTR)MessageProcStatic);
	}
	// disable system specific event because system event process on MessageProc
	SDL_EventState(SDL_SYSWMEVENT, SDL_DISABLE);
	// enable menu
	ShowMenu();
	// need resize window
	return 1;
}

void GUI::ShowMenu()
{
	if(hWindow && !now_menu) {
		HMENU hMenu = LoadMenu((HINSTANCE)GetModuleHandle(0), MAKEINTRESOURCE(IDR_MENU1));
		translate_menu(hMenu, 0);
		SetMenu(hWindow, hMenu);
		now_menu = true;
	}
}

void GUI::HideMenu()
{
	if(hWindow && now_menu) {
		HMENU hMenu = GetMenu(hWindow);
		SetMenu(hWindow, NULL);
		DestroyMenu(hMenu);
		now_menu = false;
	}
}

LRESULT CALLBACK GUI::MessageProcStatic(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	GUI *gui = (GUI *)GetWindowLongPtr(hWnd, GWLP_USERDATA);
	if (gui) return gui->MessageProc(hWnd, iMsg, wParam, lParam);
	else return DefWindowProc(hWnd, iMsg, wParam, lParam);
}

bool GUI::NeedUpdateScreen()
{
	if (need_update_screen > 0) {
		need_update_screen = 0;
//		PostMessage(hWindow, WM_USER_PAINT, 0, 0);
		UINT flags = RDW_INVALIDATE | RDW_INTERNALPAINT | RDW_NOERASE;
		RedrawWindow(hWindow, NULL, NULL, flags);
		// send signal to main loop
		SDL_CondSignal(cond_allow_update_screen);
		return true;
	} else {
		return false;
	}
}

void GUI::PreProcessEvent()
{
	// nothing to do
}

LRESULT GUI::MessageProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
#if 0
	if (iMsg >= 0 && iMsg < MAX_WMMSG) {
		emu->out_debugf("Msg: %s", wmtab[iMsg]);
	} else {
		emu->out_debugf("Msg: %x", iMsg);
	}
#endif
	switch(iMsg) {
	case WM_PAINT: {
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);
//		emu->out_debugf("WM_PAINT: %lx", hWnd);
		UpdateScreen();
		EndPaint(hWnd, &ps);
		return 0;
   }
//	case WM_USER_PAINT:
//		UpdateScreen();
//		return 0;

	case WM_COMMAND: {
		int id = LOWORD(wParam);
		// convert to SDL user message and post again
		emu->out_debugf(_T("PostCommandMessage: %d"), id);
		PostCommandMessage(id);
		return 0;
	}
	case WM_INITMENUPOPUP:
//		if(emu) {
//			emu->mute_sound(true);
//			SystemPause(true);
//		}
		update_menu((HMENU)wParam, LOWORD(lParam));
		return 0;

	case WM_ENTERMENULOOP:
//		if(emu) {
//			SystemPause(true);
//		}
		now_menuloop = true;
		return 0;

	case WM_EXITMENULOOP:
		if(emu->is_fullscreen() && now_menuloop) {
			HideMenu();
		}
		if (emu) {
			emu->key_lost_focus();
//			emu->mute_sound(false);
			SystemPause(false);
		}
		now_menuloop = false;
		return 0;

#ifndef USE_SDL2
	case WM_MOVE:
//		if(emu) {
//			emu->move_led();
//		}
		MoveLedBox();
		return 0;

	case WM_MOUSEMOVE:
		// TODO:
#if 0
		if(emu->is_fullscreen() && !now_menuloop) {
			POINTS p = MAKEPOINTS(lParam);
			if(p.y == 0) {
				ShowMenu();
			}
			else if(p.y > 32) {
				HideMenu();
			}
		}
#endif
		break;
	case WM_DROPFILES:
		OpenDroppedFile((void *)wParam);
		return 0;
#endif
	case WM_ENTERIDLE:
		return DefWindowProc(hWnd, iMsg, wParam, lParam);
	}

	if (OrigProcPtr) return CallWindowProc(OrigProcPtr, hWnd, iMsg, wParam, lParam);
	else return DefWindowProc(hWnd, iMsg, wParam, lParam);
}

void GUI::PostProcessEvent()
{
#ifndef USE_SDL2
	// When fullscreen, no grab
	ClipCursor(NULL);
#endif
}

void GUI::ScreenModeChanged(bool fullscreen)
{
	if (fullscreen) {
		HideMenu();
	} else {
		ShowMenu();
	}
}

void GUI::SetFocusToMainWindow()
{
	::SetFocus(hWindow);
}

bool GUI::ShowRecordVideoDialog(int fps_num)
{
	bool rc = false;
	PostEtSystemPause(true);
	if (recvidbox->Show(hWindow) == IDOK) {
		PostEtStartRecordVideo(fps_num);
		rc = true;
	} else {
		PostEtSystemPause(false);
	}
	SetFocusToMainWindow();
	return rc;
}

bool GUI::ShowRecordAudioDialog(void)
{
	bool rc = false;
	PostEtSystemPause(true);
	if (recaudbox->Show(hWindow) == IDOK) {
		PostEtStartRecordSound();
		rc = true;
	} else {
		PostEtSystemPause(false);
	}
	SetFocusToMainWindow();
	return rc;
}

bool GUI::ShowRecordVideoAndAudioDialog(int fps_num)
{
	bool rc = false;
	PostEtSystemPause(true);
	if (recvidbox->Show(hWindow, true) == IDOK && recaudbox->Show(hWindow) == IDOK) {
		PostEtStartRecordVideo(fps_num);
		rc = true;
	} else {
		PostEtSystemPause(false);
	}
	SetFocusToMainWindow();
	return rc;
}

#ifdef USE_EMU_INHERENT_SPEC
bool GUI::ShowConfigureDialog()
{
	bool rc = false;
	PostEtSystemPause(true);
	if (configbox->Show(hWindow) == IDOK) {
		rc = true;
	}
	PostEtSystemPause(false);
	SetFocusToMainWindow();
	return rc;
}

bool GUI::ShowKeybindDialog()
{
	bool rc = false;
	PostEtSystemPause(true);
	if (keybindbox->Show(hWindow) == IDOK) {
		emu->save_keybind();
		rc = true;
	}
	PostEtSystemPause(false);
	SetFocusToMainWindow();
	return rc;
}

/// create volume dialog
bool GUI::ShowVolumeDialog()
{
	volumebox->Show(hWindow);
	SetFocusToMainWindow();
	return true;
}
#endif

/// create about dialog
bool GUI::ShowAboutDialog()
{
	aboutbox->Show(hWindow);
	SetFocusToMainWindow();
	return true;
}

#ifdef USE_DATAREC
bool GUI::ShowLoadDataRecDialog()
{
	const _TCHAR *filter[] =
#if defined(DATAREC_PC8801)
		{ _("Supported Files (*.cas;*.cmt;*.t88)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_BINARY_ONLY)
		{ _("Supported Files (*.cas;*.cmt)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_TAP)
		{ _("Supported Files (*.wav;*.cas;*.tap)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_MZT)
		{ _("Supported Files (*.wav;*.cas;*.mzt;*.m12)"), _("All Files (*.*)"), NULL };
#elif defined(USE_EMU_INHERENT_SPEC)
		{ _("Supported Files (*.l3;*.l3b;*.l3c;*.wav;*.t9x)"), _("All Files (*.*)"), NULL };
#else
		{ _("Supported Files (*.wav;*.cas)"),_("All Files (*.*)"), NULL };
#endif

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Play Data Recorder Tape"),
		config.initial_datarec_path,
		_T("l3"),
		false
	);
	if (rc) {
		PostEtLoadDataRecMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

bool GUI::ShowSaveDataRecDialog()
{
	const _TCHAR *filter[] =
#if defined(DATAREC_PC8801)
		{ _("Supported Files (*.cas;*.cmt)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_BINARY_ONLY)
		{ _("Supported Files (*.cas;*.cmt)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_TAP)
		{ _("Supported Files (*.wav;*.cas)"), _("All Files (*.*)"), NULL };
#elif defined(DATAREC_MZT)
		{ _("Supported Files (*.wav;*.cas)"), _("All Files (*.*)"), NULL };
#elif defined(USE_EMU_INHERENT_SPEC)
		{ _("L3 File (*.l3)"), _("L3B File (*.l3b)"), _("L3C File (*.l3c)"), _("Wave File (*.wav)"), _("T9X File (*.t9x)"), _("All Files (*.*)"), NULL };
#else
		{ _("Supported Files (*.wav;*.cas)"), _("All Files (*.*)"), NULL };
#endif

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Record Data Recorder Tape"),
		config.initial_datarec_path,
		_T("l3"),
		true
	);
	if (rc) {
		PostEtSaveDataRecMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

void GUI::set_datarec_file_menu(HMENU hMenu, UINT uItem)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	if (emu->datarec_opened(true)) {
		_tcscpy(str,_("Play ["));
		_tcscat(str, UTILITY::trim_center(config.opened_datarec_path.path, TRIM_STRING_SIZE));
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_("Play..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);

	if (emu->datarec_opened(false)) {
		_tcscpy(str,_("Rec ["));
		_tcscat(str, UTILITY::trim_center(config.opened_datarec_path.path, TRIM_STRING_SIZE));
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_("Rec..."));
	}
	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);
	SetMenuItemInfo(hMenu, uItem + 1, FALSE, &minfo);

//	CheckMenuRadioItem(hMenu, uItem, uItem + 1, uItem + opened_tape_type - 1, MF_BYCOMMAND);
	CheckMenuItem(hMenu, uItem, IsOpenedLoadDataRecFile() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, uItem + 1, IsOpenedSaveDataRecFile() ? MF_CHECKED : MF_UNCHECKED);
}
#endif

#ifdef USE_FD1
bool GUI::ShowOpenFloppyDiskDialog(int drv)
{
	const _TCHAR *filter[] =
#ifndef USE_EMU_INHERENT_SPEC
		{ _("Supported Files (*.d88;*.d77;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)"), _("All Files (*.*)"), NULL };
#else
		{ _("Supported Files (*.d88;*.td0;*.imd;*.dsk;*.fdi;*.hdm;*.tfd;*.xdf;*.2d;*.sf7)"), _("All Files (*.*)"), NULL };
#endif

	DWORD flags;
	_TCHAR title[128];
	_stprintf(title, _("Open Floppy Disk %d"), drv);
	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		title,
		config.initial_disk_path,
		_T("d88"),
		false
	);
	flags = fbox.GetFlags();
	if(rc) {
		PostEtOpenFloppyMessage(drv, fbox.GetPath(), 0, flags, true);
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

int GUI::ShowSelectFloppyDriveDialog(int drv)
{
//	PostEtSystemPause(true);
	SelDrvBox dlg(hInstance, font, emu, this);
	dlg.SetDrive(drv);
	dlg.SetPrefix(_("FDD"));
	int new_drv = (int)dlg.Show(hWindow);
	return new_drv >= 0 ? new_drv : drv;
}

void GUI::set_disk_file_menu(HMENU hMenu, UINT uItem, int drv)
{
	MENUITEMINFO minfo;
	_TCHAR str[_MAX_PATH];

	minfo.cbSize = sizeof(MENUITEMINFO);
	minfo.fMask = MIIM_TYPE;
	minfo.fType = MFT_STRING;

	bool ins = InsertedFloppyDisk(drv);

	if (ins) {
		_tcscpy(str,_("Open ["));
		_tcscat(str,UTILITY::trim_center(config.opened_disk_path[drv].path, TRIM_STRING_SIZE));
		if (config.opened_disk_path[drv].num > 0) {
			_stprintf(&str[_tcslen(str)], _T(" : %d"), config.opened_disk_path[drv].num + 1);
		}
		_tcscat(str,_T("]"));
	} else {
		_tcscpy(str,_("Open..."));
	}

	minfo.dwTypeData = str;
	minfo.cch = (UINT)_tcslen(str);

	SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
	CheckMenuItem(hMenu, uItem, ins ? MF_CHECKED : MF_UNCHECKED);
}

#ifdef USE_EMU_INHERENT_SPEC
void GUI::set_disk_side_menu(HMENU hMenu, UINT uItem, int drv)
{
	if (emu) {
		MENUITEMINFO minfo;
		_TCHAR str[40];

		minfo.cbSize = sizeof(MENUITEMINFO);
		minfo.fMask = MIIM_TYPE;
		minfo.fType = MFT_STRING;

		int side = GetSideFloppyDisk(drv);
		if (side == 1 && config.fdd_type == 1) {
			_tcscpy(str,_("Change Side to A"));
		} else {
			_tcscpy(str,_("Change Side to B"));
		}
		minfo.dwTypeData = str;
		minfo.cch = (UINT)_tcslen(str);

		SetMenuItemInfo(hMenu, uItem, FALSE, &minfo);
		EnableMenuItem(hMenu, uItem, side >= 0 && config.fdd_type == 1 ? MF_ENABLED : MF_GRAYED);
	}
}
#endif

#endif	// USE_FD1

#ifdef USE_CART1
bool GUI::ShowOpenCartridgeDialog(int drv)
{
	const _TCHAR *filter[] =
#if defined(_GAMEGEAR)
		{ _("Supported Files (*.rom;*.bin;*.hex;*.gg;*.col)"), _("All Files (*.*)"), NULL };
#elif defined(_MASTERSYSTEM)
		{ _("Supported Files (*.rom;*.bin;*.hex;*.sms)"), _("All Files (*.*)"), NULL };
#elif defined(_PC6001) || defined(_PC6001MK2) || defined(_PC6001MK2SR) || defined(_PC6601) || defined(_PC6601SR)
		{ _("Supported Files (*.rom;*.bin;*.hex;*.60)"), _("All Files (*.*)"), NULL };
#elif defined(_PCENGINE) || defined(_X1TWIN)
		{ _("Supported Files (*.rom;*.bin;*.hex;*.pce)"), _("All Files (*.*)"), NULL };
#else
		{ _("Supported Files (*.rom;*.bin;*.hex)"), _("All Files (*.*)"), NULL };
#endif

	const _TCHAR *title =
#if defined(_PCENGINE) || defined(_X1TWIN)
		_("Open HuCARD");
#else
		_("Open Cartridge");
#endif

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		title,
		config.initial_cart_path,
		_T("rom"),
		false
	);
	if (rc) {
		PostEtOpenCartridgeMessage(drv, fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifdef USE_QD1
bool GUI::ShowOpenQuickDiskDialog(int drv)
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.mzt;*.q20;*.qdf)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Open Quick Disk"),
		config.initial_quickdisk_path,
		_T("mzt"),
		false
	);
	if (rc) {
		PostEtOpenQuickDiskMessage(drv, fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifdef USE_MEDIA
bool GUI::ShowOpenMediaDialog()
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.bin)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Open Media"),
		config.initial_media_path,
		_T("bin"),
		false
	);
	if (rc) {
		PostEtOpenQuickDiskMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifdef USE_BINARY_FILE1
bool GUI::ShowLoadBinaryDialog(int drv)
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.ram;*.bin;*.hex)"), _("All Files (*.*)"), NULL };

	const _TCHAR *title =
#if defined(_PASOPIA) || defined(_PASOPIA7)
		_("Load RAM Pack Cartridge");
#else
		_("Load Memory Dump");
#endif

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		title,
		config.initial_binary_path,
		_T("ram"),
		false
	);
	if (rc) {
		PostEtLoadBinaryMessage(drv, fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

bool GUI::ShowSaveBinaryDialog(int drv)
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.ram;*.bin;*.hex)"), _("All Files (*.*)"), NULL };

	const _TCHAR *title =
#if defined(_PASOPIA) || defined(_PASOPIA7)
		_("Save RAM Pack Cartridge");
#else
		_("Save Memory Dump");
#endif

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		title,
		config.initial_binary_path,
		_T("ram"),
		true
	);
	if (rc) {
		PostEtSaveBinaryMessage(drv, fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifdef USE_AUTO_KEY
bool GUI::ShowOpenAutoKeyDialog()
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.txt;*.bas;*.lpt)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Open Text File"),
		config.initial_autokey_path,
		_T("txt"),
		false
	);
	if(rc) {
		PostEtLoadAutoKeyMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifdef USE_EMU_INHERENT_SPEC
bool GUI::ShowSaveStateDialog(bool cont)
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.l3r)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	config.saved_state_path.Clear();
	bool rc = fbox.Show(
		filter,
		_("Save Status Data"),
		config.initial_state_path,
		_T("l3r"),
		true
	);
	if (rc) {
		PostEtSaveStatusMessage(fbox.GetPath(), cont);
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

bool GUI::ShowLoadStateDialog()
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.l3r)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Load Status Data"),
		config.initial_state_path,
		_T("l3r"),
		false
		);
	if(rc) {
		PostEtLoadStatusMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

#ifdef USE_KEY_RECORD
bool GUI::ShowPlayRecKeyDialog()
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.l3k)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Play Recorded Keys"),
		config.initial_state_path,
		_T("l3k"),
		false
		);
	if(rc) {
		PostEtLoadRecKeyMessage(fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}

bool GUI::ShowRecordRecKeyDialog()
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.l3k)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Record Input Keys"),
		config.initial_state_path,
		_T("l3k"),
		true
		);
	if(rc) {
		PostEtSaveRecKeyMessage(fbox.GetPath(), false);
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
bool GUI::ShowRecordStateAndRecKeyDialog()
{
	bool rc = ShowSaveStateDialog(true);
	if (!rc) return rc;
	return ShowRecordRecKeyDialog();
}
#endif
#endif	// USE_EMU_INHERENT_SPEC

#ifdef USE_PRINTER
bool GUI::ShowSavePrinterDialog(int drv)
{
	const _TCHAR *filter[] =
		{ _("Supported Files (*.lpt)"), _("All Files (*.*)"), NULL };

	FileBox fbox(hWindow);

	PostEtSystemPause(true);
	bool rc = fbox.Show(
		filter,
		_("Save Printing Data"),
		config.initial_printer_path,
		_T("lpt"),
		true
	);
	if (rc) {
		PostEtSavePrinterMessage(drv, fbox.GetPath());
	} else {
		PostEtSystemPause(false);
	}
	return rc;
}
#endif

#ifndef USE_SDL2
/**
 *	open dropped file
 */
bool GUI::OpenDroppedFile(void *param)
{
	_TCHAR dropped_file[_MAX_PATH];

	HDROP hDrop=(HDROP)param;
	// get first filename
	DragQueryFile(hDrop, 0, dropped_file, _MAX_PATH);
    DragFinish(hDrop);
	if (_tcslen(dropped_file) == 0) {
		return false;
	}

	return OpenFileByExtention(dropped_file);
}
#endif

//
//
void GUI::update_control_menu(HMENU hMenu)
{
	CheckMenuItem(hMenu, ID_RESET, NowPowerOff() ? MF_UNCHECKED : MF_CHECKED); 

#ifndef _MBS1
#ifdef USE_DIPSWITCH
	for(int i = 0; i < 8; i++) {
#ifdef _BML3MK5
		CheckMenuItem(hMenu, ID_DIPSWITCH1 + i, (GetDipswitch() & (1 << i)) ? MF_CHECKED : MF_UNCHECKED);
#else
		CheckMenuItem(hMenu, ID_DIPSWITCH1 + i, !(GetDipswitch() & (1 << i)) ? MF_CHECKED : MF_UNCHECKED);
#endif
	}
#endif
#endif
#ifdef USE_EMU_INHERENT_SPEC
#ifdef _MBS1
	CheckMenuRadioItem(hMenu, ID_SYSTEM_MODE_1, ID_SYSTEM_MODE_2, ID_SYSTEM_MODE_2 - GetSystemMode(), MF_BYCOMMAND);
#endif
	CheckMenuRadioItem(hMenu, ID_FDD_TYPE_1, ID_FDD_TYPE_4, ID_FDD_TYPE_1 + NextFddType(), MF_BYCOMMAND);
#endif
#ifdef _HC80
	if(config.device_type >= 0 && config.device_type < 3) {
		CheckMenuRadioItem(hMenu, ID_HC80_RAMDISK0, ID_HC80_RAMDISK2, ID_HC80_RAMDISK0 + config.device_type, MF_BYCOMMAND);
	}
#endif
#ifdef _MZ800
	if(config.boot_mode >= 0 && config.boot_mode < 2) {
		CheckMenuRadioItem(hMenu, ID_MZ800_MODE_MZ800, ID_MZ800_MODE_MZ700, ID_MZ800_MODE_MZ800 + config.boot_mode, MF_BYCOMMAND);
	}
#endif
#ifdef _PASOPIA
	if(config.boot_mode >= 0 && config.boot_mode < 5) {
		CheckMenuRadioItem(hMenu, ID_PASOPIA_MODE_TBASIC_V1_0, ID_PASOPIA_MODE_MINI_PASCAL, ID_PASOPIA_MODE_TBASIC_V1_0 + config.boot_mode, MF_BYCOMMAND);
	}
	if(config.device_type >= 0 && config.boot_mode < 3) {
		CheckMenuRadioItem(hMenu, ID_PASOPIA_DEVICE_RAM_PAC, ID_PASOPIA_DEVICE_JOYSTICK, ID_PASOPIA_DEVICE_RAM_PAC + config.device_type, MF_BYCOMMAND);
	}
#endif
#ifdef _PC98DO
	if(config.boot_mode >= 0 && config.boot_mode < 5) {
		CheckMenuRadioItem(hMenu, ID_PC98DO_MODE_PC98, ID_PC8801_MODE_N, ID_PC98DO_MODE_PC98 + config.boot_mode, MF_BYCOMMAND);
	}
#endif
#ifdef _PC8801MA
	if(config.boot_mode >= 0 && config.boot_mode < 4) {
		CheckMenuRadioItem(hMenu, ID_PC8801_MODE_V1S, ID_PC8801_MODE_N, ID_PC8801_MODE_V1S + config.boot_mode, MF_BYCOMMAND);
	}
	if(config.device_type >= 0 && config.boot_mode < 3) {
//		CheckMenuRadioItem(hMenu, ID_PC8801_DEVICE_JOYSTICK, ID_PC8801_DEVICE_JOYMOUSE, ID_PC8801_DEVICE_JOYSTICK + config.boot_mode, MF_BYCOMMAND);
		// joymouse is not supported yet...
		CheckMenuRadioItem(hMenu, ID_PC8801_DEVICE_JOYSTICK, ID_PC8801_DEVICE_MOUSE, ID_PC8801_DEVICE_JOYSTICK + config.device_type, MF_BYCOMMAND);
	}
#endif
#if defined(_PC9801E) || defined(_PC9801VM) || defined(_PC98DO) || defined(_PC8801MA)
	if(config.cpu_clock_low) {
		CheckMenuRadioItem(hMenu, ID_PC9801_CPU_CLOCK_HIGH, ID_PC9801_CPU_CLOCK_LOW, ID_PC9801_CPU_CLOCK_LOW, MF_BYCOMMAND);
	}
	else {
		CheckMenuRadioItem(hMenu, ID_PC9801_CPU_CLOCK_HIGH, ID_PC9801_CPU_CLOCK_LOW, ID_PC9801_CPU_CLOCK_HIGH, MF_BYCOMMAND);
	}
#endif
	if(config.cpu_power >= 0 && config.cpu_power <= 5) {
		CheckMenuRadioItem(hMenu, ID_CPU_POWER0, ID_CPU_POWER5, ID_CPU_POWER0 + GetCPUPower(), MF_BYCOMMAND);
	}
#ifdef USE_EMU_INHERENT_SPEC
	CheckMenuItem(hMenu, ID_PAUSE,    NowPause() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SYNC_IRQ, NowSyncIRQ() ? MF_CHECKED : MF_UNCHECKED);
#ifdef _MBS1
	CheckMenuItem(hMenu, ID_MEMORY_NOWAIT, NowMemNoWait() ? MF_CHECKED : MF_UNCHECKED);
#endif
#endif
#ifdef USE_AUTO_KEY
	// auto key
	bool now_paste = true;
	if(emu) {
		now_paste = IsRunningAutoKey();
	}
	EnableMenuItem(hMenu, ID_AUTOKEY_OPEN, now_paste ? MF_GRAYED : MF_ENABLED);
	EnableMenuItem(hMenu, ID_AUTOKEY_START, now_paste ? MF_GRAYED : MF_ENABLED);
#endif
#ifdef USE_EMU_INHERENT_SPEC
#ifdef USE_KEY_RECORD
	CheckMenuItem(hMenu, ID_RECKEY_PLAY, NowPlayingRecKey() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_RECKEY_REC,  NowRecordingRecKey() ? MF_CHECKED : MF_UNCHECKED);
#endif
	update_recent_menu(hMenu, ID_RECENT_STATE, config.recent_state_path);
#endif
}

//
//
void GUI::update_fdd_menu(HMENU hMenu, int drv, int ID_OPEN_FD, int ID_CHANGE_FD, int ID_WRITEPROTECT_FD, int ID_RECENT_FD, int ID_SELECT_D88_BANK)
{
	update_multivolume_menu(hMenu, drv, ID_SELECT_D88_BANK);
	update_recent_menu(hMenu, ID_RECENT_FD, config.recent_disk_path[drv]);
	set_disk_side_menu(hMenu, ID_CHANGE_FD, drv);
	set_disk_file_menu(hMenu, ID_OPEN_FD, drv);
	EnableMenuItem(hMenu, ID_WRITEPROTECT_FD, InsertedFloppyDisk(drv) ? MF_ENABLED : MF_GRAYED);
	CheckMenuItem(hMenu, ID_WRITEPROTECT_FD, WriteProtectedFloppyDisk(drv) ? MF_CHECKED : MF_UNCHECKED);
}

//
//
void GUI::update_tape_menu(HMENU hMenu)
{
	update_recent_menu(hMenu, ID_RECENT_DATAREC, config.recent_datarec_path);

	set_datarec_file_menu(hMenu, ID_PLAY_DATAREC);
	// realmode
	CheckMenuItem(hMenu, ID_REAL_DATAREC, NowRealModeDataRec() ? MF_CHECKED : MF_UNCHECKED);
}

//
//
void GUI::update_screen_menu(HMENU hMenu)
{
	_TCHAR buf[_MAX_PATH];

	// recording
	bool now_rec = true;
	if(emu) {
		now_rec = NowRecordingVideo();
	}
#ifdef USE_EMU_INHERENT_SPEC
	// rec video size
	InsertMenu(hMenu, ID_SCREEN_REC_SIZE1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
	for(int i = 0; i <= (ID_SCREEN_REC_SIZE2 - ID_SCREEN_REC_SIZE1); i++) {
		DeleteMenu(hMenu, ID_SCREEN_REC_SIZE1 + i, MF_BYCOMMAND);
	}
	for(int i = 0; i <= (ID_SCREEN_REC_SIZE2 - ID_SCREEN_REC_SIZE1); i++) {
		GetRecordVideoSizeStr(i, buf);
		InsertMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND | MF_STRING, ID_SCREEN_REC_SIZE1 + i, buf);
	}
	DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);
	CheckMenuRadioItem(hMenu, ID_SCREEN_REC_SIZE1, ID_SCREEN_REC_SIZE2, ID_SCREEN_REC_SIZE1 + config.screen_video_size, MF_BYCOMMAND);
	for(UINT i = ID_SCREEN_REC_SIZE1; i<= ID_SCREEN_REC_SIZE2; i++) {
		EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
	}
#endif
	for(UINT i = ID_SCREEN_REC60; i<= ID_SCREEN_REC10; i++) {
		EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
	}
	if (now_rec) {
		CheckMenuRadioItem(hMenu, ID_SCREEN_REC60, ID_SCREEN_REC10, ID_SCREEN_REC60 + rec_fps_no, MF_BYCOMMAND);
	} else {
		CheckMenuRadioItem(hMenu, ID_SCREEN_REC60, ID_SCREEN_REC10, ID_SCREEN_REC60 - 1, MF_BYCOMMAND);
	}
//	EnableMenuItem(hMenu, ID_SCREEN_STOP, now_stop ? MF_GRAYED : MF_ENABLED);

#ifdef USE_EMU_INHERENT_SPEC
	// frame rate
	CheckMenuRadioItem(hMenu, ID_SCREEN_VFRAME, ID_SCREEN_FPS10, ID_SCREEN_VFRAME + config.fps_no + 1, MF_BYCOMMAND);
//	for(UINT i = ID_SCREEN_VFRAME; i<= ID_SCREEN_FPS10; i++) {
//		EnableMenuItem(hMenu, i, now_rec ? MF_GRAYED : MF_ENABLED);
//	}
#endif

	// screen mode
	UINT last = ID_SCREEN_WINDOW1;
	InsertMenu(hMenu, ID_SCREEN_WINDOW1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
	for(int i = 0; i < WINDOW_MODE_MAX; i++) {
		DeleteMenu(hMenu, ID_SCREEN_WINDOW1 + i, MF_BYCOMMAND);
	}
	for(int i = 0; i < GetWindowModeCount(); i++) {
		GetWindowModeStr(i, buf);
		InsertMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND | MF_STRING, ID_SCREEN_WINDOW1 + i, buf);
		last = ID_SCREEN_WINDOW1 + i;
	}
	DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);

	InsertMenu(hMenu, ID_SCREEN_FULLSCREEN1, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
	for(int i = 0; i < SCREEN_MODE_MAX; i++) {
		DeleteMenu(hMenu, ID_SCREEN_FULLSCREEN1 + i, MF_BYCOMMAND);
	}
	for(int i = 0; i < GetFullScreenModeCount(); i++) {
		MENUITEMINFO info;
		ZeroMemory(&info, sizeof(info));
		info.cbSize = sizeof(info);
		GetFullScreenModeStr(i, buf);
		info.fMask = MIIM_TYPE | MIIM_ID;
		info.fType = MFT_STRING;
		info.wID = ID_SCREEN_FULLSCREEN1 + i;
		info.dwTypeData = buf;
		InsertMenuItem(hMenu, ID_SCREEN_BOTTOM, FALSE, &info);
		EnableMenuItem(hMenu, ID_SCREEN_FULLSCREEN1 + i, IsFullScreen() ? MF_GRAYED : MF_ENABLED);
		last = ID_SCREEN_FULLSCREEN1 + i;
	}
	DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);

	if(config.window_mode >= 0 && config.window_mode < GetWindowModeCount()) {
		CheckMenuRadioItem(hMenu, ID_SCREEN_WINDOW1, last, ID_SCREEN_WINDOW1 + config.window_mode, MF_BYCOMMAND);
	}
	else if(config.window_mode >= 8 && config.window_mode < GetFullScreenModeCount() + 8) {
		CheckMenuRadioItem(hMenu, ID_SCREEN_FULLSCREEN1, last, ID_SCREEN_FULLSCREEN1 + config.window_mode - 8, MF_BYCOMMAND);
	}
	// stretch screen
	CheckMenuItem(hMenu, ID_SCREEN_STRETCH, GetStretchScreen() == 1 ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_CUTOUT, GetStretchScreen() == 2 ? MF_CHECKED : MF_UNCHECKED);
	// pixel aspect ratio
	last = GetPixelAspectModeCount();
	InsertMenu(hMenu, ID_SCREEN_PIXEL_ASPECT0, MF_BYCOMMAND | MF_STRING, ID_SCREEN_BOTTOM, _T(""));
	for(UINT i = 0; i < last; i++) {
		DeleteMenu(hMenu, ID_SCREEN_PIXEL_ASPECT0 + i, MF_BYCOMMAND);
	}
	for(UINT i = 0; i < last; i++) {
		GetPixelAspectModeStr(i, buf);
		InsertMenuA(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND | MF_STRING, ID_SCREEN_PIXEL_ASPECT0 + i, buf);
	}
	DeleteMenu(hMenu, ID_SCREEN_BOTTOM, MF_BYCOMMAND);
	CheckMenuRadioItem(hMenu, ID_SCREEN_PIXEL_ASPECT0, ID_SCREEN_PIXEL_ASPECT0 + last - 1, ID_SCREEN_PIXEL_ASPECT0 + GetPixelAspectMode(), MF_BYCOMMAND);

#ifdef USE_MONITOR_TYPE
	if(config.monitor_type >= 0 && config.monitor_type < USE_MONITOR_TYPE) {
		CheckMenuRadioItem(hMenu, ID_SCREEN_MONITOR_TYPE0, ID_SCREEN_MONITOR_TYPE0 + USE_MONITOR_TYPE - 1, ID_SCREEN_MONITOR_TYPE0 + config.monitor_type, MF_BYCOMMAND);
	}
#elif defined(USE_SCREEN_ROTATE)
	if(config.monitor_type >= 0 && config.monitor_type < 2) {
		CheckMenuRadioItem(hMenu, ID_SCREEN_MONITOR_TYPE0, ID_SCREEN_MONITOR_TYPE1, ID_SCREEN_MONITOR_TYPE0 + config.monitor_type, MF_BYCOMMAND);
	}
#endif
#ifdef USE_SCANLINE
	// scanline
	CheckMenuRadioItem(hMenu, ID_SCREEN_SCANLINE0, ID_SCREEN_SCANLINE3, ID_SCREEN_SCANLINE0 + GetDrawMode(), MF_BYCOMMAND);
#endif
#ifdef USE_EMU_INHERENT_SPEC
	CheckMenuItem(hMenu, ID_SCREEN_AFTERIMAGE1, (GetAfterImageMode() & 1) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_AFTERIMAGE2, (GetAfterImageMode() & 2) ? MF_CHECKED : MF_UNCHECKED);
#ifdef USE_KEEPIMAGE
	CheckMenuItem(hMenu, ID_SCREEN_KEEPIMAGE1, (GetKeepImageMode() & 1) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_KEEPIMAGE2, (GetKeepImageMode() & 2) ? MF_CHECKED : MF_UNCHECKED);
#endif
#ifdef _MBS1
	CheckMenuRadioItem(hMenu, ID_SCREEN_DIGITAL, ID_SCREEN_ANALOG, ID_SCREEN_DIGITAL + GetRGBTypeMode(), MF_BYCOMMAND);
#endif
#ifdef USE_OPENGL
	CheckMenuItem(hMenu, ID_SCREEN_OPENGL_SYNC, (GetOpenGLMode() & 1) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_OPENGL_ASYNC, (GetOpenGLMode() & 2) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuRadioItem(hMenu, ID_SCREEN_OPENGL_FILTER0, ID_SCREEN_OPENGL_FILTER2, ID_SCREEN_OPENGL_FILTER0 + GetOpenGLFilter(), MF_BYCOMMAND);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_SYNC, MF_ENABLED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_ASYNC, MF_ENABLED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_FILTER0, MF_ENABLED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_FILTER1, MF_ENABLED);
#else
	CheckMenuItem(hMenu, ID_SCREEN_OPENGL_SYNC, MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_OPENGL_ASYNC, MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_SCREEN_OPENGL_FILTER, MF_UNCHECKED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_SYNC, MF_GRAYED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_ASYNC, MF_GRAYED);
	EnableMenuItem(hMenu, ID_SCREEN_OPENGL_FILTER, MF_GRAYED);
#endif
#ifdef USE_D3D
	CheckMenuItem(hMenu, ID_SCREEN_WAIT_VSYNC, config.wait_vsync ? MF_CHECKED : MF_UNCHECKED);
#endif
#endif
}

//
//
void GUI::update_sound_menu(HMENU hMenu)
{
	bool now_rec = false;
	if(emu) {
		now_rec = NowRecordingSound();
	}
	EnableMenuItem(hMenu, ID_SOUND_REC, now_rec ? MF_GRAYED : MF_ENABLED);
	CheckMenuItem(hMenu, ID_SOUND_REC, now_rec ? MF_CHECKED : MF_UNCHECKED);
//		EnableMenuItem(hMenu, ID_SOUND_STOP, now_stop ? MF_GRAYED : MF_ENABLED);

	CheckMenuRadioItem(hMenu, ID_SOUND_FREQ0, ID_SOUND_FREQ7, ID_SOUND_FREQ0 + GetSoundFrequencyNum(), MF_BYCOMMAND);
	CheckMenuRadioItem(hMenu, ID_SOUND_LATE0, ID_SOUND_LATE4, ID_SOUND_LATE0 + GetSoundLatencyNum(), MF_BYCOMMAND);
#ifdef USE_SOUND_DEVICE_TYPE
	if(config.sound_device_type >= 0 && config.sound_device_type < USE_SOUND_DEVICE_TYPE) {
		CheckMenuRadioItem(hMenu, ID_SOUND_DEVICE_TYPE0, ID_SOUND_DEVICE_TYPE0 + USE_SOUND_DEVICE_TYPE - 1, ID_SOUND_DEVICE_TYPE0 + config.sound_device_type, MF_BYCOMMAND);
	}
#endif
}

//
//
void GUI::update_device_menu(HMENU hMenu)
{
	for(int i=0; i<MAX_PRINTER; i++) {
		int j = i * (ID_PRINTER1_SAVE - ID_PRINTER0_SAVE);
		int printer_buffer_size = 0;
		// printer menu
		if(emu) {
			printer_buffer_size = GetPrinterBufferSize(i);
		}
		EnableMenuItem(hMenu, ID_PRINTER0_SAVE + j,  printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
		EnableMenuItem(hMenu, ID_PRINTER0_PRINT + j, printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
//		EnableMenuItem(hMenu, ID_PRINTER0_CLEAR + j, printer_buffer_size <= 0 ? MF_GRAYED : MF_ENABLED);
		CheckMenuItem(hMenu, ID_PRINTER0_DIRECT + j, IsEnablePrinterDirect(i) ? MF_CHECKED : MF_UNCHECKED);
	}
	for(int i=0; i<MAX_COMM; i++) {
		int j = i * (ID_COMM1_SERVER - ID_COMM0_SERVER);
		// comm menu
		CheckMenuItem(hMenu, ID_COMM0_SERVER + j, IsEnableCommServer(i) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_COMM0_CONNECT + j, NowConnectingComm(i) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_COMM0_THROUGH + j, NowCommThroughMode(i) ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hMenu, ID_COMM0_BINARY + j, NowCommBinaryMode(i) ? MF_CHECKED : MF_UNCHECKED);
	}
}

//
//
void GUI::update_option_menu(HMENU hMenu)
{
	CheckMenuItem(hMenu, ID_OPTIONS_LEDBOX_SHOW, IsShownLedBox() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_OPTIONS_LEDBOX_INSIDE, IsInsidedLedBox() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_OPTIONS_MSGBOARD, IsShownMessageBoard() ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_OPTIONS_JOYPAD0, IsEnableJoypad(1) ? MF_CHECKED : MF_UNCHECKED);
#ifdef USE_PIAJOYSTICK
	CheckMenuItem(hMenu, ID_OPTIONS_JOYPAD1, IsEnableJoypad(2) ? MF_CHECKED : MF_UNCHECKED);
#endif
#ifdef USE_LIGHTPEN
	CheckMenuItem(hMenu, ID_OPTIONS_LIGHTPEN, IsEnableLightpen() ? MF_CHECKED : MF_UNCHECKED);
#endif
	CheckMenuItem(hMenu, ID_OPTIONS_VKEYBOARD, IsShownVirtualKeyboard() ? MF_CHECKED : MF_UNCHECKED);
#ifdef USE_DEBUGGER
	EnableMenuItem(hMenu, ID_OPEN_DEBUGGER0,  IsDebuggerOpened(0) ? MF_GRAYED : MF_ENABLED);
#endif
}

//
//
void GUI::update_menu(HMENU hMenu, int pos)
{
	if (hMenu != GetSubMenu(GetMenu(hWindow), pos)) return;

#ifdef MENU_POS_CONTROL
	if(pos == MENU_POS_CONTROL) {
		// control menu
		update_control_menu(hMenu);
	}
#endif
#ifdef MENU_POS_CART1
	else if(pos == MENU_POS_CART1) {
		// cartridge
		update_recent_menu(hMenu, ID_RECENT_CART1, config.recent_cart_path[0]);
	}
#endif
#ifdef MENU_POS_FD1
	else if(pos == MENU_POS_FD1) {
		// floppy drive #1
		update_fdd_menu(hMenu, 0, ID_OPEN_FD1, ID_CHANGE_FD1, ID_WRITEPROTECT_FD1, ID_RECENT_FD1, ID_SELECT_D88_BANK1);
	}
#endif
#ifdef MENU_POS_FD2
	else if(pos == MENU_POS_FD2) {
		// floppy drive #2
		update_fdd_menu(hMenu, 1, ID_OPEN_FD2, ID_CHANGE_FD2, ID_WRITEPROTECT_FD2, ID_RECENT_FD2, ID_SELECT_D88_BANK2);
	}
#endif
#ifdef MENU_POS_FD3
	else if(pos == MENU_POS_FD3) {
		// floppy drive #3
		update_fdd_menu(hMenu, 2, ID_OPEN_FD3, ID_CHANGE_FD3, ID_WRITEPROTECT_FD3, ID_RECENT_FD3, ID_SELECT_D88_BANK3);
	}
#endif
#ifdef MENU_POS_FD4
	else if(pos == MENU_POS_FD4) {
		// floppy drive #4
		update_fdd_menu(hMenu, 3, ID_OPEN_FD4, ID_CHANGE_FD4, ID_WRITEPROTECT_FD4, ID_RECENT_FD4, ID_SELECT_D88_BANK4);
	}
#endif
#ifdef MENU_POS_FD5
	else if(pos == MENU_POS_FD5) {
		// floppy drive #5
		update_fdd_menu(hMenu, 4, ID_OPEN_FD5, ID_CHANGE_FD5, ID_WRITEPROTECT_FD5, ID_RECENT_FD5, ID_SELECT_D88_BANK5);
	}
#endif
#ifdef MENU_POS_FD6
	else if(pos == MENU_POS_FD6) {
		// floppy drive #6
		update_fdd_menu(hMenu, 5, ID_OPEN_FD6, ID_CHANGE_FD6, ID_WRITEPROTECT_FD6, ID_RECENT_FD6, ID_SELECT_D88_BANK6);
	}
#endif
#ifdef MENU_POS_QD1
	else if(pos == MENU_POS_QD1) {
		// quick disk drive
		update_recent_menu(hMenu, ID_RECENT_QD1, config.recent_quickdisk_path[0]);
	}
#endif
#ifdef MENU_POS_DATAREC
	else if(pos == MENU_POS_DATAREC) {
		// data recorder
		update_tape_menu(hMenu);
	}
#endif
#ifdef MENU_POS_MEDIA
	else if(pos == MENU_POS_MEDIA) {
		// media
		update_recent_menu(hMenu, ID_RECENT_MEDIA, config.recent_media_path);
	}
#endif
#ifdef MENU_POS_BINARY1
	else if(pos == MENU_POS_BINARY1) {
		// binary #1
		update_recent_menu(hMenu, ID_RECENT_BINARY1, config.recent_binary_path[0]);
	}
#endif
#ifdef MENU_POS_BINARY2
	else if(pos == MENU_POS_BINARY2) {
		// binary #2
		update_recent_menu(hMenu, ID_RECENT_BINARY2, config.recent_binary_path[1]);
	}
#endif
#ifdef MENU_POS_SCREEN
	else if(pos == MENU_POS_SCREEN) {
		// screen menu
		update_screen_menu(hMenu);
	}
#endif
#ifdef MENU_POS_SOUND
	else if(pos == MENU_POS_SOUND) {
		// sound menu
		update_sound_menu(hMenu);
	}
#endif
#ifdef MENU_POS_DEVICE
	else if(pos == MENU_POS_DEVICE) {
		// device menu
		update_device_menu(hMenu);
	}
#endif
#ifdef MENU_POS_OPTIONS
	else if(pos == MENU_POS_OPTIONS) {
		// options menu
		update_option_menu(hMenu);
	}
#endif
	DrawMenuBar(hWindow);
}

/// translate text string on the menu
void GUI::translate_menu(HMENU hMenu, int depth)
{
	_TCHAR buf[_MAX_PATH];
	MENUITEMINFO mii;

	if (depth >= 10) return;

	int nums = GetMenuItemCount(hMenu);
	for(int i=0; i<nums; i++) {
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = sizeof(mii);
		mii.fMask = MIIM_SUBMENU | MIIM_STRING;
		mii.dwTypeData = buf;
		mii.cch = _MAX_PATH - 1;
		GetMenuItemInfo(hMenu, i, TRUE, &mii);
		if (mii.cch > 0) {
			mii.fMask = MIIM_STRING;
			_tcscpy(buf, gettext_acp(mii.dwTypeData));
			mii.cch = (UINT)_tcslen(buf);
			SetMenuItemInfo(hMenu, i, TRUE, &mii);
		}
		if (mii.hSubMenu != NULL) {
			translate_menu(mii.hSubMenu, depth + 1);
		}
	}
}

const char *GUI::gettext_acp(const char *src)
{
	char buf[_MAX_PATH];
	wchar_t wbuf[_MAX_PATH];
	int len;

	len = ::MultiByteToWideChar(CP_UTF8, 0, gettext(src), -1, wbuf, _MAX_PATH);
	if (len <= 0) return src;
	len = ::WideCharToMultiByte(CP_ACP, 0, wbuf, -1, buf, _MAX_PATH, NULL, NULL);
	if (len <= 0) return src;

	cur_msg_buf_ptr = ((cur_msg_buf_ptr + 1) % MSG_PTR_MAX);
	if (msg_buf_ptr[cur_msg_buf_ptr] != NULL) {
		delete [] msg_buf_ptr[cur_msg_buf_ptr];
	}
	msg_buf_ptr[cur_msg_buf_ptr] = new char[len+4];
	strcpy(msg_buf_ptr[cur_msg_buf_ptr], buf);

	return msg_buf_ptr[cur_msg_buf_ptr];
}

#endif /* GUI_TYPE_WINDOWS */

#ifdef GUI_TYPE_AGAR
#include "../agar/ag_gui.h"
#endif
#ifdef USE_AUTO_KEY
bool GUI::StartAutoKey(void)
{
	bool rc = true;
	if(OpenClipboard(NULL)) {
		HANDLE hClip = GetClipboardData(CF_TEXT);
		if(hClip) {
			char *str = (char*)GlobalLock(hClip);
			emu->start_auto_key(str);
		}
		CloseClipboard();
	} else {
		rc = false;
	}
	return rc;
}
#endif

bool GUI::ShowVirtualKeyboard()
{
#ifdef USE_VKEYBOARD
	if (!vkeyboard) {
		HINSTANCE hInstance = ::GetModuleHandle(NULL);
		vkeyboard = new Vkbd::VKeyboard(hInstance, hWindow);

		uint8_t *buf;
		int siz;
		emu->get_vm_key_buffer(&buf, &siz);
		vkeyboard->SetStatusBufferPtr(buf, siz); 
		vkeyboard->Create(emu->resource_path());
		vkeyboard->Show();
	} else {
		vkeyboard->Close();
	}
	return true;
#else
	return false;
#endif
}
