/** @file logging.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.02.18 -

	@brief [ logging ]
*/

#include "logging.h"
#include "vm/vm.h"
#include "utility.h"

#ifdef _WIN32
#include "clocale.h"
extern CLocale *locale;
#endif

#define MESSAGE_BUFFER 1024

Logging::Logging()
{
	fplog = NULL;
	mux = new CMutex();
}

Logging::~Logging()
{
	delete mux;
}

void Logging::open_logfile(const _TCHAR *path)
{
#ifdef _LOG_FILE
	// get application path
	fplog = new FILEIO();
	if (fplog) {
		_TCHAR file_path[_MAX_PATH];
		_tcscpy(file_path, path);
		_tcscat(file_path, _T(CONFIG_NAME));
		_tcscat(file_path, _T(".log"));
		if (!fplog->Fopen(file_path, FILEIO_WRITE_ASCII)) {
			delete fplog;
			fplog = NULL;
		}
	}
#endif
}

void Logging::close_logfile()
{
#ifdef _LOG_FILE
	mux->lock();
	if(fplog) {
		delete fplog;
		fplog = NULL;
	}
	mux->unlock();
#endif
}

void Logging::out_logv(int level, const char* format, va_list ap)
{
	char buffer[MESSAGE_BUFFER];
	char levelstr[32];

	if (ap) {
#if (defined(_WIN32) && !defined(__MINGW32__))
		UTILITY::vsprintf_utf8(buffer, MESSAGE_BUFFER, format, ap);
#else
		vsprintf(buffer, format, ap);
#endif
	} else {
		strcpy(buffer, format);
	}

	switch(level) {
	case LOG_WARN:
		strcpy(levelstr, "WARN");
		break;
	case LOG_ERROR:
		strcpy(levelstr, "ERROR");
		break;
	case LOG_DEBUG:
		strcpy(levelstr, "DEBUG");
		break;
	default:
		strcpy(levelstr, "INFO");
		break;
	}

	mux->lock();
#ifdef _LOG_CONSOLE
	fprintf(stderr, "%s: %s\n", levelstr, buffer);
	fflush(stderr);
#endif

	send_log_message(level, levelstr, buffer);

#ifdef _LOG_FILE
	if(fplog) {
		fplog->Fprintf("%10d %-5s: %s\n", SDL_GetTicks(), levelstr, buffer);
		fplog->Flush();
	}
#endif
	mux->unlock();
}

void Logging::out_logf(int level, const char* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void Logging::out_log(int level, const char* msg)
{
	out_logv(level, msg, NULL);
}

void Logging::out_debugf(const char* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}

void Logging::out_debug(const char* msg)
{
#ifdef _DEBUG_LOG
	out_log(LOG_DEBUG, msg);
#endif
}

void Logging::out_syserrlog(int level, int err_num, const char* premsg)
{
	char msg[MESSAGE_BUFFER];
	strcpy(msg, premsg);
#ifdef _WIN32
	strcat(msg, " ");
	int len = (int)strlen(msg);
	if (FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL,
		err_num, LANGIDFROMLCID(locale->GetLcid()), &msg[len], MESSAGE_BUFFER - len, NULL) == 0) {
		sprintf(&msg[len], "%d", err_num);
	}
	for(int i=(int)strlen(msg)-1; i>=0; i--) {
		if (msg[i] != ' ' && msg[i] != '\t' && msg[i] != '\r' && msg[i] != '\n') break;
		msg[i]='\0';
	}
#endif
	out_logv(level, msg, NULL);
}

#ifdef _UNICODE

void Logging::out_logv(int level, const wchar_t* format, va_list ap)
{
	wchar_t buffer[MESSAGE_BUFFER];
	wchar_t levelstr[32];

	if (ap) {
		vswprintf(buffer, MESSAGE_BUFFER, format, ap);
	} else {
		wcscpy(buffer, format);
	}

	switch(level) {
	case LOG_WARN:
		wcscpy(levelstr, L"WARN");
		break;
	case LOG_ERROR:
		wcscpy(levelstr, L"ERROR");
		break;
	case LOG_DEBUG:
		wcscpy(levelstr, L"DEBUG");
		break;
	default:
		wcscpy(levelstr, L"INFO");
		break;
	}

	mux->lock();
#ifdef _LOG_CONSOLE
	wprintf(stderr, L"%ls: %ls\n", levelstr, buffer);
	fflush(stderr);
#endif

	send_log_message(level, levelstr, buffer);

#ifdef _LOG_FILE
	if(fplog) {
		fplog->Fprintf(L"%10d %-5ls: %s\n", SDL_GetTicks(), levelstr, buffer);
		fplog->Flush();
	}
#endif
	mux->unlock();
}

void Logging::out_logf(int level, const wchar_t* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void Logging::out_log(int level, const wchar_t* msg)
{
	out_logv(level, msg, NULL);
}

void Logging::out_debugf(const wchar_t* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}

void Logging::out_debug(const wchar_t* msg)
{
#ifdef _DEBUG_LOG
	out_log(LOG_DEBUG, msg);
#endif
}

void Logging::out_syserrlog(int level, int err_num, const wchar_t* premsg)
{
	wchar_t msg[MESSAGE_BUFFER];
	wcscpy(msg, premsg);
#ifdef _WIN32
	wcscat(msg, L" ");
	int len = (int)wcslen(msg);
	if (FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL,
		err_num, LANGIDFROMLCID(locale->GetLcid()), &msg[len], MESSAGE_BUFFER - len, NULL) == 0) {
		wsprintf(&msg[len], L"%d", err_num);
	}
	for(int i=(int)wcslen(msg)-1; i>=0; i--) {
		if (msg[i] != L' ' && msg[i] != L'\t' && msg[i] != L'\r' && msg[i] != L'\n') break;
		msg[i]=L'\0';
	}
#endif
	out_logv(level, msg, NULL);
}
#endif /* _UNICODE */
