/** @file parseopt.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ parse options ]
*/

#include "parseopt.h"
#include "config.h"
#include "vm/vm.h"
#include "utility.h"
#include "sdl_main.h"

static _TCHAR app_path[_MAX_PATH];
static _TCHAR app_name[_MAX_PATH];
static _TCHAR ini_path[_MAX_PATH];
static _TCHAR ini_file[_MAX_PATH];
static _TCHAR res_path[_MAX_PATH];

static _TCHAR *tape_file;
static _TCHAR *disk_file[USE_DRIVE];
static _TCHAR *state_file;
static _TCHAR *autokey_file;
static _TCHAR *reckey_file;
#ifdef USE_DEBUGGER
static int debugger_imm_start = 0;
#endif

const _TCHAR *get_app_path()
{
	return app_path;
}

const _TCHAR *get_app_name()
{
	return app_name;
}

const _TCHAR *get_ini_path()
{
	return ini_path;
}

const _TCHAR *get_ini_file()
{
	return ini_file;
}

const _TCHAR *get_res_path()
{
	return res_path;
}

static bool get_module_file_name(_TCHAR *path, int size);

/**
 *	get and parse options
 */
int get_options(int ac, char *av[])
{
	int opt = 0;
#ifdef _DEBUG
	for(int i=0; i<ac; i++) {
		printf("%d %s\n",i,av[i]);
	}
#endif
	memset(app_path, 0, sizeof(app_path));
	memset(ini_path, 0, sizeof(ini_path));
	memset(ini_file, 0, sizeof(ini_file));
	memset(res_path, 0, sizeof(res_path));

	// allocate buffer
	tape_file = new _TCHAR[_MAX_PATH];
	memset(tape_file, 0, sizeof(_TCHAR) * _MAX_PATH);
	for(int i=0; i<USE_DRIVE; i++) {
		disk_file[i] = new _TCHAR[_MAX_PATH];
		memset(disk_file[i], 0, sizeof(_TCHAR) * _MAX_PATH);
	}
	state_file = new _TCHAR[_MAX_PATH];
	memset(state_file, 0, sizeof(_TCHAR) * _MAX_PATH);
	autokey_file = new _TCHAR[_MAX_PATH];
	memset(autokey_file, 0, sizeof(_TCHAR) * _MAX_PATH);
	reckey_file = new _TCHAR[_MAX_PATH];
	memset(reckey_file, 0, sizeof(_TCHAR) * _MAX_PATH);

	// set application path and name
	_TCHAR **w_av;
#ifdef _UNICODE
	w_av = new _TCHAR*[ac];
	for(int i=0; i<ac; i++) {
		w_av[i] = new _TCHAR[_MAX_PATH];
		UTILITY::mbs_to_wcs(w_av[i], av[i]);
	}
#else
	w_av = av;
#endif
	if (get_module_file_name(app_path, _MAX_PATH)) {
		UTILITY::get_dir_and_basename(app_path, app_name);
	} else {
		UTILITY::get_dir_and_basename(w_av[0], app_path, app_name);
	}
	UTILITY::chomp_crlf(app_path);
	UTILITY::chomp_crlf(app_name);

	// resource path
	_tcsncpy(res_path, app_path, _MAX_PATH);
	_tcsncat(res_path, _T(RESDIR), _MAX_PATH);
	_tcscpy(res_path, UTILITY::slim_path(res_path));

#if defined(__APPLE__) && defined(__MACH__)
	// When mac, app_path set upper app folder (ex. foo/bar/baz.app/../)
	UTILITY::add_path_separator(app_path);
	UTILITY::slim_path(app_path);
	UTILITY::get_ancestor_dir(app_path, 3);
	// Xcode set debug options
	if (ac >= 3 && strstr(av[1],"NSDocument")) {
		optind += 2;
	}
#endif

#if !defined(_WIN32) || defined(__MINGW32__)
#ifdef USE_DEBUGGER
	while ((opt = getopt(ac, av, "hzp:a:d:i:s:t:k:")) != -1) {
#else
	while ((opt = getopt(ac, av, "hp:a:d:i:s:t:k:")) != -1) {
#endif
		///
		switch (opt) {
		case 'i':
			memset(ini_file, 0, sizeof(ini_file));
			_tcsncpy(ini_file, optarg, _MAX_PATH-1);
			break;
		case 'd':
			for(int drv=0; drv<USE_DRIVE; drv++) {
				if (disk_file[drv][0] == _T('\0')) {
					memset(disk_file[drv], 0, sizeof(_TCHAR) * _MAX_PATH);
					_tcsncpy(disk_file[drv], optarg, _MAX_PATH-1);
					break;
				}
			}
			break;
		case 't':
			memset(tape_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(tape_file, optarg, _MAX_PATH-1);
			break;
		case 's':
			memset(state_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(state_file, optarg, _MAX_PATH-1);
			break;
		case 'a':
			memset(autokey_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(autokey_file, optarg, _MAX_PATH-1);
			break;
		case 'k':
			memset(reckey_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(reckey_file, optarg, _MAX_PATH-1);
			break;
		case 'p':
			break;
#ifdef USE_DEBUGGER
		case 'z':
			debugger_imm_start = 1;
			break;
#endif
		case 'h':
//		case '?':
			{
				_TCHAR buf[_MAX_PATH];
				UTILITY::stprintf(buf, _MAX_PATH, _T("Usage: %s "), app_name);
				_tcscat(buf, _T("[-h] [-i <ini_file>] [-t <tape_file>] [-d <disk_file>] [-s <state_file>] [-a <autokey_file>] [-k <recordkey_file>]"));
#ifdef USE_DEBUGGER
				_tcscat(buf, _T(" [-z]"));
#endif
				_tcscat(buf, _T(" [<support_file> ...]"));
				_fputts(buf, stdout);
			}
			exit(0);
			break;
		default:
			break;
		}
	}
#else
	int optind = 1;
#endif
	for (;optind < ac && av[optind][0] != '\0'; optind++) {
		int rc = GUI::CheckSupportedFile(w_av[optind]);
		switch(rc) {
		case 1:
			memset(tape_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(tape_file, w_av[optind], _MAX_PATH - 1);
			break;
		case 2:
			for(int drv=0; drv<USE_DRIVE; drv++) {
				if (disk_file[drv][0] == _T('\0')) {
					memset(disk_file[drv], 0, sizeof(_TCHAR) * _MAX_PATH);
					_tcsncpy(disk_file[drv], w_av[optind], _MAX_PATH - 1);
					break;
				}
			}
			break;
		case 3:
			memset(state_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(state_file, w_av[optind], _MAX_PATH - 1);
			break;
		case 4:
			memset(autokey_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(autokey_file, w_av[optind], _MAX_PATH - 1);
			break;
		case 5:
			memset(ini_file, 0, sizeof(ini_file));
			_tcsncpy(ini_file, w_av[optind], _MAX_PATH - 1);
			break;
		case 6:
			memset(reckey_file, 0, sizeof(_TCHAR) * _MAX_PATH);
			_tcsncpy(reckey_file, w_av[optind], _MAX_PATH - 1);
			break;
		default:
			break;
		}
	}

	// check ini file
	if (ini_file[0] != _T('\0')) {
		if (UTILITY::check_file_extension(ini_file, _T("ini"))) {
			_TCHAR ini_name[_MAX_PATH];
			UTILITY::get_dir_and_basename(ini_file, ini_path, ini_name);
		}
		UTILITY::add_path_separator(ini_path);
	} else {
		_tcsncpy(ini_path, app_path, _MAX_PATH);
		_tcsncpy(ini_file, ini_path, _MAX_PATH);
		_tcsncat(ini_file, _T(CONFIG_NAME), _MAX_PATH);
		_tcsncat(ini_file, _T(".ini"), _MAX_PATH);
	}

#ifdef _UNICODE
	for(int i=0; i<ac; i++) {
		delete w_av[i];
	}
	delete w_av;
#endif

	return 0;
}

/**
 *	open recent file
 */
void open_recent_file(GUI *gui)
{
	_TCHAR path[_MAX_PATH];

	if (gui) {
		if (tape_file[0] != '\0') {
			gui->PostEtLoadDataRecMessage(tape_file);
		}
#ifdef USE_FD1
		// auto open recent file
		for(int drv=0; drv<USE_DRIVE; drv++) {
			path[0] = _T('\0');
			int bank_num = 0;
			if (disk_file[drv][0] != _T('\0') && _tcscmp(path, disk_file[drv]) != 0) {
				// specified file in the option
				path[0] = _T('\0');
				_tcscpy(path, disk_file[drv]);
			} else if ((config.mount_fdd & (1 << drv)) != 0 && config.recent_disk_path[drv].Count() > 0 && config.recent_disk_path[drv][0]->path.Length() > 0) {
				// recent file
				path[0] = _T('\0');
				_tcscpy(path, config.recent_disk_path[drv][0]->path);
				bank_num = config.recent_disk_path[drv][0]->num;
			}
			if (path[0] != _T('\0')) {
				gui->PostEtOpenFloppyMessage(drv, path, bank_num, 0, false);
			}
		}
#endif
		if (state_file[0] != _T('\0')) {
			gui->PostEtLoadStatusMessage(state_file);
		}
		if (autokey_file[0] != _T('\0')) {
			gui->PostEtLoadAutoKeyMessage(autokey_file);
		}
		if (reckey_file[0] != _T('\0')) {
			gui->PostEtLoadRecKeyMessage(reckey_file);
		}
#ifdef USE_DEBUGGER
		if (debugger_imm_start || config.debugger_imm_start) {
			gui->PostMtOpenDebugger(0);
		}
#endif
	}

	// release buffer
	delete[] tape_file; tape_file = NULL;
	for(int i=0; i<USE_DRIVE; i++) {
		delete[] disk_file[i]; disk_file[i] = NULL;
	}
	delete[] state_file; state_file = NULL;
	delete[] autokey_file; autokey_file = NULL;
	delete[] reckey_file; reckey_file = NULL;
}

bool get_module_file_name(_TCHAR *path, int size)
{
#if defined(_WIN32)
	return (::GetModuleFileName(NULL, path, size) > 0);
#else
	return false;
#endif
}
