/** @file sdl_debugger.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ debugger console ]

	@note Modified by Sasaji at 2016.02.01 -
*/

#include "sdl_debugger.h"

#ifdef USE_DEBUGGER

#include "gui/gui.h"
#include "csurface.h"

#define DEBUGGER_PROMPT _T("- ")

#if defined(_BML3MK5)
#define ADDR_FORMAT "%04X"
#define PHYS_ADDR_FORMAT "%04X"
#define DATA_FORMAT "%02X"
#elif defined(_MBS1)
#define ADDR_FORMAT "%04X"
#define PHYS_ADDR_FORMAT "%05X"
#define DATA_FORMAT "%02X"
#else
#define ADDR_FORMAT "%08X"
#define PHYS_ADDR_FORMAT "%08X"
#define DATA_FORMAT "%08X"
#endif
#define DATA_BYTE_FORMAT "%02X"

enum enDCommandsIdx {
	ID_DUMP_MEMORY = 1,
	ID_EDIT_MEMORY_BYTE,
	ID_EDIT_MEMORY_WORD,
	ID_EDIT_MEMORY_DWORD,
	ID_EDIT_MEMORY_ASCII,
#ifdef USE_EMU_INHERENT_SPEC
	ID_DUMP_PHYSICAL_MEMORY,
	ID_EDIT_PHYSICAL_MEMORY_BYTE,
	ID_EDIT_PHYSICAL_MEMORY_WORD,
	ID_EDIT_PHYSICAL_MEMORY_DWORD,
	ID_EDIT_PHYSICAL_MEMORY_ASCII,
#endif
#ifdef USE_IOPORT_ACCESS
	ID_INPUT_IO_BYTE,
	ID_INPUT_IO_WORD,
	ID_INPUT_IO_DWORD,
	ID_OUTPUT_IO_BYTE,
	ID_OUTPUT_IO_WORD,
	ID_OUTPUT_IO_DWORD,
#endif
	ID_ACCESS_REGISTER,
	ID_ACCESS_DEVICE,
	ID_SHOW_MEMORY_MAP,
#ifdef _MBS1
	ID_ACCESS_ADDRESS_MAP,
#endif
	ID_SEARCH,
#ifdef USE_EMU_INHERENT_SPEC
	ID_SEARCH_PHYSICAL,
#endif
	ID_UNASSEMBLE,
#ifdef USE_EMU_INHERENT_SPEC
	ID_UNASSEMBLE_PHYSICAL,
#endif
	ID_CALC_HEXA,
	ID_SET_FILE_NAME,
	ID_LOAD_BINARY_FILE,
	ID_SAVE_BINARY_FILE,
#ifdef USE_EMU_INHERENT_SPEC
	ID_LOAD_BINARY_FILE_PHYSICAL,
	ID_SAVE_BINARY_FILE_PHYSICAL,
#endif
#ifdef USE_EMU_INHERENT_SPEC
	ID_SAVE_IMAGE_FILE,
#endif
	ID_SET_BREAK_POINT,
	ID_CLEAR_BREAK_POINT,
	ID_DISABLE_BREAK_POINT,
	ID_ENABLE_BREAK_POINT,
	ID_LIST_BREAK_POINT,
#ifdef USE_IOPORT_ACCESS
	ID_SET_BREAK_POINT_FOR_IO,
	ID_CLEAR_BREAK_POINT_FOR_IO,
	ID_DISABLE_BREAK_POINT_FOR_IO,
	ID_ENABLE_BREAK_POINT_FOR_IO,
	ID_LIST_BREAK_POINT_FOR_IO,
#endif
	ID_SET_BREAK_POINT_FOR_INTR,
	ID_CLEAR_BREAK_POINT_FOR_INTR,
	ID_DISABLE_BREAK_POINT_FOR_INTR,
	ID_ENABLE_BREAK_POINT_FOR_INTR,
	ID_LIST_BREAK_POINT_FOR_INTR,
	ID_SET_TRACE_POINT,
	ID_CLEAR_TRACE_POINT,
	ID_DISABLE_TRACE_POINT,
	ID_ENABLE_TRACE_POINT,
	ID_LIST_TRACE_POINT,
#ifdef USE_IOPORT_ACCESS
	ID_SET_TRACE_POINT_FOR_IO,
	ID_CLEAR_TRACE_POINT_FOR_IO,
	ID_DISABLE_TRACE_POINT_FOR_IO,
	ID_ENABLE_TRACE_POINT_FOR_IO,
	ID_LIST_TRACE_POINT_FOR_IO,
#endif
	ID_SET_TRACE_POINT_FOR_INTR,
	ID_CLEAR_TRACE_POINT_FOR_INTR,
	ID_DISABLE_TRACE_POINT_FOR_INTR,
	ID_ENABLE_TRACE_POINT_FOR_INTR,
	ID_LIST_TRACE_POINT_FOR_INTR,
	ID_EXECUTE,
	ID_TRACE,
	ID_TRACE_BACK,
	ID_SHOW_CLOCK,
	ID_OUTPUT_LOG_FILE,
	ID_SEND_TO_EMULATOR,
	ID_USAGE,
	ID_QUIT,
	ID_END
};

/// command
struct st_commands_map {
	const _TCHAR   *cmd;
	enDCommandsIdx  idx;
};
static struct st_commands_map commands_map[] = {
	{ _T("D"),	ID_DUMP_MEMORY },
	{ _T("E"),	ID_EDIT_MEMORY_BYTE },
	{ _T("EB"),	ID_EDIT_MEMORY_BYTE },
	{ _T("EW"),	ID_EDIT_MEMORY_WORD },
	{ _T("ED"),	ID_EDIT_MEMORY_DWORD },
	{ _T("EA"),	ID_EDIT_MEMORY_ASCII },
#ifdef USE_EMU_INHERENT_SPEC
	{ _T("DP"),	ID_DUMP_PHYSICAL_MEMORY },
	{ _T("EP"),	ID_EDIT_PHYSICAL_MEMORY_BYTE },
	{ _T("EPB"),ID_EDIT_PHYSICAL_MEMORY_BYTE },
	{ _T("EPW"),ID_EDIT_PHYSICAL_MEMORY_WORD },
	{ _T("EPD"),ID_EDIT_PHYSICAL_MEMORY_DWORD },
	{ _T("EPA"),ID_EDIT_PHYSICAL_MEMORY_ASCII },
#endif
#ifdef USE_IOPORT_ACCESS
	{ _T("I"),	ID_INPUT_IO_BYTE },
	{ _T("IB"),	ID_INPUT_IO_BYTE },
	{ _T("IW"),	ID_INPUT_IO_WORD },
	{ _T("ID"),	ID_INPUT_IO_DWORD },
	{ _T("O"),	ID_OUTPUT_IO_BYTE },
	{ _T("OB"),	ID_OUTPUT_IO_BYTE },
	{ _T("OW"),	ID_OUTPUT_IO_WORD },
	{ _T("OD"),	ID_OUTPUT_IO_DWORD },
#endif
	{ _T("R"),	ID_ACCESS_REGISTER },
	{ _T("RD"),	ID_ACCESS_DEVICE },
	{ _T("M"),	ID_SHOW_MEMORY_MAP },
#ifdef _MBS1
	{ _T("X"),	ID_ACCESS_ADDRESS_MAP },
#endif
	{ _T("S"),	ID_SEARCH },
#ifdef USE_EMU_INHERENT_SPEC
	{ _T("SP"),	ID_SEARCH_PHYSICAL },
#endif
	{ _T("U"),	ID_UNASSEMBLE },
#ifdef USE_EMU_INHERENT_SPEC
	{ _T("UP"), ID_UNASSEMBLE_PHYSICAL },
#endif
	{ _T("H"),	ID_CALC_HEXA },
	{ _T("DN"),	ID_SET_FILE_NAME },
	{ _T("DL"),	ID_LOAD_BINARY_FILE },
	{ _T("DS"),	ID_SAVE_BINARY_FILE },
#ifdef USE_EMU_INHERENT_SPEC
	{ _T("DPL"),ID_LOAD_BINARY_FILE_PHYSICAL },
	{ _T("DPS"),ID_SAVE_BINARY_FILE_PHYSICAL },
#endif
#ifdef USE_EMU_INHERENT_SPEC
	{ _T("DIS"),ID_SAVE_IMAGE_FILE },
#endif
	{ _T("BP"),	ID_SET_BREAK_POINT },
	{ _T("RBP"),ID_SET_BREAK_POINT },
	{ _T("WBP"),ID_SET_BREAK_POINT },
	{ _T("BC"),	ID_CLEAR_BREAK_POINT },
	{ _T("RBC"),ID_CLEAR_BREAK_POINT },
	{ _T("WBC"),ID_CLEAR_BREAK_POINT },
	{ _T("BD"),	ID_DISABLE_BREAK_POINT },
	{ _T("RBD"),ID_DISABLE_BREAK_POINT },
	{ _T("WBD"),ID_DISABLE_BREAK_POINT },
	{ _T("BE"),	ID_ENABLE_BREAK_POINT },
	{ _T("RBE"),ID_ENABLE_BREAK_POINT },
	{ _T("WBE"),ID_ENABLE_BREAK_POINT },
//	{ _T("BL"),	ID_LIST_BREAK_POINT },
//	{ _T("RBL"),ID_LIST_BREAK_POINT },
//	{ _T("WBL"),ID_LIST_BREAK_POINT },
#ifdef USE_IOPORT_ACCESS
	{ _T("IBP"),ID_SET_BREAK_POINT_FOR_IO },
	{ _T("OBP"),ID_SET_BREAK_POINT_FOR_IO },
	{ _T("IBC"),ID_CLEAR_BREAK_POINT_FOR_IO },
	{ _T("OBC"),ID_CLEAR_BREAK_POINT_FOR_IO },
	{ _T("IBD"),ID_DISABLE_BREAK_POINT_FOR_IO },
	{ _T("OBD"),ID_DISABLE_BREAK_POINT_FOR_IO },
	{ _T("IBE"),ID_ENABLE_BREAK_POINT_FOR_IO },
	{ _T("OBE"),ID_ENABLE_BREAK_POINT_FOR_IO },
//	{ _T("IBL"),ID_LIST_BREAK_POINT_FOR_IO },
//	{ _T("OBL"),ID_LIST_BREAK_POINT_FOR_IO },
#endif
	{ _T("NBP"),ID_SET_BREAK_POINT_FOR_INTR },
	{ _T("NBC"),ID_CLEAR_BREAK_POINT_FOR_INTR },
	{ _T("NBD"),ID_DISABLE_BREAK_POINT_FOR_INTR },
	{ _T("NBE"),ID_ENABLE_BREAK_POINT_FOR_INTR },
//	{ _T("NBL"),ID_LIST_BREAK_POINT_FOR_INTR },
	{ _T("TP"),	ID_SET_TRACE_POINT },
	{ _T("RTP"),ID_SET_TRACE_POINT },
	{ _T("WTP"),ID_SET_TRACE_POINT },
	{ _T("TC"),	ID_CLEAR_TRACE_POINT },
	{ _T("RTC"),ID_CLEAR_TRACE_POINT },
	{ _T("WTC"),ID_CLEAR_TRACE_POINT },
	{ _T("TD"),	ID_DISABLE_TRACE_POINT },
	{ _T("RTD"),ID_DISABLE_TRACE_POINT },
	{ _T("WTD"),ID_DISABLE_TRACE_POINT },
	{ _T("TE"),	ID_ENABLE_TRACE_POINT },
	{ _T("RTE"),ID_ENABLE_TRACE_POINT },
	{ _T("WTE"),ID_ENABLE_TRACE_POINT },
//	{ _T("TL"),	ID_LIST_TRACE_POINT },
//	{ _T("RTL"),ID_LIST_TRACE_POINT },
//	{ _T("WTL"),ID_LIST_TRACE_POINT },
#ifdef USE_IOPORT_ACCESS
	{ _T("ITP"),ID_SET_TRACE_POINT_FOR_IO },
	{ _T("OTP"),ID_SET_TRACE_POINT_FOR_IO },
	{ _T("ITC"),ID_CLEAR_TRACE_POINT_FOR_IO },
	{ _T("OTC"),ID_CLEAR_TRACE_POINT_FOR_IO },
	{ _T("ITD"),ID_DISABLE_TRACE_POINT_FOR_IO },
	{ _T("OTD"),ID_DISABLE_TRACE_POINT_FOR_IO },
	{ _T("ITE"),ID_ENABLE_TRACE_POINT_FOR_IO },
	{ _T("OTE"),ID_ENABLE_TRACE_POINT_FOR_IO },
//	{ _T("ITL"),ID_LIST_TRACE_POINT_FOR_IO },
//	{ _T("OTL"),ID_LIST_TRACE_POINT_FOR_IO },
#endif
	{ _T("NTP"),ID_SET_TRACE_POINT_FOR_INTR },
	{ _T("NTC"),ID_CLEAR_TRACE_POINT_FOR_INTR },
	{ _T("NTD"),ID_DISABLE_TRACE_POINT_FOR_INTR },
	{ _T("NTE"),ID_ENABLE_TRACE_POINT_FOR_INTR },
//	{ _T("NTL"),ID_LIST_TRACE_POINT_FOR_INTR },
	{ _T("G"),	ID_EXECUTE },
	{ _T("T"),	ID_TRACE },
	{ _T("TB"),	ID_TRACE_BACK },
	{ _T("CL"),	ID_SHOW_CLOCK },
	{ _T(">"),	ID_OUTPUT_LOG_FILE },
	{ _T("!"),	ID_SEND_TO_EMULATOR },
	{ _T("?"),	ID_USAGE },
	{ _T("Q"),	ID_QUIT },
	{ NULL, ID_END }
};

#ifdef USE_CONSOLE_WINDOW
static BOOL WINAPI ctrl_c_handler(DWORD type)
{
	return TRUE;
}
#endif

///
DebuggerConsole::DebuggerConsole(volatile debugger_thread_t *p)
{
	dp = p;

	cpu = dp->vm->get_cpu(p->cpu_index);
	mem = dp->vm->get_memory();
	debugger = (DEBUGGER *)cpu->get_debugger();
	
	debugger->now_going = 0;
	debugger->now_debugging = true;
	// wait until vm process suspend
	while(!debugger->now_suspended && !NowDisable()) {
		CDelay(10);
	}

	logfile = NULL;
	logfilename[0] = _T('\0');

	prog_addr_mask = cpu->debug_prog_addr_mask();
	data_addr_mask = cpu->debug_data_addr_mask();
	phys_type = 0;
	dump_addr = 0;
	dump_phys_addr = 0;
	dasm_addr = cpu->get_next_pc();
	dasm_phys_addr = 0;

	UTILITY::stprintf(buffer, DC_MAX_BUFFER_LEN, _T("Debugger - %s"), _T(DEVICE_NAME));
	cp932 = false;

#ifdef USE_CONSOLE_WINDOW
	AllocConsole();
	SetConsoleTitle(buffer);
	SetConsoleCtrlHandler(ctrl_c_handler, TRUE);
	
	hStdIn = GetStdHandle(STD_INPUT_HANDLE);
	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	cp932 = (GetConsoleCP() == 932);
	
	COORD coord;
	coord.X = 80;
	coord.Y = 4000;
	
	SetConsoleScreenBufferSize(hStdOut, coord);
	SetConsoleTextAttribute(hStdOut, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
	RemoveMenu(GetSystemMenu(GetConsoleWindow(), FALSE), SC_CLOSE, MF_BYCOMMAND);
#endif

	echo_bak = true;
		
	memset(command, 0, sizeof(command));
	memset(prev_command, 0, sizeof(prev_command));
	commandptr = 0;

	paramnum = 0;

	memset(key_stat_tmp, 0, KEY_STATUS_SIZE);
}

DebuggerConsole::~DebuggerConsole()
{
	// release logfile
	if(logfile != NULL) {
		if(logfile->IsOpened()) {
			logfile->Fclose();
		}
		delete logfile;
		logfile = NULL;
	}

#ifdef USE_CONSOLE_WINDOW
	// release console
	SetConsoleCtrlHandler(ctrl_c_handler, FALSE);
	FreeConsole();
#endif
}

#ifdef USE_CONSOLE_WINDOW
const WORD cw_color_map[DebuggerConsole::EndColor] = {
	0,
	FOREGROUND_BLUE | FOREGROUND_INTENSITY,
	FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_BLUE | FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY,
};
#endif
#ifdef USE_TELNET_SERVER
const _TCHAR *ts_color_map[DebuggerConsole::EndColor] = {
	_T("\x1b[30m"),
	_T("\x1b[34m"),
	_T("\x1b[31m"),
	_T("\x1b[35m"),
	_T("\x1b[32m"),
	_T("\x1b[36m"),
	_T("\x1b[33m"),
	_T("\x1b[37m")
};
#endif

void DebuggerConsole::SetTextColor(enTextColor color)
{
	if (color < EndColor) {
#ifdef USE_CONSOLE_WINDOW
		SetConsoleTextAttribute(hStdOut, cw_color_map[color]);
#endif
#ifdef USE_TELNET_SERVER
		debugger->write_data(ts_color_map[color], (int)_tcslen(ts_color_map[color]));
#endif
	}
}

void DebuggerConsole::Printf(enTextColor color, const _TCHAR *format, ...)
{
	va_list ap;
	SetTextColor(color);
	
	va_start(ap, format);
	Vprintf(format, ap);
	va_end(ap);
}

void DebuggerConsole::Printf(const _TCHAR *format, ...)
{
	va_list ap;
	
	va_start(ap, format);
	Vprintf(format, ap);
	va_end(ap);
}

void DebuggerConsole::Vprintf(const _TCHAR *format, va_list ap)
{
	_TCHAR nbuffer[DC_MAX_BUFFER_LEN];

	UTILITY::vstprintf(nbuffer, DC_MAX_BUFFER_LEN, format, ap);

	if(logfile != NULL && logfile->IsOpened()) {
		logfile->Fwrite(nbuffer, sizeof(_TCHAR), _tcslen(nbuffer));
	}
#ifdef USE_CONSOLE_WINDOW
	DWORD dwWritten;
	WriteConsole(hStdOut, nbuffer, (DWORD)_tcslen(nbuffer), &dwWritten, NULL);
#endif
#ifdef USE_TELNET_SERVER
	if (echo_bak) debugger->write_data(nbuffer, (int)_tcslen(nbuffer));
#endif
#ifdef USE_STANDARD_INOUT
	_fputts(nbuffer, stdout);
#endif
}

void DebuggerConsole::Print(enTextColor color, const _TCHAR *str, bool cr)
{
	SetTextColor(color);
	Print(str, cr);
}

void DebuggerConsole::Print(const _TCHAR *str, bool cr)
{
	if(logfile != NULL && logfile->IsOpened()) {
		logfile->Fwrite(str, sizeof(_TCHAR), _tcslen(str));
	}
#ifdef USE_CONSOLE_WINDOW
	DWORD dwWritten;
	WriteConsole(hStdOut, str, (DWORD)_tcslen(str), &dwWritten, NULL);
#endif
#ifdef USE_TELNET_SERVER
	if (echo_bak) {
		debugger->write_data(str, (int)_tcslen(str));
	}
#endif
#ifdef USE_STANDARD_INOUT
	_fputts(str, stdout);
#endif
	if (cr) Cr();
}

void DebuggerConsole::Out(enTextColor color, bool cr)
{
	SetTextColor(color);
	Out(cr);
}

void DebuggerConsole::Out(bool cr)
{
	if(logfile != NULL && logfile->IsOpened()) {
		logfile->Fwrite(buffer, sizeof(_TCHAR), _tcslen(buffer));
	}
#ifdef USE_CONSOLE_WINDOW
	DWORD dwWritten;
	WriteConsole(hStdOut, buffer, (DWORD)_tcslen(buffer), &dwWritten, NULL);
#endif
#ifdef USE_TELNET_SERVER
	if (echo_bak) {
		debugger->write_data(buffer, (int)_tcslen(buffer));
	}
#endif
#ifdef USE_STANDARD_INOUT
	_fputts(buffer, stdout);
#endif
	if (cr) Cr();
}

void DebuggerConsole::PutCh(_TCHAR c)
{
	if(logfile != NULL && logfile->IsOpened()) {
		logfile->Fwrite(&c, sizeof(_TCHAR), 1);
	}
#ifdef USE_CONSOLE_WINDOW
	DWORD dwWritten;
	WriteConsole(hStdOut, &c, 1, &dwWritten, NULL);
#endif
#ifdef USE_TELNET_SERVER
	if (echo_bak) debugger->write_data(&c, 1);
#endif
#ifdef USE_STANDARD_INOUT
	_fputtc(c, stdout);
#endif
}

void DebuggerConsole::Cr()
{
	if(logfile != NULL && logfile->IsOpened()) {
		logfile->Fputc('\n');
	}
#ifdef USE_CONSOLE_WINDOW
	DWORD dwWritten;
	WriteConsole(hStdOut, _T("\n"), (DWORD)_tcslen(_T("\n")), &dwWritten, NULL);
#endif
#ifdef USE_TELNET_SERVER
	if (echo_bak) {
		debugger->write_data(_T("\n"), (int)_tcslen(_T("\n")));
	}
#endif
#ifdef USE_STANDARD_INOUT
	_fputts(_T("\n"), stdout);
#endif
}

void DebuggerConsole::Flush()
{
#ifdef USE_CONSOLE_WINDOW
#endif
#ifdef USE_TELNET_SERVER
#endif
#ifdef USE_STANDARD_INOUT
	fflush(stdout);
#endif
}

_TCHAR *DebuggerConsole::GetBuffer(bool clear)
{
	if (clear) {
		memset(buffer, 0, sizeof(buffer));
	}
	return buffer;
}

const _TCHAR *DebuggerConsole::GetParam(int idx)
{
	if (0 <= idx && idx < paramnum) return params[idx];
	else return NULL;
}

// ---------------------------------------------------------------------------

bool DebuggerConsole::SetCommandLine(_TCHAR chr, bool echo)
{
	bool enter_done = false;
	bool prev_echo_bak = echo_bak;
	echo_bak = echo;
	if(chr == 0x0d || chr == 0x0a) {
		if(commandptr == 0 && prev_command[0] != _T('\0')) {
			memcpy(command, prev_command, sizeof(command));
			Printf(_T("%s\n"), command);
			enter_done = true;
		} else if(commandptr > 0) {
			command[commandptr] = _T('\0');
			memcpy(prev_command, command, sizeof(command));
			Cr();
			enter_done = true;
		}
	} else if(chr == 0x08 || chr == 0x7f) {
		// BS or DEL
		if(commandptr > 0) {
			commandptr--;
			PutCh(0x08);
			PutCh(_T(' '));
			PutCh(0x08);
		}
	} else if(chr >= 0x20 && chr <= 0x7e && commandptr < DC_MAX_COMMAND_LEN && !(chr == 0x20 && commandptr == 0)) {
		command[commandptr++] = chr;
		PutCh(chr);
	}
	if (enter_done) {
		commandptr = 0;
	}
	echo_bak = prev_echo_bak;
	return enter_done;
}

bool DebuggerConsole::ReadInput()
{
	bool enter_done = false;

#ifdef USE_CONSOLE_WINDOW
	DWORD dwRead;
	INPUT_RECORD ir[16];
	if(GetNumberOfConsoleInputEvents(hStdIn, &dwRead)) {
		if(dwRead > 0 && ReadConsoleInput(hStdIn, ir, 16, &dwRead)) {
			for(unsigned int i = 0; i < dwRead; i++) {
#ifdef _UNICODE
				if((ir[i].EventType & KEY_EVENT) && ir[i].Event.KeyEvent.bKeyDown && ir[i].Event.KeyEvent.uChar.UnicodeChar) {
					_TCHAR chr = ir[i].Event.KeyEvent.uChar.UnicodeChar;
#else
				if((ir[i].EventType & KEY_EVENT) && ir[i].Event.KeyEvent.bKeyDown && ir[i].Event.KeyEvent.uChar.AsciiChar) {
					_TCHAR chr = ir[i].Event.KeyEvent.uChar.AsciiChar;
#endif
					enter_done = SetCommandLine(chr, true);
					if (enter_done) {
						break;
					}
				}
			}
		}
	} 
#endif
	int readed;
#ifdef USE_TELNET_SERVER
	// input from terminal
	_TCHAR data[64];
	int ignore = 0;
	if ((readed = debugger->read_data(data, 64)) > 0) {
		for(int i = 0; i < readed; i++) {
			// ignore telnet command
			if ((uint8_t)data[i-1] == 0xff && (uint8_t)data[i] >= 0xfb && (uint8_t)data[i] <= 0xfe) {
				ignore = 2;
			} else if ((uint8_t)data[i] >= 0xf0) {
				ignore = 1;
			}
			if (ignore > 0) {
				ignore--;
				continue;
			}
			enter_done = SetCommandLine(data[i], true);
			if (enter_done) {
				break;
			}
		}
	}
#endif
#ifdef USE_STANDARD_INOUT
	// input from stdin
	fd_set sel;
	FD_ZERO(&sel);
	FD_SET(0, &sel);
	struct timeval tim;
	tim.tv_sec = 0;
	tim.tv_usec = 10000;
	int rc = select(1, &sel, NULL, NULL, &tim);
	if (rc > 0) {
		for(int i=0; i < 64 && (readed = _fgettc(stdin)) != EOF; i++) {
			enter_done = SetCommandLine((_TCHAR)readed, true);
			if (enter_done) {
				break;
			}
		}
	}
#endif
	return enter_done;
}

bool DebuggerConsole::EscapePressed()
{
	bool pressed = false;
#ifdef USE_CONSOLE_WINDOW
	if (!pressed) {
//		pressed = ((GetAsyncKeyState(VK_ESCAPE) & 0x8000) != 0 && IsConsoleActive());
		DWORD dwRead;
		INPUT_RECORD ir[16];
		if(GetNumberOfConsoleInputEvents(hStdIn, &dwRead)) {
			if(dwRead > 0 && ReadConsoleInput(hStdIn, ir, 16, &dwRead)) {
				for(unsigned int i = 0; i < dwRead; i++) {
#ifdef _UNICODE
					if((ir[i].EventType & KEY_EVENT) && ir[i].Event.KeyEvent.bKeyDown && ir[i].Event.KeyEvent.uChar.UnicodeChar) {
						_TCHAR chr = ir[i].Event.KeyEvent.uChar.UnicodeChar;
#else
					if((ir[i].EventType & KEY_EVENT) && ir[i].Event.KeyEvent.bKeyDown && ir[i].Event.KeyEvent.uChar.AsciiChar) {
						_TCHAR chr = ir[i].Event.KeyEvent.uChar.AsciiChar;
#endif
						if (chr == 0x1b) {
							pressed = true;
							break;
						}
					}
				}
			}
		} 
	}
#endif
	int readed;
#ifdef USE_TELNET_SERVER
	if (!pressed) {
		_TCHAR data[16];
		if ((readed = debugger->read_data(data, 16)) > 0) {
			for(int i = 0; i < readed; i++) {
				if (data[i] == 0x1b) {
					pressed = true;
					break;
				}
			}
		}
	}
#endif
#ifdef USE_STANDARD_INOUT
	if (!pressed) {
		fd_set sel;
		FD_ZERO(&sel);
		FD_SET(0, &sel);
		struct timeval tim;
		tim.tv_sec = 0;
		tim.tv_usec = 1000;
		int rc = select(1, &sel, NULL, NULL, &tim);
		if (rc > 0) {
			for(int i=0; i < 16 && (readed = _fgettc(stdin)) != EOF; i++) {
				if (readed == 0x1b) {
					pressed = true;
					break;
				}
			}
		}
	}
#endif
	return pressed;
}

int DebuggerConsole::ParseInput()
{
	paramnum = UTILITY::get_parameters(command, DC_MAX_COMMAND_LEN, params, 32);
	return paramnum;
}

bool DebuggerConsole::NowDisable()
{
	return config.now_power_off;
}

// ---------------------------------------------------------------------------

uint32_t DebuggerConsole::HexaToInt(const _TCHAR *str)
{
	_TCHAR tmp[DC_MAX_BUFFER_LEN], *s;
	
	if(str == NULL || _tcslen(str) == 0) {
		return 0;
	}
	memset(tmp, 0, sizeof(tmp));
	UTILITY::tcscpy(tmp, DC_MAX_BUFFER_LEN, str);
	
	if(_tcslen(tmp) == 3 && tmp[0] == _T('\'') && tmp[2] == _T('\'')) {
		// ank
		return tmp[1] & 0xff;
	} else if((s = _tcsstr(tmp, _T(":"))) != NULL) {
		// 0000:0000
		s[0] = _T('\0');
		return (HexaToInt(tmp) << 4) + HexaToInt(s + 1);
	} else if(tmp[0] == _T('%')) {
		// decimal
		return atoi(tmp + 1);
	}
	return (uint32_t)_tcstoul(tmp, NULL, 16);
}

uint8_t DebuggerConsole::HexaToByte(char *value)
{
	char tmp[3];
	tmp[0] = value[0];
	tmp[1] = value[1];
	tmp[2] = '\0';
	return (uint8_t)strtoul(tmp, NULL, 16);
}

uint16_t DebuggerConsole::HexaToWord(char *value)
{
	char tmp[5];
	tmp[0] = value[0];
	tmp[1] = value[1];
	tmp[2] = value[2];
	tmp[3] = value[3];
	tmp[4] = '\0';
	return (uint16_t)strtoul(tmp, NULL, 16);
}

bool DebuggerConsole::HexaToInt(const _TCHAR *str, uint32_t *addr, int *len)
{
	_TCHAR tmp[DC_MAX_BUFFER_LEN], *s;
	uint32_t ad = 0;
	int le = 0;
	bool rc = true;

	memset(tmp, 0, sizeof(tmp));
	UTILITY::tcscpy(tmp, DC_MAX_BUFFER_LEN, str);
	
	if(str == NULL || _tcslen(str) == 0) {
		// zero
	} else if(_tcslen(tmp) == 3 && tmp[0] == _T('\'') && tmp[2] == _T('\'')) {
		// ank
		ad = (tmp[1] & 0xff);
	} else if((s = _tcsstr(tmp, _T("-"))) != NULL) {
		// 0000-0000
		s[0] = _T('\0');
		ad = HexaToInt(tmp);
		le = HexaToInt(s + 1) - ad;
	} else if(tmp[0] == _T('%')) {
		// decimal
		ad = atoi(tmp + 1);
	} else {
		// hexa
		ad = HexaToInt(tmp);
	}
	if (addr) *addr = ad;
	if (len) *len = le;
	return rc;
}

// ---------------------------------------------------------------------------

break_point_t *DebuggerConsole::GetBreakPoint(const _TCHAR *command)
{
	if(toupper(command[0]) == _T('B')) {
		return &debugger->bp[debugger->BP_FETCH_OP];
	} else if(toupper(command[0]) == _T('T')) {
		return &debugger->tp[debugger->TP_FETCH_OP];
	} else if(toupper(command[0]) == _T('R')) {
		if (toupper(command[1]) == _T('T')) return &debugger->tp[debugger->TP_READ_MEMORY];
		else return &debugger->bp[debugger->BP_READ_MEMORY];
	} else if(toupper(command[0]) == _T('W')) {
		if (toupper(command[1]) == _T('T')) return &debugger->tp[debugger->TP_WRITE_MEMORY];
		else return &debugger->bp[debugger->BP_WRITE_MEMORY];
	} else if(toupper(command[0]) == _T('I')) {
		if (toupper(command[1]) == _T('T')) return &debugger->tp[debugger->TP_INPUT_IO];
		else return &debugger->bp[debugger->BP_INPUT_IO];
	} else if(toupper(command[0]) == _T('O')) {
		if (toupper(command[1]) == _T('T')) return &debugger->tp[debugger->TP_OUTPUT_IO];
		else return &debugger->bp[debugger->BP_OUTPUT_IO];
	} else if(toupper(command[0]) == _T('N')) {
		if (toupper(command[1]) == _T('T')) return &debugger->tp[debugger->TP_INTERRUPT];
		else return &debugger->bp[debugger->BP_INTERRUPT];
	}
	return NULL;
}

#if 0
bool DebuggerConsole::IsConsoleActive()
{
#ifdef USE_CONSOLE_WINDOW
	HWND hWnd = GetForegroundWindow();
//	return (hWnd != NULL && hWnd == GetConsoleWindow());
	return (hWnd != NULL && hWnd == FindWindow("ConsoleWindowClass", NULL));
#else
	return false;
#endif
}
#endif

// ---------------------------------------------------------------------------

int DebuggerConsole::GetAddressType(int num)
{
	int type = -1;
	if (num == 1) {
		type = phys_type;
		if (paramnum >= 2) {
			type = HexaToInt(params[1]);
		}
#ifdef _MBS1
		if (type > 5) {
#elif _BML3MK5
		if (type > 5) {
#else
		if (true) {
#endif
			Print(_T("invalid parameter of type"));
			Cr();
			return -2;
		}
	}
	return type;
}

void DebuggerConsole::UsageAddressType(int num)
{
	switch(num) {
	case 1:
		Print(_T("  <type> - specify a kind of memory."));
#ifdef _MBS1
		Print(_T("    0 : using physical address."));
		Print(_T("    1 : attribute RAM"));
		Print(_T("    3 : IG RAM (blue)"));
		Print(_T("    4 : IG RAM (red)"));
		Print(_T("    5 : IG RAM (green)"));
#elif _BML3MK5
		Print(_T("    0 : main RAM"));
		Print(_T("    1 : extended RAM"));
		Print(_T("    2 : color RAM"));
		Print(_T("    3 : IG RAM (blue)"));
		Print(_T("    4 : IG RAM (red)"));
		Print(_T("    5 : IG RAM (green)"));
#endif
		break;
	}
}

// ---------------------------------------------------------------------------

/// dump memory
///
void DebuggerConsole::CommandDumpMemory(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageDumpMemory(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum > (pos+2)) {
		Print(_T("invalid number of parameters"));
		Cr();
		UsageDumpMemory(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageDumpMemory(false, num);
		return;
	}

	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);
	uint32_t start_addr = (num == 1 ? dump_phys_addr : dump_addr);
	if(paramnum >= (1 + pos)) {
		start_addr = HexaToInt(params[pos]);
	}
	start_addr &= addr_mask;

	uint32_t end_addr = start_addr + 8 * 16 - 1;
	if(paramnum >= (2 + pos)) {
		end_addr = HexaToInt(params[pos+1]);
	}
	end_addr &= addr_mask;

	if(start_addr > end_addr) {
		end_addr = addr_mask;
	}
	for(uint64_t addr = (start_addr & ~0x0f); addr <= end_addr; addr++) {
		if(addr > addr_mask) {
			end_addr = addr_mask;
			break;
		}
		if((addr & 0x0f) == 0) {
#ifdef _MBS1
			if (num == 1) {
				Printf(_T(PHYS_ADDR_FORMAT), addr & addr_mask);
			} else {
				Printf(_T(PHYS_ADDR_FORMAT)_T(" "), mem->debug_latch_address(addr & addr_mask));
				Printf(_T(ADDR_FORMAT), addr & addr_mask);
			}
#else
			Printf(_T(ADDR_FORMAT), addr & addr_mask);
#endif
			Print(_T(" "), false);
			memset(buffer, 0, sizeof(buffer));
		}
		if(addr < start_addr) {
			Print(_T("   "), false);
			buffer[addr & 0x0f] = _T(' ');
		} else {
			uint32_t data = mem->debug_read_data8(type, addr & addr_mask);
			Printf(_T(" ")_T(DATA_BYTE_FORMAT), data);
			buffer[addr & 0x0f] = ((data >= 0x20 && data <= 0x7e) || (cp932 && data >= 0xa1 && data <= 0xdf)) ? data : _T('.');
		}
		if((addr & 0x0f) == 0x0f) {
			Print(_T("  "), false);
			Out();
		}
	}
	if((end_addr & 0x0f) != 0x0f) {
		for(uint32_t addr = (end_addr & 0x0f) + 1; addr <= 0x0f; addr++) {
			Print(_T("   "), false);
		}
		Print(_T("  "), false);
		Out();
	}
	if (num == 1) {
		dump_phys_addr = (end_addr + 1) & addr_mask;
		phys_type = type;
		prev_command[2] = _T('\0'); // remove parameters to dump continuously
	} else {
		dump_addr = (end_addr + 1) & addr_mask;
		prev_command[1] = _T('\0'); // remove parameters to dump continuously
	}
}

/// usage - dump memory
///
void DebuggerConsole::UsageDumpMemory(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr1(s, _T("D"), _T("[<start address> [<end address>]]"), _T("Dump memory."));
		break;
	case 1:
		UsageCmdStr1(s, _T("DP"), _T("[<type> [<start address> [<end address>]]]"), _T("Dump physical memory."));
		break;
	}
	if (s) return;

	UsageAddressType(num);
	Print(_T("  <start address> - specify start address."));
	Print(_T("  <end address> - specify end address."));
	Print(_T("  If end address isn't specified, end address is set start address + 128bytes."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandEditMemoryBinary(int n, int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageEditMemory(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum < (pos+2)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageEditMemory(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageEditMemory(false, num);
		return;
	}

	uint32_t addr = HexaToInt(params[pos]);
	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);

	addr &= addr_mask;

	for(int i = pos + 1; i < paramnum; i++) {
		switch(n) {
		case 1:
			mem->debug_write_data8(type, addr, HexaToInt(params[i]) & 0xff);
			addr = (addr + 1) & data_addr_mask;
			break;
		case 2:
			mem->debug_write_data16(type, addr, HexaToInt(params[i]) & 0xffff);
			addr = (addr + 2) & data_addr_mask;
			break;
		case 4:
			mem->debug_write_data32(type, addr, HexaToInt(params[i]));
			addr = (addr + 4) & data_addr_mask;
			break;
		}
	}
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandEditMemoryAscii(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageEditMemory(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum != (pos+2)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageEditMemory(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageEditMemory(false, num);
		return;
	}

	uint32_t addr = HexaToInt(params[pos]);
	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);

	addr &= addr_mask;

	const _TCHAR *token = params[pos+1];
	int len = (int)_tcslen(token);
	for(int i = 0; i < len; i++) {
		mem->debug_write_data8(type, addr, token[i]);
		addr = (addr + 1) & addr_mask;
	}
}

void DebuggerConsole::UsageEditMemory(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr2(s, _T("E[{B,W,D}]"), _T("<address> <value> [<value> ...]"), _T("Edit memory (byte,word,dword).")
			,_T("EA"), _T("<address> \"<string>\""), _T("Edit memory (ascii)."));
		break;
	case 1:
		UsageCmdStr2(s, _T("EP[{B,W,D}]"), _T("<type> <address> <value> [<value> ...]"), _T("Edit physical memory (byte,word,dword).")
			,_T("EPA"), _T("<type> <address> \"<string>\""), _T("Edit physical memory (ascii)."));
		break;
	}
	if (s) return;

	UsageAddressType(num);
	Print(_T("  <address> - specify address."));
	Print(_T("  <value> - specify value."));
	Print(_T("  <string> - specify ascii string."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandInputIOBinary(int n)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageEditIO(false);
		return;
	}
	if(paramnum != 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageEditIO(false);
		return;
	}

	switch(n) {
	case 1:
		Printf(_T("%02X"), cpu->debug_read_io8(HexaToInt(params[1])) & 0xff);
		Cr();
		break;
	case 2:
		Printf(_T("%02X"), cpu->debug_read_io16(HexaToInt(params[1])) & 0xffff);
		Cr();
		break;
	case 4:
		Printf(_T("%02X"), cpu->debug_read_io32(HexaToInt(params[1])));
		Cr();
		break;
	}
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandOutputIOBinary(int n)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageEditIO(false);
		return;
	}
	if(paramnum != 3) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageEditIO(false);
		return;
	}

	switch(n) {
	case 1:
		cpu->debug_write_io8(HexaToInt(params[1]), HexaToInt(params[2]) & 0xff);
		break;
	case 2:
		cpu->debug_write_io16(HexaToInt(params[1]), HexaToInt(params[2]) & 0xffff);
		break;
	case 4:
		cpu->debug_write_io32(HexaToInt(params[1]), HexaToInt(params[2]));
		break;
	}
}

void DebuggerConsole::UsageEditIO(bool s)
{
	UsageCmdStr2(s, _T("I[{B,W,D}]"), _T("<port>"), _T("Input port (byte,word,dword).")
		, _T("O[{B,W,D}]"), _T("<port> <value>"), _T("Output port (byte,word,dword)."));
	if (s) return;

	Print(_T("  <port> - specify i/o port address."));
	Print(_T("  <value> - specify value."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandAccessRegister()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageAccessRegister(false);
		return;
	}

	if(paramnum == 1) {
		// show
		cpu->debug_regs_info(buffer, DC_MAX_BUFFER_LEN);
		Out();
	} else if(paramnum == 3) {
		// edit
		if(!cpu->debug_write_reg(params[1], HexaToInt(params[2]))) {
			Print(_T("unknown register "), false);
			Print(params[1]);
		}
	} else {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageAccessRegister(false);
	}
}

void DebuggerConsole::UsageAccessRegister(bool s)
{
	UsageCmdStr1(s, _T("R"), _T("[<reg> <value>]"), _T("Show/Edit register(s) in cpu."));
	if (s) return;

	Print(_T("  <reg> - specify a register."));
	Print(_T("  <value> - specify a value."));
	Print(_T("  If reg and value is specified, modify specified register."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandAccessDevice()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageAccessDevice(false);
		return;
	}

	if(paramnum < 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageAccessDevice(false);
		return;
	}

	uint32_t num = 0;
	const _TCHAR *dname = NULL;
	if (!dp->vm->get_debug_device_name(params[1], &num, NULL, &dname)) {
		Print(_T("invalid device name"));
		Cr();
		UsageAccessDevice(false);
		return;
	}
	
	if(paramnum == 2) {
		// show
		dp->vm->debug_regs_info(num, buffer, DC_MAX_BUFFER_LEN);
		Out();
	} else if(paramnum == 4) {
		// edit
		if(!dp->vm->debug_write_reg(num, HexaToInt(params[2]), HexaToInt(params[3]))) {
			Print(_T("invalid register number"));
		}
	} else {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageAccessDevice(false);
	}
}

void DebuggerConsole::UsageAccessDevice(bool s)
{
	UsageCmdStr1(s, _T("RD"), _T("<device name> [<regno> <value>]"), _T("Show/Edit register(s) in specified device."));
	if (s) return;

	Print(_T("  <device name> - specify following device name."));
	dp->vm->get_debug_device_names_str(buffer, DC_MAX_BUFFER_LEN);
	Printf(_T("  %s"), buffer);
	Cr();
	Print(_T("  <regno> - specify a register number."));
	Print(_T("  <value> - specify a value."));
	Print(_T("  If regno and value is specified, modify specified register."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandSearch(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageSearch(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum < (pos+3)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageSearch(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageSearch(false, num);
		return;
	}

	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);
	uint32_t start_addr = HexaToInt(params[pos]) & addr_mask;
	uint32_t end_addr = HexaToInt(params[pos+1]) & addr_mask;
	uint8_t list[32];
	for(int i = pos+2, j = 0; i < paramnum; i++, j++) {
		list[j] = HexaToInt(params[i]);
	}
	for(uint64_t addr = start_addr; addr <= end_addr; addr++) {
		bool found = true;
		for(int i = pos+2, j = 0; i < paramnum; i++, j++) {
			if(mem->debug_read_data8(type, (addr + j) & addr_mask) != list[j]) {
				found = false;
				break;
			}
		}
		if(found) {
			Printf(_T(ADDR_FORMAT), addr);
			for(int i = pos+2, j = 0; i < paramnum; i++, j++) {
				Printf(_T(" ")_T(DATA_FORMAT), list[j]);
			}
			Cr();
		}
	}
}

void DebuggerConsole::UsageSearch(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr1(s, _T("S"), _T("<start address> <end address> <list>"), _T("Search data(s)."));
		break;
	case 1:
		UsageCmdStr1(s, _T("SP"), _T("<type> <start address> <end address> <list>"), _T("Search data(s) using physical address."));
		break;
	}
	if (s) return;

	UsageAddressType(num);
	Print(_T("  <start address> - specify start address."));
	Print(_T("  <end address> - specify end address."));
	Print(_T("  <list> - specify values separate by space to match data in memory."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandUnassemble(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageUnassemble(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum > (pos+2)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageUnassemble(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageUnassemble(false, num);
		return;
	}

	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : prog_addr_mask);

	uint32_t addr = (num == 1 ? dasm_phys_addr : dasm_addr);
	if(paramnum >= (pos+1)) {
		addr = HexaToInt(params[pos]) & addr_mask;
	}
	uint32_t end_addr = addr;
	if(paramnum >= (pos+2)) {
		end_addr = HexaToInt(params[pos+1]) & addr_mask;
	}
	for(uint32_t n = 0; (paramnum == (pos+2) ? addr < end_addr : n < 16); n++) {
		int len = cpu->debug_dasm(type, addr, buffer, DC_MAX_BUFFER_LEN, 0x10);
#ifndef USE_EMU_INHERENT_SPEC
		Printf(_T("%08X  "), addr);
		for(int i = 0; i < len; i++) {
			Printf(_T("%02X"), mem->debug_read_data8(type, (addr + i) & addr_mask));
		}
		for(int i = len; i < 8; i++) {
			Print(_T("  "), false);
		}
		Print(_T("  "), false);
#endif
		Out();
		addr = (addr + len) & addr_mask;
	}
	if (num == 1) {
		dasm_phys_addr = addr;
		phys_type = type;
		prev_command[2] = _T('\0'); // remove parameters to dump continuously
	} else {
		dasm_addr = addr;
		prev_command[1] = _T('\0'); // remove parameters to dump continuously
	}
}

void DebuggerConsole::UsageUnassemble(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr1(s, _T("U"), _T("[<start address> [<end address>]]"), _T("Unassemble data in memory."));
		break;
	case 1:
		UsageCmdStr1(s, _T("UP"), _T("<type> [<start address> [<end address>]]"), _T("Unassemble data using physical address."));
		break;
	}
	if (s) return;

	UsageAddressType(num);
	Print(_T("  <start address> - specify start address."));
	Print(_T("  <end address> - specify end address."));
	Print(_T("  If end address isn't specified, unassemble 16 steps."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandCalculateHexa()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageCalculateHexa(false);
		return;
	}

	if(paramnum != 3) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageCalculateHexa(false);
		return;
	}

	uint32_t l = HexaToInt(params[1]);
	uint32_t r = HexaToInt(params[2]);
	Printf(_T("Add:%08X  Sub:%08X"), l + r, l - r);
	Cr();
}

void DebuggerConsole::UsageCalculateHexa(bool s)
{
	UsageCmdStr1(s, _T("H"), _T("<value1> <value2>"), _T("Add and subtract hexa value1 and value2."));
	if (s) return;
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandSetFileName()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageSetFileName(false);
		return;
	}

	if(paramnum >= 2) {
		UTILITY::tcscpy(debugger->file_path, _MAX_PATH, params[1]);
	}
	Printf(_T("Current dump file name is \"%s\"."), debugger->file_path);
	Cr();
}

void DebuggerConsole::UsageSetFileName(bool s)
{
	UsageCmdStr1(s, _T("DN"), _T("[<filename>]"), _T("Show/Set filename for dump file."));
	if (s) return;

	Print(_T("  <filename> - specify file name."));
	Print(_T("  Filename is \"debug.bin\" in default."));
	Print(_T("  File format is decided by file extension:"));
	Print(_T("    \".hex\" - intel hex format."));
	Print(_T("  Otherwise binary format."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandLoadBinaryFile(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageBinaryFile(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum < (pos+1)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageBinaryFile(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageBinaryFile(false, num);
		return;
	}

	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);

	FILEIO* fio = new FILEIO();
	if(UTILITY::check_file_extension(debugger->file_path, _T(".hex"))) {
		// load from intel hex format file
		if(fio->Fopen(debugger->file_path, FILEIO_READ_ASCII)) {
			uint32_t start_addr = 0, linear = 0, segment = 0;
			if(paramnum >= (pos + 1)) {
				start_addr = HexaToInt(params[pos]);
			}
			char line[DC_MAX_BUFFER_LEN];
			while(fio->Fgets(line, sizeof(line)) != NULL) {
				if(line[0] != ':') continue;
				int type = HexaToByte(line + 7);
				if(type == 0x00) {
					uint32_t bytes = HexaToByte(line + 1);
					uint32_t addr = HexaToWord(line + 3) + start_addr + linear + segment;
					for(uint32_t i = 0; i < bytes; i++) {
						mem->debug_write_data8(type, (addr + i) & addr_mask, HexaToByte(line + 9 + 2 * i));
					}
				} else if(type == 0x01) {
					break;
				} else if(type == 0x02) {
					segment = HexaToWord(line + 9) << 4;
//					start_addr = 0;
				} else if(type == 0x04) {
					linear = HexaToWord(line + 9) << 16;
//					start_addr = 0;
				}
			}
			fio->Fclose();
			Print(_T("loaded from "), false);
		} else {
			Print(_T("can't open "), false);
		}
		Print(debugger->file_path);
	} else {
		// load from binary dump file
		if(fio->Fopen(debugger->file_path, FILEIO_READ_BINARY)) {
			uint32_t start_addr = 0, end_addr = addr_mask;
			if(paramnum >= (pos + 1)) {
				start_addr = HexaToInt(params[pos]) & addr_mask;
			}
			if(paramnum >= (pos + 2)) {
				end_addr = HexaToInt(params[pos+1]) & addr_mask;
			}
			for(uint32_t addr = start_addr; addr <= end_addr; addr++) {
				int data = fio->Fgetc();
				if(data == EOF) {
					break;
				}
				mem->debug_write_data8(type, addr & addr_mask, data);
			}
			fio->Fclose();
			Print(_T("loaded from "), false);
		} else {
			Print(_T("can't open "), false);
		}
		Print(debugger->file_path);
	}
	delete fio;
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandSaveBinaryFile(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageBinaryFile(false, num);
		return;
	}

	int pos = (num == 1 ? 2 : 1);

	if(paramnum != (pos+2)) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageBinaryFile(false, num);
		return;
	}

	int type = GetAddressType(num);
	if (type < -1) {
		UsageBinaryFile(false, num);
		return;
	}

	uint32_t addr_mask = (num == 1 ? mem->debug_physical_addr_mask(type) : data_addr_mask);

	uint32_t start_addr = HexaToInt(params[pos]) & addr_mask, end_addr = HexaToInt(params[pos+1]) & addr_mask;
	FILEIO* fio = new FILEIO();
	if(UTILITY::check_file_extension(debugger->file_path, _T(".hex"))) {
		// save to intel hex format file
		if(fio->Fopen(debugger->file_path, FILEIO_WRITE_ASCII)) {
			uint32_t addr = start_addr;
			uint32_t segment = 0;
			while(addr <= end_addr) {
				uint32_t len = 0;
				uint32_t sum = 0;
				if (segment != ((start_addr & 0xf0000) >> 4)) {
					// type 02 segment
					len = 2;
					segment = ((start_addr & 0xf0000) >> 4);
					sum = len + 0x02 + ((segment >> 8) & 0xff) + (segment & 0xff);
					fio->Fprintf(":%02X000002%04X%02X\n", len, segment & 0xffff, (0x100 - (sum & 0xff)) & 0xff);
				}
				// type 00 data
				len = SDL_min(end_addr - addr + 1, 16);
				sum = len + ((addr >> 8) & 0xff) + (addr & 0xff) + 0x00;
				fio->Fprintf(":%02X%04X%02X", len, addr & 0xffff, 0x00);
				for(uint32_t i = 0; i < len; i++) {
					uint8_t data = mem->debug_read_data8(type, (addr++) & addr_mask);
					sum += data;
					fio->Fprintf("%02X", data);
				}
				fio->Fprintf("%02X\n", (0x100 - (sum & 0xff)) & 0xff);
			}
			// tyoe 01 eof
			fio->Fprintf(":00000001FF\n");
			fio->Fclose();
			Print(_T("saved to "), false);
		} else {
			Print(_T("can't open "), false);
		}
		Print(debugger->file_path);
	} else {
		// save to binary dump file
		if(fio->Fopen(debugger->file_path, FILEIO_WRITE_BINARY)) {
			for(uint32_t addr = start_addr; addr <= end_addr; addr++) {
				fio->Fputc(mem->debug_read_data8(type, addr & addr_mask));
			}
			fio->Fclose();
			Print(_T("saved to "), false);
		} else {
			Print(_T("can't open "), false);
		}
		Print(debugger->file_path);
	}
	delete fio;
}

void DebuggerConsole::UsageBinaryFile(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr3(s
			, _T("DL"), _T("<start address> [<end address>]"), _T("Load from a dump file.")
			, _T("DL"), _T("<offset address>"), _T("Load from a intel hex format file.")
			, _T("DS"), _T("<start address> <end address>"), _T("Save memory data to a dump file."));
		break;
	case 1:
		UsageCmdStr3(s
			, _T("DPL"), _T("<type> <start address> [<end address>]"), _T("Load from a dump file to memory of specified type.")
			, _T("DPL"), _T("<type> <offset address>"), _T("Load from a intel hex format file.")
			, _T("DPS"), _T("<type> <start address> <end address>"), _T("Save memory of specified type to a dump file."));
		break;
	}
	if (s) return;

	UsageAddressType(num);
	Print(_T("  <start address> - specify start address."));
	Print(_T("  <end address> - specify end address."));
	Print(_T("  <offset address> - specify offset from start address (when intel hex format)."));
	Print(_T("  File format is decided by file extension:"));
	Print(_T("    \".hex\" - intel hex format."));
	Print(_T("  Otherwise binary format."));
	Print(_T("  When load from intel hex format file, start address is set by data in file."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandSaveImageFile(int num)
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageImageFile(false, num);
		return;
	}

	if(paramnum != 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageImageFile(false, num);
		return;
	}

	int type =  HexaToInt(params[1]);

#ifdef _BML3MK5
	if (type >= 4) {
#endif
#ifdef _MBS1
	if (type >= 8) {
#endif
		Print(_T("invalid type"));
		Cr();
		UsageImageFile(false, num);
		return;
	}

	int w = 0, h = 0;
	if (!mem->get_debug_graphic_memory_size(type, &w, &h) || w <= 0 || h <= 0) {
		Print(_T("fatal error. (invalid size)"));
		Cr();
		return;
	}

	SDL_PixelFormat *fmt = dp->emu->get_pixel_format();
	CSurface imgSuf(w, h, fmt);

	if (!imgSuf.IsEnable()) {
		Print(_T("fatal error. (cannot create surface)"));
		Cr();
		return;
	}

	if (!mem->debug_draw_graphic(type, w, h, (scrntype *)imgSuf.GetBuffer())) {
		Print(_T("fatal error. (cannot create image)"));
		Cr();
		return;
	}

	if (dp->emu->debugger_save_image(w, h, &imgSuf)) {
		Print(_T("image file was successfully saved"));
	} else {
		Print(_T("fatal error. (cannot save image)"));
	}

	Cr();
}

void DebuggerConsole::UsageImageFile(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr1(s
			, _T("DIS"), _T("<type>"), _T("Save image file in memory."));
		break;
	}
	if (s) return;

	Print(_T("  <type> - specify type of memory."));
	Print(_T("    0 : IG RAM"));
	Print(_T("    1 : IG RAM (blue)"));
	Print(_T("    2 : IG RAM (red)"));
	Print(_T("    3 : IG RAM (green)"));
#ifdef _MBS1
	Print(_T("    4 : GRAPHIC RAM"));
	Print(_T("    5 : GRAPHIC RAM (blue)"));
	Print(_T("    6 : GRAPHIC RAM (red)"));
	Print(_T("    7 : GRAPHIC RAM (green)"));
#endif
	Print(_T("  File name will be named as YYYY-MM-DD-HH-MI-SS.bmp."));
	Print(_T("  If snapshot path is set, image file is put on it."));
}

// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------

/// @param[in] num : 0:addr 1:ioport 2:interrupt / 8:trace addr
void DebuggerConsole::CommandSetBreakPoint(int num)
{
	int cat = (num & 8);
	num &= 7;

	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageSetBreakPoint(false, cat | num);
		return;
	}

	if(paramnum < 2) {
		CommandListBreakPoint(cat | num);
		return;
	}

	const _TCHAR *bptrstr = (cat == 8 ? _T("trace") : _T("break"));

	break_point_t *bp = GetBreakPoint(params[0]);

	uint32_t addr = 0;
	uint32_t mask = 0;
	int len = 0;
	int idx = 0;
	const _TCHAR *iname = NULL;

	switch(num) {
	case 2:
		// interrupt
		if (!cpu->get_debug_signal_name_index(params[1], &addr, &idx, &iname)) {
			Print(_T("invalid signal name"));
			Cr();
			UsageSetBreakPoint(false, cat | num);
			return;
		}
		break;
	case 1:
		// ioport
		mask = 0xff;
		if(paramnum == 3) {
			mask = HexaToInt(params[2]);
		}
		HexaToInt(params[1], &addr, &len);
		break;
	default:
		// address
#ifdef _MBS1
		if(paramnum == 3) {
			mask = HexaToInt(params[2]);
		}
		if (mask > 1) mask = 1;
#else
		mask = 0;
#endif

		if (params[1][0] == _T('.')) {
			// set current address
			addr = cpu->get_debug_pc(mask);
		} else if (_tcsicmp(params[1], _T("n")) == 0) {
			// set next address
			addr = cpu->get_debug_next_pc(mask);
		} else {
			HexaToInt(params[1], &addr, &len);
		}

		if (mask == 0) addr &= prog_addr_mask;
		break;
	}

	bool found = false;
	int i = 0;
	for(; i < MAX_BREAK_POINTS && !found; i++) {
		if(bp->table[i].status == 0 || (bp->table[i].addr == addr && bp->table[i].mask == mask && bp->table[i].len == len)) {
			bp->table[i].addr = addr;
			bp->table[i].mask = mask;
			bp->table[i].len  = len;
			bp->table[i].status = 1;
			found = true;
			break;
		}
	}
	if(found) {
		Printf(_T("set %spoint%d "), bptrstr, i + 1);
		PrintBreakPointInfo(num, bp, i, iname);
		Cr();
	} else {
		Printf(_T("too many %spoints"), bptrstr);
		Cr();
	}
}

void DebuggerConsole::UsageSetBreakPoint(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr2(s, _T("BP"), _T("[<address(es)> [<type>]]"), _T("Set/Show breakpoint(s).")
			, _T("{R,W}BP"), _T("[<address(es)> [<type>]]"), _T("Set/Show breakpoint(s) (break at memory access)."));
		break;
	case 1:
		UsageCmdStr1(s, _T("{I,O}BP"), _T("[<port> [<mask>]]"), _T("Set/Show breakpoint(s) (break at i/o access)."));
		break;
	case 2:
		UsageCmdStr1(s, _T("NBP"), _T("[<signal name>]"), _T("Set/Show breakpoint(s) (break when interrupt occured)."));
		break;
	case 8:
		UsageCmdStr2(s, _T("TP"), _T("[<address(es)> [<type>]]"), _T("Set/Show tracepoint(s).")
			, _T("{R,W}TP"), _T("[<address(es)> [<type>]]"), _T("Set/Show tracepoint(s) (memory access)."));
		break;
	case 9:
		UsageCmdStr1(s, _T("{I,O}TP"), _T("[<port> [<mask>]]"), _T("Set/Show tracepoint(s) (i/o access)."));
		break;
	case 10:
		UsageCmdStr1(s, _T("NTP"), _T("[<signal name>]"), _T("Set/Show tracepoint(s) (interrupt occured)."));
		break;
	}
	if (s) return;

	switch(num) {
	case 0:
		Print(_T("  BP : break it when cpu fetches a opcode at specified address."));
		Print(_T("  RBP: break it when cpu read a data at specified address."));
		Print(_T("  WBP: break it when cpu writes a data at specified address."));
		Print(_T("  <address(es)> - set one <address> or set range <start address>-<end address>."));
		Print(_T("  * set current address if specify \".\"(period)."));
		Print(_T("  * set next address if specify \"n\" or \"N\"."));
#ifdef _MBS1
		Print(_T("  <type> - set 1 if specify physical address."));
#else
		Print(_T("  <type> - reserved."));
#endif
		break;
	case 1:
		Print(_T("  IBP: break it when cpu read a data at specified port."));
		Print(_T("  OBP: break it when cpu write a data at specified port."));
		break;
	case 2:
		Print(_T("  NBP: break it when cpu received a interrupt signal."));
		cpu->get_debug_signal_names_str(buffer, DC_MAX_BUFFER_LEN);
		Printf(_T("  <signal name> - %s"), buffer);
		Cr();
		break;
	case 8:
		Print(_T("  TP : start trace when cpu fetches a opcode at specified address."));
		Print(_T("  RTP: start trace when cpu read a data at specified address."));
		Print(_T("  WTP: start trace when cpu writes a data at specified address."));
		Print(_T("  <address(es)> - set one <address> or set range <start address>-<end address>."));
		Print(_T("  * set current address if specify \".\"(period)."));
		Print(_T("  * set next address if specify \"n\" or \"N\"."));
#ifdef _MBS1
		Print(_T("  <type> - set 1 if specify physical address."));
#else
		Print(_T("  <type> - reserved."));
#endif
		break;
	case 9:
		Print(_T("  ITP: start trace when cpu read a data at specified port."));
		Print(_T("  OTP: start trace when cpu write a data at specified port."));
		break;
	case 10:
		Print(_T("  NTP: start trace when cpu received a interrupt signal."));
		cpu->get_debug_signal_names_str(buffer, DC_MAX_BUFFER_LEN);
		Printf(_T("  <signal name> - %s"), buffer);
		Cr();
		break;
	}
}

void DebuggerConsole::PrintBreakPointInfo(int num, break_point_t *bp, int i, const _TCHAR *iname)
{
	if (num == 2) {
		Printf(_T(" %s"), iname);
	} else if (num == 1) {
		Printf(_T("at port ")_T(ADDR_FORMAT)_T(" mask ")_T(DATA_FORMAT)
			, bp->table[i].addr, bp->table[i].mask);
	} else {
		_stprintf(buffer, _T("%s %saddress %s%s%s")
			, bp->table[i].len > 0 ? _T("between") : _T("at")
	#ifdef _MBS1
			, bp->table[i].mask ? _T("physical ") : _T("logical ")
			, bp->table[i].mask ? _T(PHYS_ADDR_FORMAT) : _T(ADDR_FORMAT)
	#else
			, _T("")
			, _T(ADDR_FORMAT)
	#endif
			, bp->table[i].len > 0 ? _T(" and ") : _T("")
	#ifdef _MBS1
			, bp->table[i].len > 0 ? (bp->table[i].mask ? _T(PHYS_ADDR_FORMAT) : _T(ADDR_FORMAT)) : _T("")
	#else
			, bp->table[i].len > 0 ? _T(ADDR_FORMAT) : _T("")
	#endif
		);
		if (bp->table[i].len > 0) {
			Printf(buffer, bp->table[i].addr, bp->table[i].addr + bp->table[i].len);
		} else {
			Printf(buffer, bp->table[i].addr);
		}
	}
}

// ---------------------------------------------------------------------------

/// @param[in] num : 0:addr 1:ioport 2:interrupt / 8:trace point
void DebuggerConsole::CommandClearBreakPoint(int num)
{
	int cat = (num & 8);
	num &= 7;

	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageChangeBreakPoint(false, cat | num);
		return;
	}

	break_point_t *bp = GetBreakPoint(params[0]);

	const _TCHAR *bptrstr = (cat == 8 ? _T("trace") : _T("break"));

	if(paramnum == 2 && (_tcsicmp(params[1], _T("*")) == 0 || _tcsicmp(params[1], _T("ALL")) == 0)) {
		memset(bp->table, 0, sizeof(bp->table));
		Printf(_T("cleared all %sepoint(s)"), bptrstr);
		Cr();
	} else if(paramnum >= 2) {
		for(int i = 1; i < paramnum; i++) {
			int index = HexaToInt(params[i]);
			if(!(index >= 1 && index <= MAX_BREAK_POINTS)) {
				Printf(_T("invalid index %d"), index);
				Cr();
			} else {
				if (bp->table[index - 1].status) {
					Printf(_T("cleared %spoint%d"), bptrstr, index);
					Cr();
				}
				bp->table[index - 1].addr = bp->table[index - 1].mask = 0;
				bp->table[index - 1].status = 0;
			}
		}
	} else {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageChangeBreakPoint(false, cat | num);
	}
}

// ---------------------------------------------------------------------------

/// @param[in] num : 0:addr 1:ioport 2:interrupt / 8:trace point
void DebuggerConsole::CommandChangeBreakPoint(int num)
{
	int cat = (num & 8);
	num &= 7;

	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageChangeBreakPoint(false, cat | num);
		return;
	}

	break_point_t *bp = GetBreakPoint(params[0]);

	const _TCHAR *bptrstr = (cat == 8 ? _T("trace") : _T("break"));

	bool enabled = (params[0][1] == _T('E') || params[0][1] == _T('e') || params[0][2] == _T('E') || params[0][2] == _T('e'));
	const _TCHAR *enastr = (enabled ? _T("enabled") : _T("disabled"));

	if(paramnum == 2 && (_tcsicmp(params[1], _T("*")) == 0 || _tcsicmp(params[1], _T("ALL")) == 0)) {
		for(int i = 0; i < MAX_BREAK_POINTS; i++) {
			if(bp->table[i].status != 0) {
				bp->table[i].status = enabled ? 1 : -1;
			}
		}
		Printf(_T("%s all %spoint(s)"), enastr, bptrstr);
		Cr();
	} else if(paramnum >= 2) {
		for(int i = 1; i < paramnum; i++) {
			int index = HexaToInt(params[i]);
			if(!(index >= 1 && index <= MAX_BREAK_POINTS)) {
				Printf(_T("invalid index %d"), index);
				Cr();
			} else if(bp->table[index - 1].status == 0) {
				Printf(_T("%spoint%d is null"), bptrstr, index);
				Cr();
			} else {
				bp->table[index - 1].status = enabled ? 1 : -1;
				Printf(_T("%s %spoint%d"), enastr, bptrstr, index);
				Cr();
			}
		}
	} else {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageChangeBreakPoint(false, cat | num);
	}
}

void DebuggerConsole::UsageChangeBreakPoint(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr2(s, _T("B{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable breakpoint(s).")
			, _T("{R,W}B{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable breakpoint(s) (memory access)."));
		break;
	case 1:
		UsageCmdStr1(s, _T("{I,O}B{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable breakpoint(s) (i/o access)."));
		break;
	case 2:
		UsageCmdStr1(s, _T("NB{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable breakpoint(s) (interrupt)."));
		break;
	case 8:
		UsageCmdStr2(s, _T("T{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable tracepoint(s).")
			, _T("{R,W}T{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable tracepoint(s) (memory access)."));
		break;
	case 9:
		UsageCmdStr1(s, _T("{I,O}T{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable tracepoint(s) (i/o access)."));
		break;
	case 10:
		UsageCmdStr1(s, _T("NT{C,D,E}"), _T("{*,ALL,<list>}"), _T("Clear/Disable/Enable tracepoint(s) (interrupt)."));
		break;
	}
	if (s) return;

	switch(num) {
	case 0:
	case 1:
	case 2:
		Print(_T("  <list> - specify number."));
		Print(_T("  * or ALL - perform it about all breakpoints."));
		break;
	case 8:
	case 9:
	case 10:
		Print(_T("  <list> - specify number."));
		Print(_T("  * or ALL - perform it about all tracepoints."));
		break;
	}
}

// ---------------------------------------------------------------------------

/// @param[in] num : 0:addr 1:ioport 2:interrupt / 8:trace point
void DebuggerConsole::CommandListBreakPoint(int num)
{
	int cat = (num & 8);
	num &= 7;

	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageListBreakPoint(false, cat | num);
		return;
	}

	if(paramnum != 1) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageListBreakPoint(false, cat | num);
		return;
	}

	const _TCHAR *bptrstr = (cat == 8 ? _T("trace") : _T("break"));

	break_point_t *bp = GetBreakPoint(params[0]);
	bool found = false;
	for(int i = 0; i < MAX_BREAK_POINTS; i++) {
		if(bp->table[i].status) {
			found = true;
			const _TCHAR *iname = NULL;
			if (num == 2) {
				cpu->get_debug_signal_name_index(bp->table[i].addr, NULL, &iname);
			}
			Printf(_T("%d: %s"), i + 1, bp->table[i].status == 1 ? _T("enable ") : _T("disable "));
			PrintBreakPointInfo(num, bp, i, iname);
			Cr();

		}
	}
	if (!found) {
		Printf(_T("No %spoints exist."), bptrstr);
		Cr();
	}
}

void DebuggerConsole::UsageListBreakPoint(bool s, int num)
{
	switch(num) {
	case 0:
		UsageCmdStr2(s, _T("BL"), NULL, _T("List breakpoint(s).")
			, _T("{R,W}BL"), NULL, _T("List breakpoint(s) (memory access)."));
		break;
	case 1:
		UsageCmdStr1(s, _T("{I,O}BL"), NULL, _T("List breakpoint(s) (i/o access)."));
		break;
	case 2:
		UsageCmdStr1(s, _T("NBL"), NULL, _T("List breakpoint(s) (interrupt)."));
		break;
	case 8:
		UsageCmdStr2(s, _T("TL"), NULL, _T("List tracepoint(s).")
			, _T("{R,W}TL"), NULL, _T("List tracepoint(s) (memory access)."));
		break;
	case 9:
		UsageCmdStr1(s, _T("{I,O}TL"), NULL, _T("List tracepoint(s) (i/o access)."));
		break;
	case 10:
		UsageCmdStr1(s, _T("NTL"), NULL, _T("List tracepoint(s) (interrupt)."));
		break;
	}
	if (s) return;
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandExecute()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageExecute(false);
		return;
	}

	if(paramnum > 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageExecute(false);
		return;
	}

	if (NowDisable()) {
		Print(Yellow, _T("cannot execute now."));
		SetTextColor(White);
		return;
	}

	if(paramnum >= 2) {
		debugger->store_break_points();
		break_point_table_t *bpt = &debugger->bp[debugger->BP_FETCH_OP].table[0];
		bpt->addr = HexaToInt(params[1]) & prog_addr_mask;
		bpt->mask = prog_addr_mask;
		bpt->status = 1;
	}

	SetKeyStatus();

	uint32_t breaked = 0;
	while (!dp->request_terminate && !breaked && !debugger->now_breakpoint) {

		debugger->now_going = -1;
		debugger->now_suspended = false;
		while(!dp->request_terminate && !debugger->now_suspended) {
			if (EscapePressed()) {
				breaked |= 1;
				break;
			} else if (NowDisable()) {
				breaked |= 2;
				break;
			}
			CDelay(10);
		}
		// break cpu
		debugger->now_going = 0;
		while(!dp->request_terminate && !debugger->now_suspended && !NowDisable()) {
			CDelay(10);
		}
		dasm_addr = cpu->get_next_pc();
					
		TraceCurrent();

		cpu->debug_regs_info(buffer, DC_MAX_BUFFER_LEN);
		Out(White);

		TraceNext();

		debugger->set_trace_point(false);
	}

	PrintBreakedReason(breaked);

	SetTextColor(White);

	if(paramnum >= 2) {
		debugger->restore_break_points();
	}
}

void DebuggerConsole::UsageExecute(bool s)
{
	UsageCmdStr1(s, _T("G"), _T("[<address>]"), _T("Continue processing."));
	if (s) return;

	Print(_T("  <address> - specify address to break it."));
	Print(_T("  If address is specified, ignore any other breakpoints."));
	Print(_T("  Break it if esc key was pressed."));
}

void DebuggerConsole::PrintBreakedReason(uint32_t reason)
{
	int hit = debugger->hit_break_point();
	if(hit >= 0) {
		SetTextColor(Red);
		switch(hit) {
		case DEBUGGER::BP_FETCH_OP:
			_stprintf(buffer, _T("breaked at %s"), _T(ADDR_FORMAT));
			Printf(buffer, cpu->get_next_pc());
			break;
		case DEBUGGER::BP_READ_MEMORY:
			_stprintf(buffer, _T("breaked at %s : memory %s was read at %s"), _T(ADDR_FORMAT), _T(ADDR_FORMAT), _T(ADDR_FORMAT));
			Printf(buffer, cpu->get_next_pc(), debugger->bp[hit].hit_addr, cpu->get_pc());
			break;
		case DEBUGGER::BP_WRITE_MEMORY:
			_stprintf(buffer, _T("breaked at %s : memory %s was written at %s"), _T(ADDR_FORMAT), _T(ADDR_FORMAT), _T(ADDR_FORMAT));
			Printf(buffer, cpu->get_next_pc(), debugger->bp[hit].hit_addr, cpu->get_pc());
			break;
		case DEBUGGER::BP_INPUT_IO:
			_stprintf(buffer, _T("breaked at %s : port %s was read at %s"), _T(ADDR_FORMAT), _T(ADDR_FORMAT), _T(ADDR_FORMAT));
			Printf(buffer, cpu->get_next_pc(), debugger->bp[hit].hit_addr, cpu->get_pc());
			break;
		case DEBUGGER::BP_OUTPUT_IO:
			_stprintf(buffer, _T("breaked at %s : port %s was written at %s"), _T(ADDR_FORMAT), _T(ADDR_FORMAT), _T(ADDR_FORMAT));
			Printf(buffer, cpu->get_next_pc(), debugger->bp[hit].hit_addr, cpu->get_pc());
			break;
		case DEBUGGER::BP_INTERRUPT: {
			const _TCHAR *iname = NULL;
			cpu->get_debug_signal_name_index(debugger->bp[hit].hit_addr, NULL, &iname);
			_stprintf(buffer, _T("breaked at %s : %s occured"), _T(ADDR_FORMAT), iname);
			Printf(buffer, cpu->get_next_pc());
			}
			break;
		}
		Cr();
		debugger->set_break_point(false);
	} else if (reason & 1) {
		_stprintf(buffer, _T("breaked at %s: esc key was pressed"), _T(ADDR_FORMAT));
		Printf(Red, buffer, cpu->get_next_pc());
		Cr();
	} else if (reason & 2) {
		_stprintf(buffer, _T("breaked at %s: cannot execute"), _T(ADDR_FORMAT));
		Printf(Red, buffer, cpu->get_next_pc());
		Cr();
	}
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandTrace()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageTrace(false);
		return;
	}

	if(paramnum > 3) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageTrace(false);
		return;
	}

	if (NowDisable()) {
		Print(Yellow, _T("cannot trace now."));
		SetTextColor(White);
		return;
	}

	int skips = 0;
	int steps = 1;
	if(paramnum >= 3) {
		skips = HexaToInt(params[2]);
	}
	if(paramnum >= 2) {
		steps = HexaToInt(params[1]);
	}

	debugger->now_going = skips;

	SetKeyStatus();

	uint32_t breaked = 0;
	int step = 0;
	while (!dp->request_terminate && !breaked && !debugger->now_breakpoint && step < steps) {
		while(step < steps) {
			if (debugger->now_going <= 0) {
				debugger->now_going = skips;
			}
			debugger->now_suspended = false;
			while(!dp->request_terminate && !debugger->now_suspended && !NowDisable()) {
				CDelay(10);
			}
			dasm_addr = cpu->get_next_pc();
						
			TraceCurrent();

			cpu->debug_regs_info(buffer, DC_MAX_BUFFER_LEN);
			Out(White);
						
			if(debugger->now_breakpoint
			|| debugger->now_tracepoint
			|| (dp->request_terminate)) {
				break;
			} else if (EscapePressed()) {
				breaked |= 1;
				break;
			} else if (NowDisable()) {
				breaked |= 2;
				break;
			}

			step++;
		}

		TraceNext();

		debugger->set_trace_point(false);
	}

	PrintBreakedReason(breaked);

	SetTextColor(White);
}

void DebuggerConsole::UsageTrace(bool s)
{
	UsageCmdStr1(s, _T("T"), _T("[<count> [<skip count>]]"), _T("Trace memory while step processing."));
	if (s) return;

	Print(_T("  <count> - specify number of step."));
	Print(_T("  <skip count> - specify number of step to skip trace."));
	Print(_T("  Total steps become <count> * <skip count>."));
	Print(_T("  Stop tracing if esc key was pressed."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::TraceFirst()
{
	cpu->debug_regs_info(buffer, DC_MAX_BUFFER_LEN);
	Out(White);
	
	TraceNext();

	_stprintf(buffer, _T("breaked at %s"), _T(ADDR_FORMAT));
	Printf(Red, buffer, cpu->get_next_pc());
	Cr();

	SetTextColor(White);
}

void DebuggerConsole::TraceCurrent()
{
	cpu->debug_dasm(-1, cpu->get_pc(), buffer, DC_MAX_BUFFER_LEN, 1);
#ifndef USE_EMU_INHERENT_SPEC
	Printf(Cyan, _T("done\t%08X  "), cpu->get_pc());
#else
	Print(Cyan, _T("  "), false);
#endif
	Out();
}

void DebuggerConsole::TraceNext()
{
	cpu->debug_dasm(-1, cpu->get_next_pc(), buffer, DC_MAX_BUFFER_LEN, 0);
#ifndef USE_EMU_INHERENT_SPEC
	Printf(Cyan, _T("next\t%08X  "), cpu->get_next_pc());
#else
	Print(Cyan, _T("> "), false);
#endif
	Out();
}

void DebuggerConsole::SetKeyStatus()
{
	uint8_t *k = dp->emu->key_buffer();
	memcpy(k, key_stat_tmp, KEY_STATUS_SIZE);
	memset(key_stat_tmp, 0, KEY_STATUS_SIZE);
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandTraceBack()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageTraceBack(false);
		return;
	}

	if(paramnum > 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageTraceBack(false);
		return;
	}

	int steps = 25;
	if(paramnum >= 2) {
		steps = HexaToInt(params[1]);
		if (steps <= 0) steps = 1;
	}
	steps--;

	for(int i=steps; i>=0; ) {
		if (cpu->debug_trace_back(i, buffer, DC_MAX_BUFFER_LEN) >= -1) {
			Print(Yellow, _T("  "), false);
			Out();
		}
		if ((i = cpu->debug_trace_back_regs(i, buffer, DC_MAX_BUFFER_LEN)) >= -1) {
			Out(White);
		}
	}
	SetTextColor(White);
}

void DebuggerConsole::UsageTraceBack(bool s)
{
	UsageCmdStr1(s, _T("TB"), _T("[<count>]"), _T("Trace back the recently processed instructions."));
	if (s) return;

	Print(_T("  <count> - specify number of traceing."));
}

// ---------------------------------------------------------------------------

#ifdef USE_EMU_INHERENT_SPEC
/// show memory map
///
void DebuggerConsole::CommandShowMemoryMap()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageShowMemoryMap(false);
		return;
	}

	if(paramnum != 1) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageShowMemoryMap(false);
		return;
	}

	mem->debug_memory_map_info(this);
}

void DebuggerConsole::UsageShowMemoryMap(bool s)
{
	UsageCmdStr1(s, _T("M"), NULL, _T("Show memory map."));
	if (s) return;
}
#endif

// ---------------------------------------------------------------------------

#ifdef _MBS1
/// show address map
///
void DebuggerConsole::CommandAccessAddressMap()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageAccessAddressMap(false);
		return;
	}

	if(paramnum > 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageAccessAddressMap(false);
		return;
	}

	int idx = -1;
	if (paramnum == 2) {
		idx = (int)HexaToInt(params[1]);
	}
	mem->debug_address_map_info(this, idx);
}

void DebuggerConsole::UsageAccessAddressMap(bool s)
{
	UsageCmdStr1(s, _T("X"), _T("[<segment number>]"), _T("Show address map."));
	if (s) return;

	Print(_T("  <segment number> - specify number of address segment."));
}
#endif

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandShowClock()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageShowClock(false);
		return;
	}

	if(paramnum > 2) {
		Print(_T("invalid number of parameter(s)"));
		Cr();
		UsageShowClock(false);
		return;
	}

	uint64_t clk = dp->emu->get_current_clock();
	Printf(_T("Dec:%lld  Hex:%llX"), clk, clk);
	Cr();
}

void DebuggerConsole::UsageShowClock(bool s)
{
	UsageCmdStr1(s, _T("CL"), NULL, _T("Show current clock."));
	if (s) return;
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandQuit()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageQuit(false);
		return;
	}

	Print(_T("Quit"));
	dp->emu->get_gui()->PostMtCloseDebugger();
}

void DebuggerConsole::UsageQuit(bool s)
{
	UsageCmdStr1(s, _T("Q"), NULL, _T("Quit debugger."));
	if (s) return;
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandOutputLogFile()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageOutputLogFile(false);
		return;
	}

	if(paramnum >= 2) {
		if(logfile != NULL && logfile->IsOpened()) {
			logfile->Fclose();
			delete logfile;
			logfile = NULL;
			logfilename[0] = _T('\0');
		}
		if (_tcsicmp(params[1], _T("CLOSE")) != 0) {
			UTILITY::tcscpy(logfilename, _MAX_PATH, params[1]);
			logfile = new FILEIO();
			if (!logfile->Fopen(logfilename, FILEIO_WRITE_ASCII)) {
				delete logfile;
				logfile = NULL;
				logfilename[0] = _T('\0');
			}
		}
	}
	if (_tcslen(logfilename) > 0) {
		Printf(_T("Current log file is \"%s\"."), logfilename);
		Cr();
	} else {
		Print(_T("No log file is opened."));
	}
}

void DebuggerConsole::UsageOutputLogFile(bool s)
{
	UsageCmdStr1(s, _T(">"), _T("[<filename>]"), _T("Open/Close a log file."));
	if (s) return;

	Print(_T("  If filename is specified \"CLOSE\", close the log file."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::CommandSendToEmulator()
{
	if(paramnum >= 2 && *params[1] == _T('?')) {
		Cr();
		UsageSendToEmulator(false);
		return;
	}

	if(paramnum == 1) {
		Print(_T("invalid parameter"));
		Cr();
		UsageSendToEmulator(false);
		return;
	}

	if(_tcsicmp(params[1], _T("RESET")) == 0) {
		CommandSendResetToEmulator();
	} else if(_tcsicmp(params[1], _T("KEY")) == 0) {
		CommandSendKeyToEmulator();
	} else {
		Print(_T("unknown command ! "), false);
		Print(params[1]);
		Cr();
		UsageSendToEmulator(false);
	}
}

void DebuggerConsole::CommandSendResetToEmulator()
{
	if(paramnum == 2) {
//			dp->vm->reset();
		dp->emu->get_gui()->PostEtReset();

	} else if(paramnum == 3) {
		if(_tcsicmp(params[2], _T("ALL")) == 0) {
//				dp->vm->reset();
			dp->emu->get_gui()->PostEtReset();

		} if(_tcsicmp(params[2], _T("CPU")) == 0) {
			cpu->reset();
		} else {
			Print(_T("unknown device "), false);
			Print(params[2]);
		}
	} else {
		Print(_T("invalid parameter"));
		Cr();
		UsageSendToEmulator(false);
	}
}

void DebuggerConsole::CommandSendKeyToEmulator()
{
	if(paramnum == 3 || paramnum == 4) {
		int code =  HexaToInt(params[2]);
		if (code < 0 || KEY_STATUS_SIZE <= code) {
			Print(_T("invalid key code"));
			Cr();
			UsageSendToEmulator(false);
			return;
		}
		int msec = 100;
		if(paramnum == 4) {
			msec = HexaToInt(params[3]);
		}
#ifdef SUPPORT_VARIABLE_TIMING
		key_stat_tmp[code] = SDL_max((int)(p->vm->frame_rate() * (double)msec / 1000.0 + 0.5), 1);
#else
		key_stat_tmp[code] = SDL_max((int)(FRAMES_PER_SEC * (double)msec / 1000.0 + 0.5), 1);
#endif
#ifdef NOTIFY_KEY_DOWN
		dp->vm->key_down(code, false);
#endif
	} else {
		Print(_T("invalid parameter"));
		Cr();
		UsageSendToEmulator(false);
	}
}

void DebuggerConsole::UsageSendToEmulator(bool s)
{
	UsageCmdStr1(s, _T("!"), NULL, _T("Send command to emulator."));
	if (s) return;

	Print(_T("! reset [all,cpu]"), false);
	Print(_T(" - "), false);
	Print(_T("Send restart."));
	Print(_T("! key <code> [<msec>]"), false);
	Print(_T(" - "), false);
	Print(_T("Send a key code."));
}

// ---------------------------------------------------------------------------

void DebuggerConsole::UsageCmdStr1(bool s, const _TCHAR *cmd1, const _TCHAR *args1, const _TCHAR *desc1)
{
	Printf(_T("%-8s"), cmd1);
	Print(_T(" - "), false);
	Print(desc1);
	if (s) return;
	Cr();
	if (!args1) return;
	Print(_T("Usage: "), false);
	Print(cmd1, false);
	Print(_T(" "), false);
	Print(args1);
}

void DebuggerConsole::UsageCmdStr2(bool s, const _TCHAR *cmd1, const _TCHAR *args1, const _TCHAR *desc1
	, const _TCHAR *cmd2, const _TCHAR *args2, const _TCHAR *desc2)
{
	Printf(_T("%-8s"), cmd1);
	Print(_T(" - "), false);
	Print(desc1);
	Printf(_T("%-8s"), cmd2);
	Print(_T(" - "), false);
	Print(desc2);
	if (s) return;
	Cr();
	if (!args1) return;
	Print(_T("Usage: "), false);
	Print(cmd1, false);
	Print(_T(" "), false);
	Print(args1);
	if (!args2) return;
	Print(_T("       "), false);
	Print(cmd2, false);
	Print(_T(" "), false);
	Print(args2);
}

void DebuggerConsole::UsageCmdStr3(bool s, const _TCHAR *cmd1, const _TCHAR *args1, const _TCHAR *desc1
	, const _TCHAR *cmd2, const _TCHAR *args2, const _TCHAR *desc2
	, const _TCHAR *cmd3, const _TCHAR *args3, const _TCHAR *desc3)
{
	Printf(_T("%-8s"), cmd1);
	Print(_T(" - "), false);
	Print(desc1);
	Printf(_T("%-8s"), cmd2);
	Print(_T(" - "), false);
	Print(desc2);
	Printf(_T("%-8s"), cmd3);
	Print(_T(" - "), false);
	Print(desc3);
	if (s) return;
	Cr();
	if (!args1) return;
	Print(_T("Usage: "), false);
	Print(cmd1, false);
	Print(_T(" "), false);
	Print(args1);
	if (!args2) return;
	Print(_T("       "), false);
	Print(cmd2, false);
	Print(_T(" "), false);
	Print(args2);
	if (!args3) return;
	Print(_T("       "), false);
	Print(cmd3, false);
	Print(_T(" "), false);
	Print(args3);
}

// ---------------------------------------------------------------------------

void DebuggerConsole::Usage()
{
	UsageDumpMemory(true, 0);
#ifdef USE_EMU_INHERENT_SPEC
	UsageDumpMemory(true, 1);
#endif
	UsageEditMemory(true, 0);
#ifdef USE_EMU_INHERENT_SPEC
	UsageEditMemory(true, 1);
#endif
#ifdef USE_IOPORT_ACCESS
	UsageEditIO(true);
#endif
	UsageAccessRegister(true);
	UsageAccessDevice(true);
#ifdef USE_EMU_INHERENT_SPEC
	UsageShowMemoryMap(true);
#endif
#ifdef _MBS1
	UsageAccessAddressMap(true);
#endif
	UsageSearch(true, 0);
#ifdef USE_EMU_INHERENT_SPEC
	UsageSearch(true, 1);
#endif
	UsageUnassemble(true, 0);
#ifdef USE_EMU_INHERENT_SPEC
	UsageUnassemble(true, 1);
#endif

	UsageCalculateHexa(true);
	UsageSetFileName(true);
	UsageBinaryFile(true, 0);
#ifdef USE_EMU_INHERENT_SPEC
	UsageBinaryFile(true, 1);
#endif
#ifdef USE_EMU_INHERENT_SPEC
	UsageImageFile(true, 0);
#endif

	UsageSetBreakPoint(true, 0);
#ifdef USE_IOPORT_ACCESS
	UsageSetBreakPoint(true, 1);
#endif
	UsageSetBreakPoint(true, 2);
	UsageChangeBreakPoint(true, 0);
#ifdef USE_IOPORT_ACCESS
	UsageChangeBreakPoint(true, 1);
#endif
	UsageChangeBreakPoint(true, 2);
//	UsageListBreakPoint(true, 0);
#ifdef USE_IOPORT_ACCESS
//	UsageListBreakPoint(true, 1);
#endif
//	UsageListBreakPoint(true, 2);

	UsageExecute(true);
	UsageTrace(true);
	UsageTraceBack(true);

	UsageSetBreakPoint(true, 8);
#ifdef USE_IOPORT_ACCESS
	UsageSetBreakPoint(true, 9);
#endif
	UsageSetBreakPoint(true, 10);
	UsageChangeBreakPoint(true, 8);
#ifdef USE_IOPORT_ACCESS
	UsageChangeBreakPoint(true, 9);
#endif
	UsageChangeBreakPoint(true, 10);
//	UsageListBreakPoint(true, 8);
#ifdef USE_IOPORT_ACCESS
//	UsageListBreakPoint(true, 9);
#endif
//	UsageListBreakPoint(true, 10);

	UsageShowClock(true);

	UsageOutputLogFile(true);

	UsageSendToEmulator(true);
				
	UsageQuit(true);
	Cr();
	Print(_T("<value> - hexa, decimal(%d), ascii('a')"));
}

void DebuggerConsole::CommandUsage()
{
	if (paramnum >= 2) {
		int idx = -1;
		for(int i=0; commands_map[i].cmd != NULL; i++) {
			if (_tcsicmp(params[1], commands_map[i].cmd) == 0) {
				idx = i;
				break;
			}
		}
		if (idx >= 0) {
			switch(commands_map[idx].idx) {
			case ID_DUMP_MEMORY:
				UsageDumpMemory(false, 0);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_DUMP_PHYSICAL_MEMORY:
				UsageDumpMemory(false, 1);
				break;
#endif
			case ID_EDIT_MEMORY_BYTE:
			case ID_EDIT_MEMORY_WORD:
			case ID_EDIT_MEMORY_DWORD:
			case ID_EDIT_MEMORY_ASCII:
				UsageEditMemory(false, 0);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_EDIT_PHYSICAL_MEMORY_BYTE:
			case ID_EDIT_PHYSICAL_MEMORY_WORD:
			case ID_EDIT_PHYSICAL_MEMORY_DWORD:
			case ID_EDIT_PHYSICAL_MEMORY_ASCII:
				UsageEditMemory(false, 1);
				break;
#endif
#ifdef USE_IOPORT_ACCESS
			case ID_INPUT_IO_BYTE:
			case ID_INPUT_IO_WORD:
			case ID_INPUT_IO_DWORD:
			case ID_OUTPUT_IO_BYTE:
			case ID_OUTPUT_IO_WORD:
			case ID_OUTPUT_IO_DWORD:
				UsageEditIO(false);
				break;
#endif
			case ID_ACCESS_REGISTER:
				UsageAccessRegister(false);
				break;
			case ID_ACCESS_DEVICE:
				UsageAccessDevice(false);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_SHOW_MEMORY_MAP:
				UsageShowMemoryMap(false);
				break;
#endif
#ifdef _MBS1
			case ID_ACCESS_ADDRESS_MAP:
				UsageAccessAddressMap(false);
				break;
#endif
			case ID_SEARCH:
				UsageSearch(false, 0);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_SEARCH_PHYSICAL:
				UsageSearch(false, 1);
				break;
#endif
			case ID_UNASSEMBLE:
				UsageUnassemble(false, 0);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_UNASSEMBLE_PHYSICAL:
				UsageUnassemble(false, 1);
				break;
#endif
			case ID_CALC_HEXA:
				UsageCalculateHexa(false);
				break;
			case ID_SET_FILE_NAME:
				UsageSetFileName(false);
				break;
			case ID_LOAD_BINARY_FILE:
			case ID_SAVE_BINARY_FILE:
				UsageBinaryFile(false, 0);
				break;
#ifdef USE_EMU_INHERENT_SPEC
			case ID_LOAD_BINARY_FILE_PHYSICAL:
			case ID_SAVE_BINARY_FILE_PHYSICAL:
				UsageBinaryFile(false, 1);
				break;
#endif
#ifdef USE_EMU_INHERENT_SPEC
			case ID_SAVE_IMAGE_FILE:
				UsageImageFile(false, 0);
				break;
#endif
			case ID_SET_BREAK_POINT:
				UsageSetBreakPoint(false, 0);
				break;
			case ID_CLEAR_BREAK_POINT:
			case ID_ENABLE_BREAK_POINT:
			case ID_DISABLE_BREAK_POINT:
				UsageChangeBreakPoint(false, 0);
				break;
			case ID_LIST_BREAK_POINT:
				UsageListBreakPoint(false, 0);
				break;
#ifdef USE_IOPORT_ACCESS
			case ID_SET_BREAK_POINT_FOR_IO:
				UsageSetBreakPoint(false, 1);
				break;
			case ID_CLEAR_BREAK_POINT_FOR_IO:
			case ID_ENABLE_BREAK_POINT_FOR_IO:
			case ID_DISABLE_BREAK_POINT_FOR_IO:
				UsageChangeBreakPoint(false, 1);
				break;
			case ID_LIST_BREAK_POINT_FOR_IO:
				UsageListBreakPoint(false, 1);
				break;
#endif
			case ID_SET_BREAK_POINT_FOR_INTR:
				UsageSetBreakPoint(false, 2);
				break;
			case ID_CLEAR_BREAK_POINT_FOR_INTR:
			case ID_ENABLE_BREAK_POINT_FOR_INTR:
			case ID_DISABLE_BREAK_POINT_FOR_INTR:
				UsageChangeBreakPoint(false, 2);
				break;
			case ID_LIST_BREAK_POINT_FOR_INTR:
				UsageListBreakPoint(false, 2);
				break;
			case ID_EXECUTE:
				UsageExecute(false);
				break;
			case ID_TRACE:
				UsageTrace(false);
				break;
			case ID_TRACE_BACK:
				UsageTraceBack(false);
				break;
			case ID_SHOW_CLOCK:
				UsageShowClock(false);
				break;
			case ID_SET_TRACE_POINT:
				UsageSetBreakPoint(false, 8);
				break;
			case ID_CLEAR_TRACE_POINT:
			case ID_ENABLE_TRACE_POINT:
			case ID_DISABLE_TRACE_POINT:
				UsageChangeBreakPoint(false, 8);
				break;
			case ID_LIST_TRACE_POINT:
				UsageListBreakPoint(false, 8);
				break;
#ifdef USE_IOPORT_ACCESS
			case ID_SET_TRACE_POINT_FOR_IO:
				UsageSetBreakPoint(false, 9);
				break;
			case ID_CLEAR_TRACE_POINT_FOR_IO:
			case ID_ENABLE_TRACE_POINT_FOR_IO:
			case ID_DISABLE_TRACE_POINT_FOR_IO:
				UsageChangeBreakPoint(false, 9);
				break;
			case ID_LIST_TRACE_POINT_FOR_IO:
				UsageListBreakPoint(false, 9);
				break;
#endif
			case ID_SET_TRACE_POINT_FOR_INTR:
				UsageSetBreakPoint(false, 10);
				break;
			case ID_CLEAR_TRACE_POINT_FOR_INTR:
			case ID_ENABLE_TRACE_POINT_FOR_INTR:
			case ID_DISABLE_TRACE_POINT_FOR_INTR:
				UsageChangeBreakPoint(false, 10);
				break;
			case ID_LIST_TRACE_POINT_FOR_INTR:
				UsageListBreakPoint(false, 10);
				break;
			case ID_OUTPUT_LOG_FILE:
				UsageOutputLogFile(false);
				break;
			case ID_SEND_TO_EMULATOR:
				UsageSendToEmulator(false);
				break;
			case ID_QUIT:
				UsageQuit(false);
				break;
			default:
				Usage();
				break;
			}
		} else {
			Usage();
		}
	} else {
		Usage();
	}
}

void DebuggerConsole::PrintPrompt()
{
	Print(DEBUGGER_PROMPT, false);
	Flush();
}

void DebuggerConsole::Process()
{
	TraceFirst();

	while(!dp->request_terminate) {
		// prompt
		PrintPrompt();
		
		// get command
		bool enter_done = false;
		while(!dp->request_terminate && !enter_done) {
			enter_done = ReadInput();
			CDelay(10);
		}
		
		// process command
		if(!dp->request_terminate && enter_done) {
			ParseInput();
			int idx = -1;
			for(int i=0; commands_map[i].cmd != NULL; i++) {
				if (_tcsicmp(params[0], commands_map[i].cmd) == 0) {
					idx = i;
					break;
				}
			}
			if (idx >= 0) {
				switch(commands_map[idx].idx) {
				case ID_DUMP_MEMORY:
					CommandDumpMemory(0);
					break;
#ifdef USE_EMU_INHERENT_SPEC
				case ID_DUMP_PHYSICAL_MEMORY:
					CommandDumpMemory(1);
					break;
#endif
				case ID_EDIT_MEMORY_BYTE:
					CommandEditMemoryBinary(1, 0);
					break;
				case ID_EDIT_MEMORY_WORD:
					CommandEditMemoryBinary(2, 0);
					break;
				case ID_EDIT_MEMORY_DWORD:
					CommandEditMemoryBinary(4, 0);
					break;
				case ID_EDIT_MEMORY_ASCII:
					CommandEditMemoryAscii(0);
					break;
#ifdef USE_EMU_INHERENT_SPEC
				case ID_EDIT_PHYSICAL_MEMORY_BYTE:
					CommandEditMemoryBinary(1, 1);
					break;
				case ID_EDIT_PHYSICAL_MEMORY_WORD:
					CommandEditMemoryBinary(2, 1);
					break;
				case ID_EDIT_PHYSICAL_MEMORY_DWORD:
					CommandEditMemoryBinary(4, 1);
					break;
				case ID_EDIT_PHYSICAL_MEMORY_ASCII:
					CommandEditMemoryAscii(1);
					break;
#endif
#ifdef USE_IOPORT_ACCESS
				case ID_INPUT_IO_BYTE:
					CommandInputIOBinary(1);
					break;
				case ID_INPUT_IO_WORD:
					CommandInputIOBinary(2);
					break;
				case ID_INPUT_IO_DWORD:
					CommandInputIOBinary(4);
					break;
				case ID_OUTPUT_IO_BYTE:
					CommandOutputIOBinary(1);
					break;
				case ID_OUTPUT_IO_WORD:
					CommandOutputIOBinary(2);
					break;
				case ID_OUTPUT_IO_DWORD:
					CommandOutputIOBinary(4);
					break;
#endif
				case ID_ACCESS_REGISTER:
					CommandAccessRegister();
					break;
				case ID_ACCESS_DEVICE:
					CommandAccessDevice();
					break;
				case ID_SHOW_MEMORY_MAP:
					CommandShowMemoryMap();
					break;
#ifdef _MBS1
				case ID_ACCESS_ADDRESS_MAP:
					CommandAccessAddressMap();
					break;
#endif
				case ID_SEARCH:
					CommandSearch(0);
					break;
#ifdef USE_EMU_INHERENT_SPEC
				case ID_SEARCH_PHYSICAL:
					CommandSearch(1);
					break;
#endif
				case ID_UNASSEMBLE:
					CommandUnassemble(0);
					break;
#ifdef USE_EMU_INHERENT_SPEC
				case ID_UNASSEMBLE_PHYSICAL:
					CommandUnassemble(1);
					break;
#endif
				case ID_CALC_HEXA:
					CommandCalculateHexa();
					break;
				case ID_SET_FILE_NAME:
					CommandSetFileName();
					break;
				case ID_LOAD_BINARY_FILE:
					CommandLoadBinaryFile(0);
					break;
				case ID_SAVE_BINARY_FILE:
					CommandSaveBinaryFile(0);
					break;
#ifdef USE_EMU_INHERENT_SPEC
				case ID_LOAD_BINARY_FILE_PHYSICAL:
					CommandLoadBinaryFile(1);
					break;
				case ID_SAVE_BINARY_FILE_PHYSICAL:
					CommandSaveBinaryFile(1);
					break;
#endif
#ifdef USE_EMU_INHERENT_SPEC
				case ID_SAVE_IMAGE_FILE:
					CommandSaveImageFile(0);
					break;
#endif
				case ID_SET_BREAK_POINT:
					CommandSetBreakPoint(0);
					break;
				case ID_CLEAR_BREAK_POINT:
					CommandClearBreakPoint(0);
					break;
				case ID_DISABLE_BREAK_POINT:
				case ID_ENABLE_BREAK_POINT:
					CommandChangeBreakPoint(0);
					break;
				case ID_LIST_BREAK_POINT:
					CommandListBreakPoint(0);
					break;
#ifdef USE_IOPORT_ACCESS
				case ID_SET_BREAK_POINT_FOR_IO:
					CommandSetBreakPoint(1);
					break;
				case ID_CLEAR_BREAK_POINT_FOR_IO:
					CommandClearBreakPoint(1);
					break;
				case ID_DISABLE_BREAK_POINT_FOR_IO:
				case ID_ENABLE_BREAK_POINT_FOR_IO:
					CommandChangeBreakPoint(1);
					break;
				case ID_LIST_BREAK_POINT_FOR_IO:
					CommandListBreakPoint(1);
					break;
#endif
				case ID_SET_BREAK_POINT_FOR_INTR:
					CommandSetBreakPoint(2);
					break;
				case ID_CLEAR_BREAK_POINT_FOR_INTR:
					CommandClearBreakPoint(2);
					break;
				case ID_DISABLE_BREAK_POINT_FOR_INTR:
				case ID_ENABLE_BREAK_POINT_FOR_INTR:
					CommandChangeBreakPoint(2);
					break;
				case ID_LIST_BREAK_POINT_FOR_INTR:
					CommandListBreakPoint(2);
					break;
				case ID_SET_TRACE_POINT:
					CommandSetBreakPoint(8);
					break;
				case ID_CLEAR_TRACE_POINT:
					CommandClearBreakPoint(8);
					break;
				case ID_DISABLE_TRACE_POINT:
				case ID_ENABLE_TRACE_POINT:
					CommandChangeBreakPoint(8);
					break;
				case ID_LIST_TRACE_POINT:
					CommandListBreakPoint(8);
					break;
#ifdef USE_IOPORT_ACCESS
				case ID_SET_TRACE_POINT_FOR_IO:
					CommandSetBreakPoint(9);
					break;
				case ID_CLEAR_TRACE_POINT_FOR_IO:
					CommandClearBreakPoint(9);
					break;
				case ID_DISABLE_TRACE_POINT_FOR_IO:
				case ID_ENABLE_TRACE_POINT_FOR_IO:
					CommandChangeBreakPoint(9);
					break;
				case ID_LIST_TRACE_POINT_FOR_IO:
					CommandListBreakPoint(9);
					break;
#endif
				case ID_SET_TRACE_POINT_FOR_INTR:
					CommandSetBreakPoint(10);
					break;
				case ID_CLEAR_TRACE_POINT_FOR_INTR:
					CommandClearBreakPoint(10);
					break;
				case ID_DISABLE_TRACE_POINT_FOR_INTR:
				case ID_ENABLE_TRACE_POINT_FOR_INTR:
					CommandChangeBreakPoint(10);
					break;
				case ID_LIST_TRACE_POINT_FOR_INTR:
					CommandListBreakPoint(10);
					break;
				case ID_EXECUTE:
					CommandExecute();
					break;
				case ID_TRACE:
					CommandTrace();
					break;
				case ID_TRACE_BACK:
					CommandTraceBack();
					break;
				case ID_SHOW_CLOCK:
					CommandShowClock();
					break;
				case ID_QUIT:
					CommandQuit();
					break;
				case ID_OUTPUT_LOG_FILE:
					CommandOutputLogFile();
					break;
				case ID_SEND_TO_EMULATOR:
					CommandSendToEmulator();
					break;
				case ID_USAGE:
					CommandUsage();
					break;
				default:
					Printf(_T("unsupported command %s\n"), params[0]);
					break;
				}
			} else {
				Printf(_T("unknown command %s\n"), params[0]);
			}
		}
	}
	
	// stop debugger
	try {
		debugger->now_going = 0;
		debugger->now_debugging = debugger->now_suspended = false;
	} catch(...) {
	}
}


int SDLCALL debugger_thread(void *lpx)
{
	volatile debugger_thread_t *p = (debugger_thread_t *)lpx;
	p->running = true;

	// initialize
	DebuggerConsole *dc = new DebuggerConsole(p);

	dc->Process();

	// release
	delete dc;
	
	p->running = false;

	return 0;
}

// ----------------------------------------------------------------------------
	
void EMU::EMU_DEBUGGER()
{
	hDebuggerThread = NULL;
}

void EMU::initialize_debugger()
{
	now_debugging = false;
}

void EMU::release_debugger()
{
	close_debugger();
}

void EMU::open_debugger(int cpu_index)
{
	if(!(now_debugging && debugger_thread_param.cpu_index == cpu_index)) {
		close_debugger();
		if(vm->get_cpu(cpu_index) != NULL && vm->get_cpu(cpu_index)->get_debugger() != NULL) {
			debugger_thread_param.emu = this;
			debugger_thread_param.vm = vm;
			debugger_thread_param.cpu_index = cpu_index;
			debugger_thread_param.request_terminate = false;
#ifndef USE_SDL2
			if ((hDebuggerThread = SDL_CreateThread(debugger_thread, &debugger_thread_param)) != NULL) {
#else
			if ((hDebuggerThread = SDL_CreateThread(debugger_thread, "debugger_thread", &debugger_thread_param)) != NULL) {
#endif
				stop_rec_sound();
				stop_rec_video();
#ifdef USE_TELNET_SERVER
				// acceptable from telnet terminal
				DEBUGGER *debugger = (DEBUGGER *)vm->get_cpu(debugger_thread_param.cpu_index)->get_debugger();
				debugger->enable_server(true);
#endif
				now_debugging = true;

				out_log(LOG_INFO, _("Debugger was started."));
			} else {
				out_log(LOG_ERROR, _("Cannot start debugger."));
			}
		}
	}
}

void EMU::close_debugger()
{
	if(now_debugging) {
		if(debugger_thread_param.running) {
			debugger_thread_param.request_terminate = true;
			int status;
			SDL_WaitThread(hDebuggerThread, &status);
		}
#ifdef USE_TELNET_SERVER
		_TCHAR data[32];
		DEBUGGER *debugger = (DEBUGGER *)vm->get_cpu(debugger_thread_param.cpu_index)->get_debugger();
		// back to default color
		_tcscpy(data, _T("\x1b[39m\x1b[49m"));
		debugger->write_data(data, (int)_tcslen(data));
		// disconnect
		debugger->enable_server(false);
#endif
		hDebuggerThread = NULL;
		now_debugging = false;

		out_log(LOG_INFO, _("Debugger was stopped."));
	}
}

bool EMU::debugger_enabled(int cpu_index)
{
	return (vm->get_cpu(cpu_index) != NULL && vm->get_cpu(cpu_index)->get_debugger() != NULL);
}

void EMU::debugger_terminal_accepted(DEBUGGER *debugger)
{
	// send telnet command
	_TCHAR data[128];
	// telnet options
	_tcscpy(data, _T("\xff\xfb\x03"));	// WILL Suppress Go Ahead
	_tcscat(data, _T("\xff\xfb\x01"));	// WILL Echo
	debugger->write_data(data, (int)_tcslen(data));
	// send color code
	_tcscpy(data, _T("\x1b[37m\x1b[40m"));
	debugger->write_data(data, (int)_tcslen(data));
	// message
	UTILITY::stprintf(data, 128, _T("Connected to %s debugger. Type ? to see help.\n"), CLASS_NAME);
	debugger->write_data(data, (int)_tcslen(data));
	// prompt
	_tcscpy(data, DEBUGGER_PROMPT);
	debugger->write_data(data, (int)_tcslen(data));
}

#include "video/rec_video.h"

bool EMU::debugger_save_image(int width, int height, CSurface *surface)
{
	VmRectWH size;

	size.x = 0; size.y = 0; size.w = width; size.h = height;
	return rec_video->Capture(SAVE_IMAGE_TYPE, &size, surface, &size);
}

#endif

