/** @file sdl_screen.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl screen ]

	@note
	This code is based on win32_screen.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "gui/gui.h"
#include "config.h"
#ifdef USE_MESSAGE_BOARD
#include "msgboard.h"
#endif
#include "csurface.h"
#include "video/rec_video.h"
#include "sdl_utils.h"

static const VmSize window_proposal[] = {
	{640,400},{640,480},{800,600},{1024,768},{0,0}
};

void EMU::EMU_SCREEN()
{
	screen_width = SCREEN_WIDTH;
	screen_height = SCREEN_HEIGHT;
	screen_width_aspect = SCREEN_WIDTH_ASPECT;
	screen_height_aspect = SCREEN_HEIGHT_ASPECT;
//	window_width = WINDOW_WIDTH;
//	window_height = WINDOW_HEIGHT;
	min_window_width = MIN_WINDOW_WIDTH;
	min_window_height = MIN_WINDOW_HEIGHT;
	screen_size_changed = true;
	desktop_width = MIN_WINDOW_WIDTH;
	desktop_height = MIN_WINDOW_HEIGHT;
	display_width = MIN_WINDOW_WIDTH;
	display_height = MIN_WINDOW_HEIGHT;
	display_margin.left = 0;
	display_margin.top = 0;
	display_margin.right = 0;
	display_margin.bottom = 0;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = source_height_aspect = -1;
	stretch_pow_x = stretch_pow_y = -1;
//	display_stretched_width = display_stretched_height = 1;
	stretched_width = SCREEN_WIDTH;
	stretched_height = SCREEN_HEIGHT;
	stretched_dest_x = stretched_dest_y = 0;
	stretched_dest_real_x = stretched_dest_real_y = 0;
	mixed_width = source_width;
	mixed_height = source_height;
	mixed_dest_x = mixed_dest_y = 0;
	mixed_width_ratio = mixed_height_ratio = 1;

	// screen mode
	screen_mode_count = 0;
	window_mode_count = 0;
	window_mode_power = 10;
	now_fullscreen = false;
	prev_window_mode = 0;

	// initialize video recording
	rec_video_size[0].w = MIN_WINDOW_WIDTH;
	rec_video_size[0].h = MIN_WINDOW_HEIGHT;
	rec_video_size[0].x = (MAX_WINDOW_WIDTH - MIN_WINDOW_WIDTH) / 2;
	rec_video_size[0].y = (MAX_WINDOW_HEIGHT - MIN_WINDOW_HEIGHT) / 2;
	rec_video_size[1].w = MAX_WINDOW_WIDTH;
	rec_video_size[1].h = MAX_WINDOW_HEIGHT;
	rec_video_size[1].x = 0;
	rec_video_size[1].y = 0;
	rec_video_stretched_size = rec_video_size[0];

	rec_video = new REC_VIDEO(this);
	now_recording_video = rec_video->GetNowRecordingPtr();

	// initialize update flags
	first_draw_screen = false;
	first_invalidate = self_invalidate = false;
	skip_frame = false;

	//
	screen = NULL;
	srcsurface = new CSurface();
	mixsurface = new CSurface();
	screen_flags = 0;

#ifdef USE_SDL2
	window = NULL;
#ifdef USE_OPENGL
	glcontext = NULL;
#endif
#endif

#ifdef USE_OPENGL
	mix_texture_name = 0;
	src_tex_l = src_tex_t = 0.0;
	src_tex_r = src_tex_b = 1.0;
#endif

	// for lock screen
	mux_update_screen = new CMutex();
	screen_changing = false;
}

void EMU::initialize_screen()
{
	out_debug(_T("EMU::initialize_screen"));

#ifdef USE_OPENGL
	initialize_opengl();
#endif
	change_pixel_aspect(config.pixel_aspect);
	change_rec_video_size(config.screen_video_size);

#ifdef USE_MESSAGE_BOARD
	msgboard = new MsgBoard(this);
	if (msgboard) {
		msgboard->InitScreen(&pixel_format, screen_width, screen_height);
		msgboard->SetMessagePos(4 + reMix.x, -4 - reMix.y, 2);
		msgboard->SetInfoPos(-4 - reMix.x, 4 + reMix.y, 1);
		msgboard->SetVisible(FLG_SHOWMSGBOARD ? true : false);
	}
#endif
	if (gui) {
		gui->CreateLedBox(res_path.GetN(), &pixel_format);
	}
}

void EMU::release_screen()
{
	delete srcsurface;
	delete mixsurface;

#ifdef USE_OPENGL
	release_opengl();
#endif
	if (gui) {
		gui->ReleaseLedBox();
	}

	delete mux_update_screen;

	delete rec_video;
}

void EMU::release_screen_on_emu_thread()
{
	// stop video recording
	stop_rec_video();
}

bool EMU::create_screen(int width, int height, uint32_t flags)
{
	bool rc = true;
//	Uint32 set_flags = 0;
	VmRect margin = { 0, 0, 0, 0 };

	lock_screen();

	if (now_fullscreen) {
		flags |= SDL_WINDOW_FULLSCREEN;
	} else {
		flags &= ~SDL_WINDOW_FULLSCREEN;
		margin = display_margin;
	}
	flags &= ~SDL_WINDOW_OPENGL;
#ifdef USE_OPENGL
	if (use_opengl) {
		flags |= SDL_WINDOW_OPENGL;
		release_texture();
		set_opengl_attr();
	}
#endif
#ifndef USE_SDL2
	screen = SDL_SetVideoMode(margin.left + width + margin.right, margin.top + height + margin.bottom, _RGB888, flags);
	if (screen == NULL) {
		out_logf(LOG_ERROR, _T("SDL_SetVideoMode: %s."), SDL_GetError());
		rc = false;
		goto FIN;
	}
	screen_flags = screen->flags;

#else
#ifdef USE_OPENGL
	if (glcontext) {
		SDL_GL_DeleteContext(glcontext);
	}
#endif
	if (window) {
		if ((flags & SDL_WINDOW_FULLSCREEN) == 0) {
			// go window mode. So, set window size.
			if (SDL_SetWindowFullscreen(window, flags) < 0) {
				out_logf(LOG_WARN, _T("SDL_SetWindowFullscreen: %s."), SDL_GetError());
			}
			SDL_SetWindowSize(window, margin.left + width + margin.right, margin.top + height + margin.bottom);
		} else {
			// go fullscreen mode. So, set screen mode
			SDL_DisplayMode mode;
			if (SDL_GetWindowDisplayMode(window, &mode) < 0) {
				out_logf(LOG_WARN, _T("SDL_GetWindowDisplayMode: %s."), SDL_GetError());
			}
			mode.w = margin.left + width + margin.right;
			mode.h = margin.top + height + margin.bottom;
			mode.format = 0;
			mode.refresh_rate = 0;
//			mode.driverdata = NULL;
			if (SDL_SetWindowDisplayMode(window, &mode) < 0) {
				out_logf(LOG_WARN, _T("SDL_SetWindowDisplayMode: %s."), SDL_GetError());
			}
			if (SDL_SetWindowFullscreen(window, flags) < 0) {
				out_logf(LOG_WARN, _T("SDL_SetWindowFullscreen: %s."), SDL_GetError());
			}
		}
	} else {
#ifndef GUI_USE_FOREIGN_WINDOW
		window = SDL_CreateWindow(DEVICE_NAME, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
			margin.left + width + margin.right, margin.top + height + margin.bottom, flags);
		if (window == NULL) {
			out_logf(LOG_ERROR, _T("SDL_CreateWindow: %s."), SDL_GetError());
			rc = false;
			goto FIN;
		}
#else
		void *data = gui->GetWindowData();
		if (data != NULL) {
			window = SDL_CreateWindowFrom(data);
		}
		if (window == NULL) {
			out_logf(LOG_ERROR, _T("SDL_CreateWindowFrom: %s."), SDL_GetError());
			rc = false;
			goto FIN;
		}
#ifdef USE_OPENGL
		// TODO: set flags on window ...
#endif
#endif
	}

	screen_flags = SDL_GetWindowFlags(window);
	screen = SDL_GetWindowSurface(window);
	if (screen == NULL) {
		out_logf(LOG_ERROR, _T("SDL_GetWindowSurface: %s."), SDL_GetError());
		rc = false;
		goto FIN;
	}
#endif
	if (flags != screen_flags) {
		out_debugf(_T("SDL: different flags 0x%x -> 0x%x."), flags, screen_flags);
	}
	pixel_format = *(screen->format);

#ifdef _DEBUG_LOG
	out_debugf(_T("screen_flags:0x%x"),screen_flags);
	out_debugf(_T("screen->w:%d h:%d"),screen->w, screen->h);
	out_debugf(_T("screen->pitch:%d"), screen->pitch);
	out_debugf(_T("screen->format->BitsPerPixel:%d"),screen->format->BitsPerPixel);
	out_debugf(_T("screen->format->BytesPerPixel:%d"),screen->format->BytesPerPixel);
	out_debugf(_T("screen->format->Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),screen->format->Rmask,screen->format->Gmask,screen->format->Bmask,screen->format->Amask);
	out_debugf(_T("screen->format->Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),screen->format->Rshift,screen->format->Gshift,screen->format->Bshift,screen->format->Ashift);
	out_debugf(_T("screen->format->Rloss:%d, Gloss:%d, Bloss:%d, Aloss:%d"),screen->format->Rloss,screen->format->Gloss,screen->format->Bloss,screen->format->Aloss);
#ifndef USE_SDL2
	out_debugf(_T("screen->format->colorkey:%d"),screen->format->colorkey);
	out_debugf(_T("screen->format->alpha:%d"),screen->format->alpha);
#endif
#endif
#ifdef USE_OPENGL
	if (flags & SDL_WINDOW_OPENGL) {
#ifdef USE_SDL2
		glcontext = SDL_GL_CreateContext(window);
		if (glcontext == NULL) {
			out_logf(LOG_WARN, _T("SDL_GL_CreateContext: %s."), SDL_GetError());
		}
#endif
		GLubyte *vender   = (GLubyte *)glGetString(GL_VENDOR);
		GLubyte *renderer = (GLubyte *)glGetString(GL_RENDERER);
		GLubyte *version  = (GLubyte *)glGetString(GL_VERSION);
		out_logf(LOG_DEBUG, _T("Use OPENGL: VENDER:%s RENDERER:%s VERSION:%s"), vender, renderer, version);

		// set pixel format
		SDL_UTILS::set_pixel_format_for_opengl(&pixel_format);

		// setting polygon
		set_opengl_poly(screen->w, screen->h);

		// create texture
		create_texture();
	} else {
		use_opengl = 0;
	}
#endif
	// create gui window
	if (gui->CreateWidget(screen, screen->w, screen->h) == -1) {
		rc = false;
		goto FIN;
	}

FIN:
	unlock_screen();

	return rc;
}

bool EMU::create_offlinesurface()
{
	if (screen != NULL) {
		if (!srcsurface->Create(screen_width, screen_height, &pixel_format)) {
			out_logf(LOG_ERROR, _T("CSurface::Create: srcsurface: %s."),SDL_GetError());
			return false;
		}
		// mixsurface size normalize 2^n for opengl
		mixed_max_width = 0;
		mixed_max_height = 0;
		// 65536,32768,16384,8192,4096,2048,1024,512,256,128,64,32,16
		for(int i=16; i>=4; i--) {
			if (screen_width <= (1 << i)) mixed_max_width = (1 << i);
			if (screen_height <= (1 << i)) mixed_max_height = (1 << i);
		}
		if (!mixsurface->Create(mixed_max_width, mixed_max_height, &pixel_format)) {
			out_logf(LOG_ERROR, _T("CSurface::Create: mixsurface: %s."),SDL_GetError());
			return false;
		}
	}
	return true;
}

void EMU::get_rgbformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift)
{
	if (screen) {
		*r_mask = pixel_format.Rmask;
		*g_mask = pixel_format.Gmask;
		*b_mask = pixel_format.Bmask;
		*r_shift = pixel_format.Rshift;
		*g_shift = pixel_format.Gshift;
		*b_shift = pixel_format.Bshift;
	} else {
		*r_mask = 0;
		*g_mask = 0;
		*b_mask = 0;
		*r_shift = 0;
		*g_shift = 0;
		*b_shift = 0;
	}
}

void EMU::get_rgbaformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint32_t *a_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift, uint8_t *a_shift)
{
	if (screen) {
		*r_mask = pixel_format.Rmask;
		*g_mask = pixel_format.Gmask;
		*b_mask = pixel_format.Bmask;
		*a_mask = ~(pixel_format.Rmask | pixel_format.Gmask | pixel_format.Bmask);
		*r_shift = pixel_format.Rshift;
		*g_shift = pixel_format.Gshift;
		*b_shift = pixel_format.Bshift;
		for(int i=0; i<32; i++) {
			if (((*a_mask) >> i) & 1) {
				*a_shift = i;
				break;
			}
		}
	} else {
		*a_mask = 0;
		*r_mask = 0;
		*g_mask = 0;
		*b_mask = 0;
		*a_shift = 0;
		*r_shift = 0;
		*g_shift = 0;
		*b_shift = 0;
	}
}

void EMU::get_rgbcolor(scrntype pixel, uint8_t *r, uint8_t *g, uint8_t *b)
{
	if (screen) {
		SDL_GetRGB((Uint32)pixel, &pixel_format, r, g, b);
	} else {
		*r = 0;
		*g = 0;
		*b = 0;
	}
}

scrntype EMU::map_rgbcolor(uint8_t r, uint8_t g, uint8_t b)
{
	if (screen) {
		return (scrntype)SDL_MapRGB(&pixel_format, r, g, b);
	} else {
		return 0;
	}
}

#if 0
int EMU::get_window_width(int mode)
{
	return window_width + screen_width_aspect * mode;
}

int EMU::get_window_height(int mode)
{
	return window_height + screen_height_aspect * mode;
}
#endif

/// setting window or fullscreen size
/// @param [in] width
/// @param [in] height
/// @param [in] power magnify x 10
/// @param [in] window_mode  true:window / false:fullscreen
void EMU::set_display_size(int width, int height, int power, bool window_mode)
{
	bool stretch_changed = false;

	if(width != -1 && (display_width != width || display_height != height)) {
		display_width = width;
		display_height = height;
		stretch_changed = true;
	}

//	int source_width_aspect, source_height_aspect;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = screen_width_aspect;
	source_height_aspect = screen_height_aspect;

	// fullscreen and stretch screen
	if(config.stretch_screen && !window_mode && display_width >= source_width && display_height >= source_height) {
		if (config.stretch_screen == 1) {
			// fit to full screen
			mixed_width = source_width;
			mixed_height = source_height;
			if (mixed_width_ratio < mixed_height_ratio) {
				mixed_height = mixed_height * mixed_width_ratio / mixed_height_ratio;
			} else {
				mixed_width = mixed_width * mixed_height_ratio / mixed_width_ratio;
			}
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

			stretched_width = (display_height * source_width_aspect) / source_height_aspect;
			stretched_height = display_height;
			stretched_dest_real_x = - mixed_dest_x * display_height / source_height_aspect;
			stretched_dest_real_y = - mixed_dest_y * display_height / source_height_aspect;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * source_height_aspect) / source_width_aspect;
				stretched_dest_real_x = - mixed_dest_x * display_width / source_width_aspect;
				stretched_dest_real_y = - mixed_dest_y * display_width / source_width_aspect;
			}
			if (mixed_width_ratio < mixed_height_ratio) {
				stretched_dest_real_y = stretched_dest_real_y * mixed_height_ratio / mixed_width_ratio;
			} else {
				stretched_dest_real_x = stretched_dest_real_x * mixed_width_ratio / mixed_height_ratio;
			}
			stretched_dest_x = (display_width - stretched_width) / 2;
			stretched_dest_y = (display_height - stretched_height) / 2;
			stretched_dest_real_x += stretched_dest_x;
			stretched_dest_real_y += stretched_dest_y;
		} else {
			// fit text area to full screen (cut off padding area)
			int sw;
			int sh;
			sw = LIMIT_MIN_WINDOW_WIDTH;
			sh = LIMIT_MIN_WINDOW_HEIGHT;
			int mw = sw;
			int mh = sh;
			if (mixed_width_ratio < mixed_height_ratio) {
				mh = mh * mixed_height_ratio / mixed_width_ratio;
			} else {
				mw = mw * mixed_width_ratio / mixed_height_ratio;
			}
			int mmw = 0;
			int mmh = 0;

			stretched_width = (display_height * mw) / mh;
			stretched_height = display_height;
			stretched_dest_real_x = 0;
			stretched_dest_real_y = 0;
			mmw = (display_width * mh / display_height) - mw;
			mmh = 0;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * mh) / mw;
				mmw = 0;
				mmh = (display_height * mw / display_width) - mh;

				stretched_height = display_height;

				if (sh + mmh > source_height) {
					mmh = sh + mmh - source_height;
					stretched_height -= mmh * display_width / mw;
					stretched_dest_real_y = 0;
				} else {
					stretched_dest_real_y = (sh + mmh - source_height) * display_width / mw / 2;
				}
				stretched_dest_real_x = (sw - source_width) * display_width / sw / 2;
			} else {
				stretched_width = display_width;

				if (sw + mmw > source_width) {
					mmw = sw + mmw - source_width;
					stretched_width -= mmw * display_height / mh;
					stretched_dest_real_x = 0;
				} else {
					stretched_dest_real_x = (sw + mmw - source_width) * display_height / mh / 2;
				}
				stretched_dest_real_y = (sh - source_height) * display_height / sh / 2;
			}
			mixed_width  = sw + mmw;
			mixed_height = sh + mmh;
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

			stretched_dest_x = (stretched_width  < display_width  ? (display_width  - stretched_width) / 2 : 0);
			stretched_dest_y = (stretched_height < display_height ? (display_height - stretched_height) / 2 : 0);
			stretched_dest_real_x += stretched_dest_x;
			stretched_dest_real_y += stretched_dest_y;
		}
	}
	// window or non-streach mode
	else {
		for(int n = 0; n <= 1; n++) {
			if (n == 0) {
				mixed_width = display_width * 10 / power;
				mixed_height = display_height * 10 / power;
			} else {
				mixed_width = source_width;
				mixed_height = source_height;
			}
			if (mixed_width_ratio < mixed_height_ratio) {
				mixed_height = mixed_height * mixed_width_ratio / mixed_height_ratio;
			} else {
				mixed_width = mixed_width * mixed_height_ratio / mixed_width_ratio;
			}
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			if (mixed_dest_x >= 0 && mixed_dest_y >= 0) {
				break;
			}
		}
		mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

		stretched_width = source_width_aspect * power / 10;
		stretched_height = source_height_aspect * power / 10;
		stretched_dest_x = (display_width - stretched_width) / 2;
		stretched_dest_y = (display_height - stretched_height) / 2;
		stretched_dest_real_x = - mixed_dest_x * power / 10;
		stretched_dest_real_y = - mixed_dest_y * power / 10;
		if (mixed_width_ratio < mixed_height_ratio) {
			stretched_dest_real_y = stretched_dest_real_y * mixed_height_ratio / mixed_width_ratio;
		} else {
			stretched_dest_real_x = stretched_dest_real_x * mixed_width_ratio / mixed_height_ratio;
		}
		if (stretched_dest_x < 0) {
			stretched_dest_x = 0;
			stretched_width = display_width;
		}
		if (stretched_dest_y < 0) {
			stretched_dest_y = 0;
			stretched_height = display_height;
		}
		stretched_dest_real_x += stretched_dest_x;
		stretched_dest_real_y += stretched_dest_y;
	}

	reMix.x = mixed_dest_x; reMix.y = mixed_dest_y; reMix.w = mixed_width; reMix.h = mixed_height;
	reSuf.x = stretched_dest_x; reSuf.y = stretched_dest_y; reSuf.w = stretched_width; reSuf.h = stretched_height;

#ifdef USE_OPENGL
	src_tex_l = (GLdouble)mixed_dest_x / mixed_max_width;
	src_tex_t = (GLdouble)mixed_dest_y / mixed_max_height;
	src_tex_r = (GLdouble)(mixed_dest_x + mixed_width) / mixed_max_width;
	src_tex_b = (GLdouble)(mixed_dest_y + mixed_height) / mixed_max_height;
	reTex.x = stretched_dest_x;
	reTex.y = stretched_dest_y;
	reTex.w = stretched_width;
	reTex.h = stretched_height;
#endif

	change_rec_video_size(config.screen_video_size);

#ifdef _DEBUG_LOG
	out_debugf(_T("set_display_size: w:%d h:%d power:%d %s"),width,height,power,window_mode ? _T("window") : _T("fullscreen"));
	out_debugf(_T("         display: w:%d h:%d"),display_width, display_height);
	out_debugf(_T("          screen: w:%d h:%d"), screen_width, screen_height);
	out_debugf(_T("          source: w:%d h:%d"), source_width, source_height);
	out_debugf(_T("           mixed: w:%d h:%d"), mixed_width, mixed_height);
	out_debugf(_T("         stretch: w:%d h:%d"), stretched_width, stretched_height);
	out_debugf(_T("      mixed dest: x:%d y:%d"), mixed_dest_x, mixed_dest_y);
	out_debugf(_T("    stretch dest: x:%d y:%d"), stretched_dest_x, stretched_dest_y);
	out_debugf(_T(" stretch dest re: x:%d y:%d"), stretched_dest_real_x, stretched_dest_real_y);
#ifdef USE_OPENGL
	out_debugf(_T("     src texture: l:%.5f t:%.5f r:%.5f b:%.5f"), src_tex_l, src_tex_t, src_tex_r, src_tex_b);
	out_debugf(_T("         texture: x:%d y:%d w:%d h:%d"), reTex.x, reTex.y, reTex.w, reTex.h);
#endif
#endif
	if (window_mode) {
		stretched_dest_x += display_margin.left;
		stretched_dest_y += display_margin.top;
		reSuf.x += display_margin.left;
		reSuf.y += display_margin.top;
#ifdef _DEBUG_LOG
		out_debugf(_T(" margin      suf: l:%d t:%d w:%d h:%d"),reSuf.x,reSuf.y,reSuf.w,reSuf.h);
#endif
#ifdef USE_OPENGL
		reTex.x += display_margin.left;
		reTex.y += display_margin.top;
#ifdef _DEBUG_LOG
		out_debugf(_T("             tex: l:%d t:%d w:%d h:%d"),reTex.x,reTex.y,reTex.w,reTex.h);
#endif
#endif
	}

	// send display size to vm
	set_vm_display_size();

	first_draw_screen = false;
	first_invalidate = true;

	if (gui) {
		gui->SetLedBoxPosition(window_mode, mixed_dest_x, mixed_dest_y, mixed_width, mixed_height, config.led_pos);
	}
#ifdef USE_MESSAGE_BOARD
	if (msgboard) {
		msgboard->SetSize(source_width, source_height);
		msgboard->SetMessagePos(4 + mixed_dest_x,  - 4 - source_height + mixed_dest_y + mixed_height, 2);
		msgboard->SetInfoPos(-4 - mixed_dest_x, 4 + mixed_dest_y, 1);
	}
#endif

}

#if 0
void EMU::change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh)
{
}
#endif

/// set vm drawing area
void EMU::set_vm_display_size()
{
	if (vm) {
		int new_size = config.screen_video_size;
		int l = mixed_dest_x;
		int t = mixed_dest_y;
		int r = l + mixed_width;
		int b = t + mixed_height;

		// vm drawing area is same as area to capture and record screen
		if (rec_video_size[new_size].w > mixed_width) {
			l = rec_video_size[new_size].x;
			r = rec_video_size[new_size].x + rec_video_size[new_size].w;
		}

		if (rec_video_size[new_size].h > mixed_height) {
			t = rec_video_size[new_size].y;
			b = rec_video_size[new_size].y + rec_video_size[new_size].h;
		}

		vm->set_display_size(l,t,r,b);
	}
}

/// reserve top margin area for ledbox
int EMU::adjust_y_position(int h, int y)
{
	if (20 < h - LIMIT_MIN_WINDOW_HEIGHT && h - LIMIT_MIN_WINDOW_HEIGHT < 42) {
		y -= (42 - h + LIMIT_MIN_WINDOW_HEIGHT) / 2;
	}
	return y;
}

///
/// draw src screen from virtual machine
///
void EMU::draw_screen()
{
	// draw screen
	lock_screen();

	if (srcsurface->Lock()) {
		if (!config.now_power_off) {
			vm->draw_screen();
		} else {
			fill_gray();
		}
		srcsurface->Unlock();
	}

	if (FLG_SHOWLEDBOX && gui) {
		gui->DrawLedBox(srcsurface);
	}

	unlock_screen();

	first_draw_screen = true;
}

void EMU::fill_gray()
{
	scrntype color = map_rgbcolor(0x40, 0x40, 0x40);
	for(int y=0; y<source_height; y++) {
		scrntype *line = screen_buffer(y);
		for(int x=0; x<source_width; x++) {
			*line = color;
			line++;
		}
	}
}

///
/// copy src screen to mix screen and overlap a message
/// @return false: cannot mix (no allocate mix surface)
///
bool EMU::mix_screen()
{
	if (!mixsurface) return false;

	srcsurface->Blit(NULL, *mixsurface, NULL);
	if (gui) {
		gui->UpdateIndicator(update_led());
	}
#ifdef USE_MESSAGE_BOARD
	if (msgboard && FLG_SHOWMSGBOARD) {
		msgboard->Draw(mixsurface->Get());
	}
#endif

#ifdef USE_OPENGL
	if (screen_flags & SDL_WINDOW_OPENGL) {
		if (mix_texture_name > 0) {
			GLenum gl_errno;

			// draw texture using screen pixel buffer
			glBindTexture(GL_TEXTURE_2D, mix_texture_name);

			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, mixsurface->Width(), mixsurface->Height(), 0, GL_RGBA,
					GL_UNSIGNED_BYTE, (GLvoid *)mixsurface->GetBuffer());
			if ((gl_errno = glGetError()) != GL_NO_ERROR) {
				out_logf(LOG_ERROR, _T("glTexImage2D: 0x%x"), gl_errno);
			}

			// fill black
			glBegin(GL_QUADS);
				glColor3ub(0, 0, 0);
				glVertex3i(0, 0, 0);
				glVertex3i(0, display_height, 0);
				glVertex3i(display_width, display_height, 0);
				glVertex3i(display_width, 0, 0);
			glEnd();

			glEnable(GL_TEXTURE_2D);
			glColor3ub(255, 255, 255);
			glBegin(GL_QUADS);
				// anticlockwise
				glTexCoord2d(src_tex_l, src_tex_t);
				glVertex3i(reTex.x, reTex.y, 0);
				glTexCoord2d(src_tex_l, src_tex_b);
				glVertex3i(reTex.x, reTex.y + reTex.h, 0);
				glTexCoord2d(src_tex_r, src_tex_b);
				glVertex3i(reTex.x + reTex.w, reTex.y + reTex.h, 0);
				glTexCoord2d(src_tex_r, src_tex_t);
				glVertex3i(reTex.x + reTex.w, reTex.y, 0);
			glEnd();
			glDisable(GL_TEXTURE_2D);

			glBindTexture(GL_TEXTURE_2D, 0);

			if (!self_invalidate) {
				// call this only once at first.
				change_opengl_attr();
			}
		}
	} else
#endif
	{
#ifdef USE_SDL2
		screen = SDL_GetWindowSurface(window);
#endif
		// fill black on screen
		if (first_invalidate) {
			SDL_Rect re;
			re.x = 0; re.y = 0; re.w = screen->w; re.h = screen->h;
			SDL_FillRect(screen, &re, SDL_MapRGB(screen->format, 0, 0, 0));
			first_invalidate = false;
		}
		// render screen
		if(stretched_width == mixed_width && stretched_height == mixed_height) {
			SDL_UTILS::blit_surface(mixsurface->Get(), &reMix, screen, &reSuf);
		} else {
			SDL_UTILS::zoomin_surface(mixsurface->Get(), &reMix, screen, &reSuf);
		}
	}

	// invalidate window
	self_invalidate = true;
//	skip_frame = false;

	return true;
}

///
///
///
void EMU::need_update_screen()
{
}

///
/// return pointer on source screen
///
scrntype* EMU::screen_buffer(int y)
{
	return srcsurface->GetBuffer() + screen_width * y;
}

int EMU::screen_buffer_offset()
{
	return screen_width;
}

///
/// render screen and display window
/// @note must be called by main thread
void EMU::update_screen()
{
	lock_screen();

	if (mix_screen()) {

#ifdef USE_OPENGL
		if (screen_flags & SDL_WINDOW_OPENGL) {
			gui->MixSurface();
#ifndef USE_SDL2
			SDL_GL_SwapBuffers();
#else
			if (window) SDL_GL_SwapWindow(window);
#endif
		} else
#endif
		{
			// GuiMixScreen include SDL_UpdateRect process.
			if (!gui->MixSurface()) {
#ifndef USE_SDL2
//				SDL_UpdateRect(screen, 0, 0, display_width, display_height);
				SDL_UpdateRect(screen, 0, 0, 0, 0);
#else
				if (window) SDL_UpdateWindowSurfaceRects(window, NULL, 0);
#endif
			}
		}
	}
	unlock_screen();
}


bool EMU::now_skip_frame()
{
	return skip_frame && !now_rec_video();
}

///
/// capture current screen and save to a file
///
void EMU::capture_screen()
{
	int size = config.screen_video_size;
	rec_video->Capture(CAPTURE_SCREEN_TYPE, &rec_video_stretched_size, srcsurface, &rec_video_size[size]);
}

///
/// start recording video
///
bool EMU::start_rec_video(int type, int fps, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	int size = config.screen_video_size;
	return rec_video->Start(type, fps, &rec_video_size[size], srcsurface, show_dialog);
#else
	return false;
#endif
}

///
/// stop recording video
///
void EMU::stop_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Stop();
#endif
}

///
/// restart recording video
///
void EMU::restart_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Restart();
#endif
}

///
/// record video
///
void EMU::record_rec_video()
{
#ifdef USE_REC_VIDEO
	if (rec_video->IsRecordFrame()) {
		int size = config.screen_video_size;
		rec_video->Record(&rec_video_stretched_size, srcsurface, &rec_video_size[size]);
	}
#endif
}

bool EMU::now_rec_video()
{
#ifdef USE_REC_VIDEO
	return *now_recording_video;
#else
	return false;
#endif
}

void EMU::resize_rec_video(int num)
{
#ifdef USE_EMU_INHERENT_SPEC
	config.screen_video_size = num;
	change_rec_video_size(num);

	// send display size to vm
	set_vm_display_size();
#endif
}
void EMU::change_rec_video_size(int num)
{
	rec_video_stretched_size = rec_video_size[num];
	if (mixed_width_ratio < mixed_height_ratio) {
		rec_video_stretched_size.h = rec_video_stretched_size.h * mixed_width_ratio / mixed_height_ratio;
		rec_video_stretched_size.y += (rec_video_size[num].h - rec_video_stretched_size.h) / 2; 
	} else {
		rec_video_stretched_size.w = rec_video_stretched_size.w * mixed_height_ratio / mixed_width_ratio;
		rec_video_stretched_size.x += (rec_video_size[num].w - rec_video_stretched_size.w) / 2; 
	}
	rec_video_stretched_size.y = adjust_y_position(rec_video_stretched_size.h, rec_video_stretched_size.y);

#ifdef _DEBUG_LOG
	out_debugf(_T("          rec_video_size: num:%d x:%d y:%d w:%d h:%d")
		,num,rec_video_size[num].x,rec_video_size[num].y,rec_video_size[num].w,rec_video_size[num].h);
	out_debugf(_T("rec_video_stretched_size:       x:%d y:%d w:%d h:%d")
		,rec_video_stretched_size.x,rec_video_stretched_size.y,rec_video_stretched_size.w,rec_video_stretched_size.h);
#endif
}

bool EMU::rec_video_enabled(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->IsEnabled(type);
#else
	return false;
#endif
}

const _TCHAR **EMU::get_rec_video_codec_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetCodecList(type);
#else
	return NULL;
#endif
}

const _TCHAR **EMU::get_rec_video_quality_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetQualityList(type);
#else
	return NULL;
#endif
}

//----------

int EMU::qsort_by_w_h_n(const void *a, const void *b)
{
	struct window_mode_st *a1 = (struct window_mode_st *)a;
	struct window_mode_st *b1 = (struct window_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	if (!rc) rc = (a1->power - b1->power);
	return rc;
}

int EMU::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	int aa = a1->width * a1->height;
	int bb = b1->width * b1->height;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (bb - aa);
	return rc;
}

/// enumerate screen mode for window
/// calcurate magnify range
/// @param [in] max_width
/// @param [in] max_height
void EMU::enum_window_mode(int max_width, int max_height)
{
	// enumerate screen mode for window
	int w,h;
	bool minsize = false;
	bool maxsize = false;

	window_mode_count = 0;
	for (int i=0; window_proposal[i].cx != 0 && window_mode_count < WINDOW_MODE_MAX; i++) {
		w = window_proposal[i].cx;
		h = window_proposal[i].cy;
		// minimum window size?
		if (w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT) {
			continue;
		}
		if (!minsize) {
			w = MIN_WINDOW_WIDTH;
			h = MIN_WINDOW_HEIGHT;
			i--;
			minsize = true;
		}
		if (w > MAX_WINDOW_WIDTH || h > MAX_WINDOW_HEIGHT) {
			if (!maxsize) {
				w = MAX_WINDOW_WIDTH;
				h = MAX_WINDOW_HEIGHT;
				maxsize = true;
				i--;
			} else {
				break;
			}
		}
		for (int pow = 10; pow <= 30 && window_mode_count < WINDOW_MODE_MAX; pow+=5) {
			if ((w*pow/10) <= max_width && (h*pow/10) <= max_height) {
				bool found = false;
				for(int j = 0; j < window_mode_count; j++) {
					if(window_mode[j].power == pow && window_mode[j].width == (w*pow/10) && window_mode[j].height == (h*pow/10)) {
						found = true;
						break;
					}
				}
				if (!found) {
					window_mode[window_mode_count].power = pow;
					window_mode[window_mode_count].width = w * pow / 10;
					window_mode[window_mode_count].height = h * pow / 10;
					out_debugf(_T("window_mode:%d %dx%d x%.1f"),window_mode_count, w * pow / 10, h * pow / 10, (double)pow / 10.0);
					window_mode_count++;
				}
			} else {
				break;
			}
		}
	}
	qsort(&window_mode, window_mode_count, sizeof(struct window_mode_st), qsort_by_w_h_n);
}

void EMU::enum_screen_mode(uint32_t flags)
{
	// enumerate display device
	int max_desktop_width = 0;
	int max_desktop_height = 0;
	int disp_device_count = 1;

	// enumerate screen mode for fullscreen
	int w,h,bpp;
	screen_mode_count = 0;

#ifndef USE_SDL2
	SDL_Rect **list = NULL;

	flags |= SDL_FULLSCREEN;

	const SDL_VideoInfo *video = SDL_GetVideoInfo();
#else
	SDL_DisplayMode mode;
	disp_device_count = SDL_GetNumVideoDisplays();
#endif

	for(int disp_no = 0; disp_no < disp_device_count && screen_mode_count < SCREEN_MODE_MAX; disp_no++) {
#ifndef USE_SDL2
		list = SDL_ListModes(video->vfmt, flags);
		if (list == NULL || list == (SDL_Rect **)-1) break;
		for(int i = 0; list[i] != NULL && screen_mode_count < SCREEN_MODE_MAX; i++) {
			w = list[i]->w;
			h = list[i]->h;
			bpp = video->vfmt->BitsPerPixel;
#else
		int mode_count = SDL_GetNumDisplayModes(disp_no);
		for(int i = 0; i < mode_count && screen_mode_count < SCREEN_MODE_MAX; i++) {
			if (SDL_GetDisplayMode(disp_no, i, &mode) < 0) {
				continue;
			}
			w = mode.w;
			h = mode.h;
			bpp = SDL_BITSPERPIXEL(mode.format);
#endif
			if(w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT || w < 640 || h < 480) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if(((w * 30 / h) != 40)
			&& ((w * 40 / h) != 50)
			&& ((w * 90 / h) != 160)
			&& ((w * 50 / h) != 80)) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if (w <= 1280 && (w != 640 && w != 800 && w != 1024 && w != 1280)) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			bool found = false;
			for(int j = 0; j < screen_mode_count; j++) {
				if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == w && screen_mode[j].height == h) {
					found = true;
					out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp already exist"), disp_no, w, h, bpp);
					break;
				}
			}
			if(!found) {
				screen_mode[screen_mode_count].disp_no = disp_no;
				screen_mode[screen_mode_count].width = w;
				screen_mode[screen_mode_count].height = h;
				out_debugf(_T("screen_mode:%2d [%d] %dx%d %dbpp"), screen_mode_count, disp_no, w, h, bpp);
				screen_mode_count++;

				if(max_desktop_width <= w) {
					max_desktop_width = w;
				}
				if(max_desktop_height <= h) {
					max_desktop_height = h;
				}
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);

	// if cannot get modes, add default screen size mode.
	if (screen_mode_count == 0) {
		screen_mode[screen_mode_count].disp_no = 0;
		screen_mode[screen_mode_count].width = desktop_width;
		screen_mode[screen_mode_count].height = desktop_height;
		screen_mode_count = 1;

		max_desktop_width = desktop_width;
		max_desktop_height = desktop_height;
	}
}

/// matching window or fullscreen size
void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		// matching window size at last time
		bool find = false;
		int i = config.window_mode;
		if (i < window_mode_count) {
			if (window_mode[i].width == config.screen_width
			&&  window_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	} else {
		// matching fullscreen size at last time
		bool find = false;
		for(int i = 0; i < screen_mode_count; i++) {
			if (screen_mode[i].disp_no == config.disp_device_no
			&&  screen_mode[i].width == config.screen_width
			&&  screen_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i + 8;
				break;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	}
}

/// change window size / switch over fullscreen and window
/// @param[in] mode 0 - 7: window size  8 -:  fullscreen size  -1: switch over  -2: shift window mode
void EMU::change_screen_mode(int mode)
{
//	out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
//	if (mode == config.window_mode) return;

	if (mode == -1) {
		// switch over fullscreen and window
		if (now_fullscreen) {
			// go window mode
			mode = prev_window_mode;
		}
	} else if (mode == -2) {
		// shift window mode 
		if (now_fullscreen) {
			// no change
			return;
		} else {
			mode = ((config.window_mode + 1) % window_mode_count);
		}
	}
	if (!now_fullscreen) {
		prev_window_mode = config.window_mode;
	}
//	out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
	set_window(mode, desktop_width, desktop_height);
	if (!create_screen(config.screen_width, config.screen_height, screen_flags)) {
		exit(1);
	}

}

void EMU::change_stretch_screen(int num)
{
	if (num < 0) {
		num = (config.stretch_screen + 1);
	}
	num = (num % 3);
	if (config.stretch_screen == num) {
		num = 0;
	}
	config.stretch_screen = num;
	if (now_fullscreen) {
		set_display_size(-1, -1, 10, !now_fullscreen);
	}
}

#if 0
void EMU::change_cutout_screen()
{
	config.cutout_screen = !config.cutout_screen;
	if (now_fullscreen) set_display_size(-1, -1, 10, !now_fullscreen);
}
#endif

void EMU::change_pixel_aspect(int mode)
{
	if (mode < 0) {
		mode = (config.pixel_aspect + 1) % get_pixel_aspect_count();
	}
	config.pixel_aspect = get_pixel_aspect(mode, &mixed_width_ratio, &mixed_height_ratio);

	set_display_size(config.screen_width, config.screen_height, now_fullscreen ? 10 : window_mode_power, !now_fullscreen);
}

int EMU::get_pixel_aspect_count()
{
	return 3;
}

int EMU::get_pixel_aspect(int mode, int *wratio, int *hratio)
{
	int new_mode = mode;
	switch(mode) {
	case 1:
		*wratio = 10;
		*hratio = 11;
		break;
	case 2:
		*wratio = 10;
		*hratio = 12;
		break;
	default:
		*wratio = 1;
		*hratio = 1;
		new_mode = 0;
		break;
	}
	return new_mode;
}

/// setting window or fullscreen size to display
/// @param [in] mode 0 .. 7 window mode / 8 .. 23 fullscreen mode / -1 want to go fullscreen, but unknown mode
/// @param [in] cur_width  current desktop width if mode is -1
/// @param [in] cur_height current desktop height if mode is -1
void EMU::set_window(int mode, int cur_width, int cur_height)
{
	out_debugf(_T("set_window: mode:%d"), mode);
	if(mode >= 0 && mode < 8) {
		// go window
		if (mode >= window_mode_count) mode = 0;
		int width = window_mode[mode].width;
		int height = window_mode[mode].height;

		config.window_mode = mode;
		config.disp_device_no = 0;
		config.screen_width = width;
		config.screen_height = height;
		window_mode_power = window_mode[mode].power;

		now_fullscreen = false;

		gui->ScreenModeChanged(now_fullscreen);

		out_debugf(_T("set_window: f->w mode:%d w:%d h:%d nf:%d"), mode, config.screen_width,screen_height,now_fullscreen ? 1 : 0);

		// set screen size to emu class
		set_display_size(width, height, window_mode_power, true);
	}
	else if(!now_fullscreen) {
		// go fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int disp_no = (mode == -1) ? 0 : screen_mode[mode - 8].disp_no;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;

		if(mode == -1) {
			// matching width and height
			for(int i = 0; i < screen_mode_count; i++) {
				if(screen_mode[i].disp_no == disp_no && screen_mode[i].width == width && screen_mode[i].height == height) {
					mode = i + 8;
					break;
				}
			}
		}
		config.window_mode = mode;
		config.disp_device_no = disp_no;
		config.screen_width = width;
		config.screen_height = height;

		now_fullscreen = true;

		gui->ScreenModeChanged(now_fullscreen);

		out_debugf(_T("set_window: w->f mode:%d w:%d h:%d nf:%d"), mode, config.screen_width,screen_height,now_fullscreen ? 1 : 0);

		// set screen size to emu class
		set_display_size(width, height, 10, false);
	} else {
		// now fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;

		out_debugf(_T("set_window: f->f mode:%d w:%d h:%d nf:%d"), mode, config.screen_width,screen_height,now_fullscreen ? 1 : 0);

		// set screen size to emu class
		set_display_size(width, height, 10, false);
	}
}

window_mode_t *EMU::get_window_mode(int num)
{
	if (num < window_mode_count) {
		return &window_mode[num];
	}
	return NULL;
}

screen_mode_t *EMU::get_screen_mode(int num)
{
	if (num < screen_mode_count) {
		return &screen_mode[num];
	}
	return NULL;
}

VmRectWH *EMU::get_screen_record_size(int num)
{
	if (0 <= num && num <= 1) {
		return &rec_video_size[num];
	}
	return NULL;
}

// ----------
#ifdef USE_OPENGL

void EMU::initialize_opengl()
{
	if (!use_opengl) {
		return;
	}
//	create_texture();
}

void EMU::create_texture()
{
	GLenum gl_errno;

#ifdef USE_SDL2
	if (!window) return;
	if (SDL_GL_SetSwapInterval(use_opengl == 1 ? 1 : 0) < 0) {
		out_logf(LOG_WARN, _T("SDL_GL_SetSwapInterval: %s."), SDL_GetError());
	}
#endif
	if (mix_texture_name != 0) return;

	// create
#if defined(_RGB555) || defined(_RGB565)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
#elif defined(_RGB888)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
#else
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
#endif

	glGenTextures(1, &mix_texture_name);
	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
		out_logf(LOG_ERROR, _T("glGenTextures: %d"), gl_errno);
	} else {
		out_debugf(_T("glGenTextures: name:%d"), mix_texture_name);
	}
	glBindTexture(GL_TEXTURE_2D, mix_texture_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_S, GL_CLAMP);	// no wrap texture
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, SCREEN_WIDTH, SCREEN_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
//	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
//		out_logf(LOG_ERROR, "glTexImage2D: %d", gl_errno);
//	}
	glBindTexture(GL_TEXTURE_2D, 0);	// unbound

	// fill black
	glBegin(GL_QUADS);
	glColor3ub(0, 0, 0);
	glVertex3i(0, 0, 0);
	glVertex3i(0, display_height, 0);
	glVertex3i(display_width, display_height, 0);
	glVertex3i(display_width, 0, 0);
	glEnd();
#ifndef USE_SDL2
	SDL_GL_SwapBuffers();
#else
	SDL_GL_SwapWindow(window);
#endif
}

void EMU::change_opengl_attr()
{
	if (!use_opengl) {
		return;
	}

	if (mix_texture_name != 0 && glIsTexture(mix_texture_name) == GL_TRUE) {
		glBindTexture(GL_TEXTURE_2D, mix_texture_name);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
		glBindTexture(GL_TEXTURE_2D, 0);	// unbound
	}
}

void EMU::release_opengl()
{
	if (!use_opengl) {
		return;
	}
	release_texture();
#ifdef USE_SDL2
	if (glcontext) {
		SDL_GL_DeleteContext(glcontext);
	}
#endif
}

void EMU::release_texture()
{
	if (mix_texture_name != 0 && glIsTexture(mix_texture_name) == GL_TRUE) {
		// release texture
		glDeleteTextures(1, &mix_texture_name);
	}
	mix_texture_name = 0;
}

void EMU::set_opengl_attr()
{
	// set OpenGL atributes
#if defined(_RGB555)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 16 );
#elif defined(_RGB565)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 6 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 16 );
#elif defined(_RGB888)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 32 );
#endif
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
//		SDL_GL_SetAttribute( SDL_GL_ACCELERATED_VISUAL, 0 );
#ifdef __WIN32__
	// if windows, multisample must be spcify
	SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 0 );
	SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, 1 );
#endif
	if (use_opengl) {
#ifndef USE_SDL2
		if (SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, use_opengl == 1 ? 1 : 0 ) < 0) {
			out_logf(LOG_WARN, _T("SDL_GL_SWAP_CONTROL: %s."), SDL_GetError());
		}
#endif
	}
}

void EMU::set_opengl_poly(int width, int height)
{
#ifdef _DEBUG_LOG
	int v;
	SDL_GL_GetAttribute(SDL_GL_RED_SIZE, &v); out_debugf(_T("SDL_GL_RED_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE, &v); out_debugf(_T("SDL_GL_GREEN_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE, &v); out_debugf(_T("SDL_GL_BLUE_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_ALPHA_SIZE, &v); out_debugf(_T("SDL_GL_ALPHA_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &v); out_debugf(_T("SDL_GL_DEPTH_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_BUFFER_SIZE, &v); out_debugf(_T("SDL_GL_BUFFER_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER, &v); out_debugf(_T("SDL_GL_DOUBLEBUFFER:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS, &v); out_debugf(_T("SDL_GL_MULTISAMPLEBUFFERS:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES, &v); out_debugf(_T("SDL_GL_MULTISAMPLESAMPLES:%d"), v);
#ifndef USE_SDL2
	SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &v); out_logf(LOG_DEBUG, _T("SDL_GL_SWAP_CONTROL:%d"), v);
#else
	out_logf(LOG_DEBUG, _T("SDL_GL_GetSwapInterval:%d"), SDL_GL_GetSwapInterval());
#endif
#endif

	// shading
	glShadeModel( GL_SMOOTH );
	// culling
	glFrontFace( GL_CCW );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );
	// clear bg color
	glClearColor( 0.0, 0.0, 0.0, 0.0 );

    // Setup our viewport : same as window size
    glViewport( 0, 0, width, height );

	glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

	// change axis unit
	glOrtho(0.0, (GLdouble)width, (GLdouble)height, 0.0, -1.0, 1.0);

}

#endif // USE_OPENGL

inline void EMU::lock_screen()
{
	mux_update_screen->lock();
	screen_changing = true;
}

inline void EMU::unlock_screen()
{
	screen_changing = false;
	mux_update_screen->unlock();
}

