/** @file sdl_sound.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl sound ]

	@note
	This code is based on win32_sound.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "video/rec_audio.h"

//#define DSOUND_BUFFER_SIZE (uint32_t)(sound_samples * 8)
//#define DSOUND_BUFFER_HALF (uint32_t)(sound_samples * 4)

#define SOUND_RECORD_IN_MIXER

void EMU::EMU_SOUND()
{
	sound_rate = 0;
	sound_samples = 0;
	sound_latency = 0;
	sound_ok = false;
	now_mute = false;
	sound_prev_time = 0;
	memset(&audio_desired,0,sizeof(audio_desired));

	rec_audio = new REC_AUDIO(this);
}

void EMU::initialize_sound(int rate, int samples, int latency)
{
	sound_rate = rate;
	sound_samples = samples;
	sound_latency = latency;
	vm->initialize_sound(sound_rate, sound_samples);

#if 0 /* USE_SDL2 */
	const int drvs = SDL_GetNumAudioDevices(0);
	for(int i=0; i <drvs; i++) {
		out_debugf(_T("audio_device %d %s"),i,SDL_GetAudioDeviceName(i, 0));
	}
#endif

	memset(&audio_desired,0,sizeof(audio_desired));
	audio_desired.freq = rate;
#ifdef USE_AUDIO_U8
	audio_desired.format = AUDIO_U8;
#else
	audio_desired.format = AUDIO_S16;
#endif
	audio_desired.channels = 2;
	audio_desired.samples = (samples >> 1);
	audio_desired.callback = &update_sound;
	audio_desired.userdata = (void *)this;

	if (SDL_OpenAudio(&audio_desired,&audio_obtained) != 0) {
		out_logf(LOG_ERROR,_T("SDL_OpenAudio: %s"),SDL_GetError());
	}
#ifdef _DEBUG_LOG
	out_debugf(_T("audio_obtained.freq:%d"),audio_obtained.freq);
	out_debugf(_T("audio_obtained.format:%04x"),audio_obtained.format);
	out_debugf(_T("audio_obtained.channels:%d"),audio_obtained.channels);
	out_debugf(_T("audio_obtained.samples:%d"),audio_obtained.samples);
	out_debugf(_T("audio_obtained.silence:%d"),audio_obtained.silence);
	out_debugf(_T("audio_obtained.size:%d"),audio_obtained.size);
#endif

	out_logf(LOG_DEBUG, _T("sound ok: rate:%d samples:%d latency:%d"), rate, samples, latency);
	sound_ok = true;

	sound_prev_time = SDL_GetTicks();
}

//void EMU::initialize_sound(int rate, int samples)
//{
//	initialize_sound(rate, samples, 100);
//}

void EMU::initialize_sound()
{
	// load sound config
#ifdef SUPPORT_SOUND_FREQ_55467HZ
	// PC-8801/9801 series
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 55467, 96000};
#else
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 48000, 96000};	// Hz
#endif
	static int late_table[5] = {50, 100, 200, 300, 400};	// ms

	if(!(0 <= config.sound_frequency && config.sound_frequency < 8)) {
		config.sound_frequency = 6;	// default: 48KHz
	}
	if(!(0 <= config.sound_latency && config.sound_latency < 5)) {
		config.sound_latency = 1;	// default: 100msec
	}
	int frequency = freq_table[config.sound_frequency];
	int samples = (int)((double)(frequency * late_table[config.sound_latency]) / 1000.0 + 0.5);
	int latency = late_table[config.sound_latency];

	initialize_sound(frequency, samples, latency);
}

void EMU::release_sound()
{
	SDL_CloseAudio();

	delete rec_audio;
}

void EMU::release_sound_on_emu_thread()
{
	end_sound();
}

void EMU::start_sound()
{
	SDL_PauseAudio(0);

	sound_ok = true;
}

void EMU::end_sound()
{
	SDL_PauseAudio(1);
	sound_ok = false;

	// stop recording
	stop_rec_sound();
}

/// @attention this function is called by another thread.
/// event callback so static function
void EMU::update_sound(void *userdata, Uint8 *stream, int len)
{
	EMU *emu = (EMU *)userdata;

//	emu->out_debugf("EMU::update_sound start: len:%d stream:%llx",len,stream);

//	if (emu->sound_prev_time + emu->sound_latency > sound_now_time) {
//		return;
//	}

	int extra_frames = 0;

	if(emu->sound_ok) {
		// sound buffer must be updated
#ifdef USE_AUDIO_U8
		uint8_t* sound_buffer = emu->vm->create_sound(&extra_frames, emu->audio_obtained.samples);
#else
		int16_t* sound_buffer = emu->vm->create_sound(&extra_frames, emu->audio_obtained.samples);
#endif
#ifndef SOUND_RECORD_IN_MIXER
		record_sound(sound_buffer, sound_samples);
#endif
		if(sound_buffer) {
#ifndef USE_SDL2
			SDL_MixAudio(stream, (Uint8 *)sound_buffer, len, SDL_MIX_MAXVOLUME);
#else
			// TODO: should be SDL_MixAudioFormat
			SDL_memcpy(stream, (Uint8 *)sound_buffer, len);
#endif
		}
	}

//	emu->out_debug("EMU::update_sound end");
}

void EMU::mute_sound(bool mute)
{
	if (mute) {
// sound is continuous driving during mute
//		SDL_PauseAudio(1);
		now_mute = true;
	} else {
		if (sound_ok && !vm_pause){
//			SDL_PauseAudio(0);
			now_mute = false;
		}
	}
}

void EMU::set_volume(int volume)
{
	// set volume for sound devices.
	if (vm) {
		vm->set_volume();
	}
}


void EMU::start_rec_sound(int type)
{
#ifdef USE_REC_AUDIO
	rec_audio->Start(type, sound_rate, false);
#endif
}

void EMU::stop_rec_sound()
{
#ifdef USE_REC_AUDIO
	rec_audio->Stop();
#endif
}

void EMU::restart_rec_sound()
{
#ifdef USE_REC_AUDIO
	rec_audio->Restart();
#endif
}

#if 0
void EMU::record_rec_sound(uint8_t *buffer, int samples)
{
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
}

void EMU::record_rec_sound(int16_t *buffer, int samples)
{
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
}
#endif

void EMU::record_rec_sound(int32_t *buffer, int samples)
{
#ifdef SOUND_RECORD_IN_MIXER
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
#endif
}

bool EMU::now_rec_sound()
{
	return rec_audio->NowRecording();
}

const _TCHAR **EMU::get_rec_sound_codec_list(int type)
{
#ifdef USE_REC_AUDIO
	return rec_audio->GetCodecList(type);
#else
	return NULL;
#endif
}

bool *EMU::get_now_rec_sound_ptr()
{
	return rec_audio->GetNowRecordingPtr();
}

bool EMU::rec_sound_enabled(int type)
{
#ifdef USE_REC_AUDIO
	return rec_audio->IsEnabled(type);
#else
	return false;
#endif
}

void EMU::lock_sound_buffer()
{
	SDL_LockAudio();
}
void EMU::unlock_sound_buffer()
{
	SDL_UnlockAudio();
}
