/** @file sdl_utils.cpp

 Skelton for retropc emulator
 SDL edition

 @author Sasaji
 @date   2015.05.20

 @brief [ sdl utilities ]

 */

#include "sdl_utils.h"

#define CP_RECT(src, dst) { \
	(dst).x = (src).x; \
	(dst).y = (src).y; \
	(dst).w = (src).w; \
	(dst).h = (src).h; \
}

namespace SDL_UTILS
{

/// blit surface
int blit_surface(SDL_Surface *src, VmRectWH *srcrect, SDL_Surface *dst, VmRectWH *dstrect)
{
	SDL_Rect s_srcrect,s_dstrect;
	CP_RECT(*srcrect, s_srcrect);
	CP_RECT(*dstrect, s_dstrect);
	return SDL_BlitSurface(src, &s_srcrect, dst, &s_dstrect);
}

/// copy surface
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void copy_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;
	Uint8 r, g, b, a;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	for(int sy=reSrc->y, dy = reDst->y; sy < sbottom && dy < dbottom; sy++, dy++) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; sx++, dx++) {
			r = ((pSrc[sx] & src->format->Rmask) >> (src->format->Rshift)) & 0xff;
			g = ((pSrc[sx] & src->format->Gmask) >> (src->format->Gshift)) & 0xff;
			b = ((pSrc[sx] & src->format->Bmask) >> (src->format->Bshift)) & 0xff;
			a = ((pSrc[sx] & src->format->Amask) >> (src->format->Ashift)) & 0xff;
			pDst[dx] =
			((r << dst->format->Rshift) & dst->format->Rmask)
			| ((g << dst->format->Gshift) & dst->format->Gmask)
			| ((b << dst->format->Bshift) & dst->format->Bmask)
			| ((a << dst->format->Ashift) & dst->format->Amask);
		}
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

/// copy surface
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void copy_surface(SDL_Surface *src, VmRectWH *reSrc, SDL_Surface *dst, VmRectWH *reDst)
{
	SDL_Rect s_reSrc, s_reDst;
	CP_RECT(*reSrc, s_reSrc);
	CP_RECT(*reDst, s_reDst);
#ifndef USE_SDL2
	copy_surface(src, &s_reSrc, dst, &s_reDst);
#else
	SDL_BlitSurface(src, &s_reSrc, dst, &s_reDst);
#endif
}

/// copy surface (flipped top and bottom)
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void copy_surface_flipped(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;
	Uint8 r, g, b, a;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	for(int sy=reSrc->y, dy = dbottom - 1; sy < sbottom && dy >= reDst->y ; sy++, dy--) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; sx++, dx++) {
			r = ((pSrc[sx] & src->format->Rmask) >> (src->format->Rshift)) & 0xff;
			g = ((pSrc[sx] & src->format->Gmask) >> (src->format->Gshift)) & 0xff;
			b = ((pSrc[sx] & src->format->Bmask) >> (src->format->Bshift)) & 0xff;
			a = ((pSrc[sx] & src->format->Amask) >> (src->format->Ashift)) & 0xff;
			pDst[dx] =
			((r << dst->format->Rshift) & dst->format->Rmask)
			| ((g << dst->format->Gshift) & dst->format->Gmask)
			| ((b << dst->format->Bshift) & dst->format->Bmask)
			| ((a << dst->format->Ashift) & dst->format->Amask);
		}
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

/// copy surface (flipped top and bottom)
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void copy_surface_flipped(SDL_Surface *src, VmRectWH *reSrc, SDL_Surface *dst, VmRectWH *reDst)
{
	SDL_Rect s_reSrc, s_reDst;
	CP_RECT(*reSrc, s_reSrc);
	CP_RECT(*reDst, s_reDst);
	copy_surface_flipped(src, &s_reSrc, dst, &s_reDst);
}

/// zoom in surface
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void zoomin_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;
	Uint8 r, g, b, a;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);
	int x_mag = (reSrc->w << 12) / reDst->w;
	int y_mag = (reSrc->h << 12) / reDst->h;
	int sx_m = 0;
	int sy_m = 0;

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	sy_m = (reSrc->y << 12);
	for(int sy=reSrc->y, dy = reDst->y; sy < sbottom && dy < dbottom; dy++) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		sx_m = (reSrc->x << 12);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; dx++) {
			r = ((pSrc[sx] & src->format->Rmask) >> (src->format->Rshift)) & 0xff;
			g = ((pSrc[sx] & src->format->Gmask) >> (src->format->Gshift)) & 0xff;
			b = ((pSrc[sx] & src->format->Bmask) >> (src->format->Bshift)) & 0xff;
			a = ((pSrc[sx] & src->format->Amask) >> (src->format->Ashift)) & 0xff;
			pDst[dx] =
			((r << dst->format->Rshift) & dst->format->Rmask)
			| ((g << dst->format->Gshift) & dst->format->Gmask)
			| ((b << dst->format->Bshift) & dst->format->Bmask)
			| ((a << dst->format->Ashift) & dst->format->Amask);
			sx_m+=x_mag;
			sx = (sx_m >> 12);
		}
		sy_m+=y_mag;
		sy = (sy_m >> 12);
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

/// zoom in surface
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void zoomin_surface(SDL_Surface *src, VmRectWH *reSrc, SDL_Surface *dst, VmRectWH *reDst)
{
	SDL_Rect s_reSrc, s_reDst;
	CP_RECT(*reSrc, s_reSrc);
	CP_RECT(*reDst, s_reDst);
#ifndef USE_SDL2
	zoomin_surface(src, &s_reSrc, dst, &s_reDst);
#else
	SDL_BlitScaled(src, &s_reSrc, dst, &s_reDst);
#endif
}

/// zoom in surface (flipped top and bottom)
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void zoomin_surface_flipped(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;
	Uint8 r, g, b, a;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);
	int x_mag = (reSrc->w << 12) / reDst->w;
	int y_mag = (reSrc->h << 12) / reDst->h;
	int sx_m = 0;
	int sy_m = 0;

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	sy_m = (reSrc->y << 12);
	for(int sy=reSrc->y, dy = dbottom - 1; sy < sbottom && dy >= reDst->y; dy--) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		sx_m = (reSrc->x << 12);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; dx++) {
			r = ((pSrc[sx] & src->format->Rmask) >> (src->format->Rshift)) & 0xff;
			g = ((pSrc[sx] & src->format->Gmask) >> (src->format->Gshift)) & 0xff;
			b = ((pSrc[sx] & src->format->Bmask) >> (src->format->Bshift)) & 0xff;
			a = ((pSrc[sx] & src->format->Amask) >> (src->format->Ashift)) & 0xff;
			pDst[dx] =
			((r << dst->format->Rshift) & dst->format->Rmask)
			| ((g << dst->format->Gshift) & dst->format->Gmask)
			| ((b << dst->format->Bshift) & dst->format->Bmask)
			| ((a << dst->format->Ashift) & dst->format->Amask);
			sx_m+=x_mag;
			sx = (sx_m >> 12);
		}
		sy_m+=y_mag;
		sy = (sy_m >> 12);
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

/// zoom in surface (flipped top and bottom)
/// @param[in] src
/// @param[in] reSrc : area to copy from src (cannot null)
/// @param[out] dst
/// @param[in] reDst : area to copy on dst (cannot null)
/// @note no error check. so you must specify valid rect.
void zoomin_surface_flipped(SDL_Surface *src, VmRectWH *reSrc, SDL_Surface *dst, VmRectWH *reDst)
{
	SDL_Rect s_reSrc, s_reDst;
	CP_RECT(*reSrc, s_reSrc);
	CP_RECT(*reDst, s_reDst);
	zoomin_surface_flipped(src, &s_reSrc, dst, &s_reDst);
}

/// fill data on surface
/// @param[in,out] src
/// @param[in] data : pixel data to fill
/// @param[in] mask : mask data
/// @note no error check. so you must specify valid rect.
void fill_surface(SDL_Surface *src, scrntype data, scrntype mask)
{
	scrntype *pSrc;
	SDL_LockSurface(src);
	for(int sy=0; sy < src->h; sy++) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		for(int sx=0; sx < src->w; sx++) {
			*pSrc = ((data & mask) | (*pSrc & ~mask));
			pSrc++;
		}
	}
	SDL_UnlockSurface(src);
}

/// create SDL surface
/// @param[in] srcrect surface size (width, height)
/// @param[in] srcformat refer to pixel format
/// @param[in] force_format 0:same as srcformat 1:for OpenGL(RGBA) 2:BGRA 3:ARGB 4:ABGR
/// @return new surface
SDL_Surface *create_surface(SDL_Rect *srcrect, SDL_PixelFormat *srcformat, int force_format)
{
	SDL_PixelFormat format = *srcformat;
	switch(force_format) {
	case 1:
		set_pixel_format_for_opengl(&format);
		break;
	case 2:
		set_pixel_format_bgra(&format);
		break;
	case 3:
		set_pixel_format_argb(&format);
		break;
	case 4:
		set_pixel_format_rgba(&format);
		break;
	}

	SDL_Surface *surface = SDL_CreateRGBSurface(SDL_SWSURFACE, srcrect->w, srcrect->h
	  , format.BitsPerPixel, format.Rmask, format.Gmask, format.Bmask, format.Amask);
	return surface;
}

/// create SDL surface
/// @param[in] srcrect surface size (width, height)
/// @param[in] srcformat refer to pixel format
/// @param[in] force_format 0:same as srcformat 1:for OpenGL(RGBA) 2:BGRA 3:ARGB 4:ABGR
/// @return new surface
SDL_Surface *create_surface(VmRectWH *srcrect, SDL_PixelFormat *srcformat, int force_format)
{
	SDL_Rect s_srcrect;
	CP_RECT(*srcrect, s_srcrect);
	return create_surface(&s_srcrect, srcformat, force_format);
}

/// create SDL surface
/// @param[in] srcrect surface size (width, height)
/// @param[in] srcsurface refer to pixel format on this
/// @param[in] force_format 0:same as srcsurface 1:for OpenGL(RGBA) 2:BGRA 3:ARGB 4:ABGR
/// @return new surface
SDL_Surface *create_surface(SDL_Rect *srcrect, SDL_Surface *srcsurface, int force_format)
{
	SDL_PixelFormat format = *srcsurface->format;
	switch(force_format) {
	case 1:
		set_pixel_format_for_opengl(&format);
		break;
	case 2:
		set_pixel_format_bgra(&format);
		break;
	case 3:
		set_pixel_format_argb(&format);
		break;
	case 4:
		set_pixel_format_rgba(&format);
		break;
	}

	SDL_Surface *surface = SDL_CreateRGBSurface(SDL_SWSURFACE, srcrect->w, srcrect->h
	  , format.BitsPerPixel, format.Rmask, format.Gmask, format.Bmask, format.Amask);
	return surface;
}

/// create SDL surface
/// @param[in] srcrect surface size (width, height)
/// @param[in] srcsurface refer to pixel format on this
/// @param[in] force_format 0:same as srcsurface 1:for OpenGL(RGBA) 2:BGRA 3:ARGB 4:ABGR
/// @return new surface
SDL_Surface *create_surface(VmRectWH *srcrect, SDL_Surface *srcsurface, int force_format)
{
	SDL_Rect s_srcrect;
	CP_RECT(*srcrect, s_srcrect);
	return create_surface(&s_srcrect, srcsurface, force_format);
}

/// create SDL surface
/// @param[in] width
/// @param[in] height
/// @param[in] pixel_format
/// @return new surface
SDL_Surface *create_surface(int width, int height, SDL_PixelFormat *pixel_format)
{
	SDL_Surface *surface = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height
	  , pixel_format->BitsPerPixel, pixel_format->Rmask, pixel_format->Gmask, pixel_format->Bmask, pixel_format->Amask);
	return surface;
}

SDL_Surface *release_surface(SDL_Surface *surface)
{
	if (surface) {
		SDL_FreeSurface(surface);
	}
	return NULL;
}

void set_pixel_format_for_opengl(SDL_PixelFormat *pixel_format)
{
	pixel_format->palette = NULL;
	if (pixel_format->BytesPerPixel == 4) {
		// OpenGL RGBA masks
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x000000FF;
		pixel_format->Gmask = 0x0000FF00;
		pixel_format->Bmask = 0x00FF0000;
		pixel_format->Amask = 0xFF000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 8;
		pixel_format->Bshift = 16;
		pixel_format->Ashift = 24;
#else
		pixel_format->Rmask = 0xFF000000;
		pixel_format->Gmask = 0x00FF0000;
		pixel_format->Bmask = 0x0000FF00;
		pixel_format->Amask = 0x000000FF;
		pixel_format->Rshift = 24;
		pixel_format->Gshift = 16;
		pixel_format->Bshift = 8;
		pixel_format->Ashift = 0;
#endif
	} else if (pixel_format->BytesPerPixel == 2) {
		// OpenGL RGB masks ?
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x0000001F;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000F800;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 11;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x0000F800;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000001F;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 11;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 0;
#endif
	}
}

void set_pixel_format_bgra(SDL_PixelFormat *pixel_format)
{
	pixel_format->palette = NULL;
	if (pixel_format->BytesPerPixel == 4) {
		// BGRA
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x00FF0000;
		pixel_format->Gmask = 0x0000FF00;
		pixel_format->Bmask = 0x000000FF;
		pixel_format->Amask = 0xFF000000;
		pixel_format->Rshift = 16;
		pixel_format->Gshift = 8;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 24;
#else
		pixel_format->Rmask = 0x0000FF00;
		pixel_format->Gmask = 0x00FF0000;
		pixel_format->Bmask = 0xFF000000;
		pixel_format->Amask = 0x000000FF;
		pixel_format->Rshift = 8;
		pixel_format->Gshift = 16;
		pixel_format->Bshift = 24;
		pixel_format->Ashift = 0;
#endif
	} else if (pixel_format->BytesPerPixel == 2) {
		// BGR ?
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x0000F800;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000001F;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 11;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x0000001F;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000F800;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 11;
		pixel_format->Ashift = 0;
#endif
	}
}

void set_pixel_format_argb(SDL_PixelFormat *pixel_format)
{
	pixel_format->palette = NULL;
	if (pixel_format->BytesPerPixel == 4) {
		// ARGB
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x0000FF00;
		pixel_format->Gmask = 0x00FF0000;
		pixel_format->Bmask = 0xFF000000;
		pixel_format->Amask = 0x000000FF;
		pixel_format->Rshift = 8;
		pixel_format->Gshift = 16;
		pixel_format->Bshift = 24;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x00FF0000;
		pixel_format->Gmask = 0x0000FF00;
		pixel_format->Bmask = 0x000000FF;
		pixel_format->Amask = 0xFF000000;
		pixel_format->Rshift = 16;
		pixel_format->Gshift = 8;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 24;
#endif
	} else if (pixel_format->BytesPerPixel == 2) {
		// RGB ?
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x0000001F;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000F800;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 11;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x0000F800;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000001F;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 11;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 0;
#endif
	}
}

void set_pixel_format_rgba(SDL_PixelFormat *pixel_format)
{
	pixel_format->palette = NULL;
	if (pixel_format->BytesPerPixel == 4) {
		// ABGR
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0xFF000000;
		pixel_format->Gmask = 0x00FF0000;
		pixel_format->Bmask = 0x0000FF00;
		pixel_format->Amask = 0x000000FF;
		pixel_format->Rshift = 24;
		pixel_format->Gshift = 16;
		pixel_format->Bshift = 8;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x000000FF;
		pixel_format->Gmask = 0x0000FF00;
		pixel_format->Bmask = 0x00FF0000;
		pixel_format->Amask = 0xFF000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 8;
		pixel_format->Bshift = 16;
		pixel_format->Ashift = 24;
#endif
	} else if (pixel_format->BytesPerPixel == 2) {
		// RGB ?
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format->Rmask = 0x0000F800;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000001F;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 11;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 0;
		pixel_format->Ashift = 0;
#else
		pixel_format->Rmask = 0x0000001F;
		pixel_format->Gmask = 0x000007E0;
		pixel_format->Bmask = 0x0000F800;
		pixel_format->Amask = 0x00000000;
		pixel_format->Rshift = 0;
		pixel_format->Gshift = 5;
		pixel_format->Bshift = 11;
		pixel_format->Ashift = 0;
#endif
	}
}
}; /* SDL_UTILS */
