/** @file cmt.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.13 -

	@brief [ cmt ]
*/

#include <math.h>
#include "cmt.h"

#define TAPE_WAV_FREQ	4800

#define T9X_IDENTIFIER "eMB-689X CassetteTapeImageFile  "

static const _TCHAR *support_ext[6]={_T(".wav"), _T(".l3c"), _T(".l3b"), _T(".l3"), _T(".t9x"), NULL};

void CMT::initialize()
{
	// data recorder
	fio = new FILEIO();
	wav = new PARSEWAV();
	memset(data_buffer, 0x00, sizeof(data_buffer));
	memset(c_onedata, 0x00, sizeof(c_onedata));
	data_read_pos = 0;
	data_write_pos = 0;
	data_read_size = 0;
	data_write_size = 0;
	data_total_pos = 0;
	file_type = 3;
	is_bytedata = true;
	need_header = false;
	data_len = 0;

	t9x_data = 0;
	t9x_dlen = 0;

	buffer_overflow = false;
	play = rec = false;
	remote = false;
	remote_prev = false;
	send_ok = true;
	recv_ok = false;

	baud_sel = 0;

	register_id[0] = -1;
	register_id[1] = -1;

	relay_wav_enable = 0;
	relay_wav_loaded_at_first = false;

	relay_data[0] = NULL;
	relay_size[0] = 0;
	relay_data[1] = NULL;
	relay_size[1] = 0;
	relay_play_pos = 0;
	relay_play = 0;
	_tcscpy(relay_file[0], _T("relay_on.wav"));
	_tcscpy(relay_file[1], _T("relayoff.wav"));

	tape_wave_buf = NULL;
	tape_wave_buf_wpos = 0;
	tape_wave_buf_rpos = 0;

	set_baud(baud_sel);
}

void CMT::reset()
{
	load_relay_wav();

	baud_sel = 0;
	set_baud(baud_sel);
	for(int i=0; i<2; i++) {
		if (register_id[i] != -1) {
			cancel_my_event(i);
			register_my_event(i);
		}
	}
	remote = !config.now_power_off;
}

void CMT::release()
{
	close_datarec();
	cancel_my_event(0);
	cancel_my_event(1);

	delete[] tape_wave_buf;

	delete[] relay_data[1];
	delete[] relay_data[0];

	delete wav;
	delete fio;
}

void CMT::load_relay_wav()
{
	// load relay wav file
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];
	wav_header_t	 head;
	wav_fmt_chank_t  fmt;
	wav_data_chank_t data;
	size_t			 data_len;

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	for (int j=0; j<2; j++) {
		if ((relay_wav_enable & (1 << j)) == 0) {
			relay_data[j] = new uint8_t[sample_rate >> 2];	// 0.25sec
			memset(relay_data[j], 128, sizeof(uint8_t) * (sample_rate >> 2));
		}
	}

	FILEIO* fre = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		for (int j=0; j<2; j++) {
			if ((relay_wav_enable & (1 << j)) == 0) {
				UTILITY::stprintf(file_path, _MAX_PATH, _T("%s%s"), app_path, relay_file[j]);
				if(fre->Fopen(file_path, FILEIO_READ_BINARY)) {
					if (wav->util->CheckWavFormat(fre, &head, &fmt, &data, &data_len) == 0) {
						relay_size[j] = (int)wav->util->ReadWavData(fre, &fmt, data_len, relay_data[j], sample_rate, 8, sample_rate >> 2);
						relay_wav_enable |= (1 << j);
						emu->out_logf(LOG_INFO,_("%s was loaded."), relay_file[j]);
					}
					fre->Fclose();
				}
			}
		}
	}
	delete fre;

	for (int j=0; j<2; j++) {
		if ((relay_wav_enable & (1 << j)) == 0) {
			relay_size[j] = 0;
			delete[] relay_data[j];
			relay_data[j]=NULL;
			if (!relay_wav_loaded_at_first) {
				emu->out_logf(LOG_WARN,_("%s couldn't be loaded."), relay_file[j]);
			}
		}
	}
	relay_wav_loaded_at_first = true;
}

void CMT::write_io8(uint32_t addr, uint32_t data)
{
	set_data(data);
}

void CMT::set_data(uint32_t data)
{
	int len, c_len;

	if (remote && recv_ok && rec) {
		if (config.realmode_datarec) {
			set_tape_wave(data);
		}
		if (data_write_pos >= CMTDATA_BUFFER_SIZE) {
			buffer_overflow = true;
			if (d_ctrl != NULL) {
				d_ctrl->write_signal(SIG_ACIA_ERROR, 1, 1);
			}
		} else {
			switch (file_type) {
			case 0:
				// .wav
				if (need_header) {
					// wav header
					wav->SetSampleRate(config.wav_sample_rate);
					wav->SetSampleBits(config.wav_sample_bits);
					//
					data_write_pos = wav->InitFileHeader(data_buffer);
					data_len = data_write_pos;
					need_header = false;
				}
				wav->SetBaud(baud_sel);
				len = wav->SetData(data, &data_buffer[data_write_pos], &c_len);
				data_write_pos+=len;
				data_len+=len;

				data_total_pos+=c_len;
				break;
			case 1:
				// .l3c
				if (data_len >= 88) {
					// cr + lf
					data_buffer[data_write_pos]='\r';
					data_write_pos++;
					data_buffer[data_write_pos]='\n';
					data_write_pos++;
					data_len = 0;
				}
				wav->SetBaud(baud_sel);
				len = wav->SetData(data, &data_buffer[data_write_pos], &c_len);
				data_write_pos+=len;
				data_len+=len;

				data_total_pos+=c_len;
				break;
			case 2:
				// .l3b
				data |= 0x30;
				if (data_len >= 110
				&& (data_buffer[data_write_pos - 1] & 1) == 1
				&& (data & 1) == 0) {
					// cr + lf
					data_buffer[data_write_pos]='\r';
					data_write_pos++;
					data_buffer[data_write_pos]='\n';
					data_write_pos++;
					data_len = 0;
				}
				data_buffer[data_write_pos]=data & 0xff;
				data_write_pos++;
				data_len++;

				data_total_pos += 8;
				break;
			case 4:
				// .t9x
				if (need_header) {
					// set dummy header
					memset(data_buffer, 0, sizeof(uint8_t) * 64);
					data_write_pos = 64;
					data_len = data_write_pos;
					need_header = false;
				}
				t9x_data |= ((data & 1) << (t9x_dlen));
				emu->out_debugf(_T("t9x  w: %d-%d %d"), data_write_pos, t9x_dlen, data);
				t9x_dlen++;
				if (t9x_dlen >= 8) {
					data_buffer[data_write_pos]=t9x_data;
					emu->out_debugf(_T("t9x wt: %d-%d %02x"), data_write_pos, t9x_dlen, t9x_data);
					data_write_pos++;
					data_len++;
					t9x_data = 0;
					t9x_dlen = 0;
				}
				data_total_pos += 8;
				break;
			default:
				// .l3
				data_buffer[data_write_pos]=data & 0xff;
				data_write_pos++;
				data_len++;

				data_total_pos += 88;
				break;
			}
		}
	}
}

uint32_t CMT::read_io8(uint32_t addr)
{
	return get_data();
}

uint32_t CMT::get_data()
{
	uint32_t data = 0;
	static int cnt = 0;
	bool cont = false;

	// read 1byte from buffer
	if (remote) {
		do {
			cont = false;

			// read from file
			if (data_read_pos == data_read_size) {
				load_image();
			}
			if (data_read_size == 0) break;

			switch(file_type) {
			case 0:
			case 1:
				// .l3c or .wav

				// get 1 data
				data = data_buffer[data_read_pos];
				data_read_pos++;

				if (config.realmode_datarec) {
					set_tape_amp(c_onedata);
				}
				data_total_pos += (int)strlen(c_onedata);
				break;
			case 2:
				// .l3b

				// get 1 data
				data = data_buffer[data_read_pos];
				data_read_pos++;

				if ((data & 0xff) != 0x0a && (data & 0xff) != 0x0d) {
					if (config.realmode_datarec) {
						set_tape_wave(data);
					}
					data_total_pos += 8;
				} else {
					cont = true;
				}
				break;
			case 4:
				// .t9x
				if (t9x_dlen >= 8) {
					// get 1 data
					t9x_data = data_buffer[data_read_pos];
					data_read_pos++;
					t9x_dlen = 0;
				}
				data = ((t9x_data >> t9x_dlen) & 1);
//				emu->out_debugf(_T("t9x r: %d-%d %d"), data_read_pos, t9x_dlen, data);
				t9x_dlen++;

				if (config.realmode_datarec) {
					set_tape_wave(data);
				}
				data_total_pos += 8;
				break;
			default:
				// .l3 file

				// get 1 data
				data = data_buffer[data_read_pos];
				data_read_pos++;

				if (config.realmode_datarec) {
					set_tape_wave(data);
				}
				data_total_pos += 88;
				break;
			}
		} while (cont && data_read_pos <= data_read_size);
	} else {
		data = cnt ? 0xff : 0;
		cnt--;
		if (cnt < 0) cnt = 15;
	}
	return data;
}

uint32_t CMT::get_data_fast()
{
	uint32_t data = 0;
	bool cont = false;

	// read 1byte from buffer
	do {
		cont = false;

		// read from file
		if (data_read_pos == data_read_size) {
			load_image();
		}
		if (data_read_size == 0) break;

		// get 1 data
		data = data_buffer[data_read_pos];
		data_read_pos++;

		switch(file_type) {
		case 0:
		case 1:
			// .l3c or .wav
			data_total_pos += (int)strlen(c_onedata);
			break;
		case 2:
			// .l3b
			if ((data & 0xff) != 0x0a && (data & 0xff) != 0x0d) {
				data_total_pos += 8;
			} else {
				cont = true;
			}
			break;
		case 4:
			// .t3x
			data_total_pos += 64;
			break;
		default:
			// .l3 file
			data_total_pos += 88;
			break;
		}
	} while (cont && data_read_pos <= data_read_size);

	if (data_read_size == 0) {
		// end of data
		cancel_my_event(1);

		if (data_total_pos > 0) {
			// if file is not top position, will not output header.
			need_header = false;
		}

		// refresh file buffer
		fio->Fseek(0, FILEIO_SEEK_CUR);
	}
	return data;
}

void CMT::write_signal(int id, uint32_t data, uint32_t mask)
{
	switch(id) {
		case SIG_CMT_REMOTE:
			// remote on/off
			remote_prev = remote;
			remote = (data & mask) ? true : false;
			break;
		case SIG_CMT_BAUD_SEL:
			// baud select
			baud_sel = (data & mask) ? 1 : 0;
			set_baud(baud_sel);
			if (register_id[0] != -1) {
				cancel_my_event(0);
				register_my_event(0);
			}
			break;
		case SIG_ACIA_DTR:
			send_ok = (data & mask) ? true : false;
			break;
		case SIG_ACIA_RTS:
			recv_ok = (data & mask) ? true : false;
			break;
		case SIG_ACIA_TXDATA:
			// receive data from acia
			set_data(data);
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			if (now_reset) {
				reset();
			}
			break;
	}

}

void CMT::set_baud(int sel)
{
	switch(sel) {
	case 1:
		// 1200baud
		baud_rate = 1200;
		tape_wave_nums = 2;
		break;
	default:
		// 600baud
		baud_rate = 600;
		tape_wave_nums = 4;
	}
}

bool CMT::play_datarec(const _TCHAR* filename)
{
	close_datarec();

	if(fio->Fopen(filename, FILEIO_READ_BINARY)) {
		check_extension(filename);

		// load image file
		play = true;
		cancel_my_event(0);
		register_my_event(0);

		play = set_load_param();
//		if (play) load_image();

	}
	return play;
}

void CMT::rewind_datarec()
{
	if (play) {
		fio->Fseek(0, FILEIO_SEEK_SET);

		play = set_load_param();
		load_image();
	}
	if (rec) {
		fio->Fseek(0, FILEIO_SEEK_SET);

		data_write_pos = 0;
		data_len = 0;

		set_save_param();
	}

	data_total_pos = 0;
}

void CMT::fast_forward_datarec()
{
	if (play || rec) {
		register_my_event(1);
	}
}

void CMT::realmode_datarec()
{
	config.realmode_datarec = !config.realmode_datarec;
	if (register_id[0] != -1) {
		cancel_my_event(0);
		register_my_event(0);
	}
}

bool CMT::datarec_opened(bool play_mode)
{
	if (play_mode) {
		return play;
	} else {
		return rec;
	}
}

bool CMT::rec_datarec(const _TCHAR* filename)
{
	close_datarec();

	if(!(fio->Fopen(filename, FILEIO_READ_WRITE_BINARY)
	  || fio->Fopen(filename, FILEIO_WRITE_BINARY))) {
		// open failed
		return false;
	}
	fio->Fclose();
	if(fio->Fopen(filename, FILEIO_READ_WRITE_BINARY))	{
		check_extension(filename);

		data_write_pos = 0;
		data_len = 0;
		rec = true;

		cancel_my_event(0);
		register_my_event(0);

		set_save_param();

		data_total_pos = 0;
	}

	return rec;
}

void CMT::close_datarec()
{
	// close file
	if(rec) {
		save_image();
		if (file_type == 0) {
			// .wav file
			wav->SetFileHeader(fio);
		} else if (file_type == 4) {
			// .t9x
			set_file_header(fio, file_type);
		}
	}

	if(play || rec) {
		cancel_my_event(0);
		fio->Fclose();
	}
	cancel_my_event(1);

	play = rec = false;
}

bool CMT::set_load_param()
{
	if (file_type <= 1) {
		// .l3c or .wav
		if (wav->CheckFileFormat(fio, file_type) != 0) {
			emu->out_log(LOG_ERROR, wav->Errmsg());
			return false;
		}
		wav->SetReverseWave(config.wav_reverse);
		wav->SetHalfWave(config.wav_half);
		wav->SetCorrectWave(config.wav_correct);
		wav->SetCorrectType(config.wav_correct_type);
		wav->InitData(fio, file_type);
	} else {
		// .l3b or .l3 or .t9x
		if (check_file_format(fio, file_type) != 0) {
			emu->out_logf(LOG_ERROR,_T("This is not %s format file."),support_ext[file_type]);
			return false;
		}
	}
	data_read_pos = 0;
	data_read_size = 0;

	t9x_data = 0;
	t9x_dlen = 8;

	data_total_pos = 0;

	return true;
}

bool CMT::set_save_param()
{
	if (file_type <= 1) {
		// .l3c or .wav
		int rc = wav->CheckFileFormat(fio, file_type);
		if (rc > 1) {
			emu->out_log(LOG_ERROR, wav->Errmsg());
			return false;
		}
		if (rc == 0) {
			// get sample_rate and sample_bits parameters from file data.
			config.wav_sample_rate = wav->GetSampleRate();
			config.wav_sample_bits = wav->GetSampleBits();
		} else {
			wav->SetSampleRate(config.wav_sample_rate);
			wav->SetSampleBits(config.wav_sample_bits);
		}
		wav->InitData(fio, file_type);
	} else {
		// .l3b or .l3 or .t9x
		if (check_file_format(fio, file_type) > 1) {
			emu->out_logf(LOG_ERROR,_T("This is not %s format file."),support_ext[file_type]);
			return false;
		}
	}
	data_read_pos = data_write_pos;
	data_read_size = data_write_pos;

	t9x_data = 0;
	t9x_dlen = 0;

	need_header = true;

	return true;
}

// binary image
void CMT::load_image()
{
	memset(data_buffer, 0x00, sizeof(data_buffer));
	if (file_type <= 1) {
		// .l3c or .wav
		wav->SetBaud(baud_sel);
		c_onedata[0]='\0';
		data_buffer[0]='\0'; 
		data_read_size = wav->GetData((char *)data_buffer, c_onedata);
		data_read_size = (data_read_size >= 0 ? data_read_size : 0);
		data_read_pos = 0;
		data_write_pos = 0;
		data_len = 0;
	} else {
		// .l3b or .l3 or .t9x
		data_read_size = (int)fio->Fread(data_buffer, sizeof(uint8_t), sizeof(data_buffer));
		data_read_pos = 0;
		data_write_pos = 0;
		data_len = 0;
	}
}

void CMT::save_image()
{
	// binary image
	if (data_write_pos > 0) {
		fio->Fwrite(data_buffer, data_write_pos, 1);
		data_write_pos = 0;
	}
}

int CMT::check_extension(const _TCHAR* filename)
{
	int nam_len = (int)_tcslen(filename);

	int i = 0;
	file_type = -1;
	is_bytedata = false;
	while (support_ext[i] != NULL) {
		int ext_len = (int)_tcslen(support_ext[i]);
		if(nam_len >= ext_len) {
#ifdef _WIN32
			if(_tcsnicmp(&filename[nam_len - ext_len], support_ext[i], ext_len) == 0) {
#else
#ifdef _UNICODE
			if(wcsncasecmp(&filename[nam_len - ext_len], support_ext[i], ext_len) == 0) {
#else
			if(strncasecmp(&filename[nam_len - ext_len], support_ext[i], ext_len) == 0) {
#endif
#endif
				file_type = i;
				break;
			}
		}
		i++;
	}
	if (file_type < 0) {
		// default = .l3
		file_type = 3;
	}
	if (file_type == 3) {
		is_bytedata = true;
	}
	return file_type;
}

// ----------------------------------------------------------------------------
int CMT::check_file_format(FILEIO *fp, int file_type)
{
	int rc = 0;
	switch(file_type) {
	case 4:
		// .t9x
		if (fp->Fread(&t9x_header, sizeof(t9x_header), 1) != 1) {
			rc = 1;
			break;
		}
		if (memcmp(t9x_header.ident, T9X_IDENTIFIER, 32) != 0) {
			rc = 1;
			break;
		}
		break;
	default:
		break;
	}
	return rc;
}

int CMT::set_file_header(FILEIO *fp, int file_type)
{
	int rc = 0;
	switch(file_type) {
	case 4:
		// .t9x
		{
			int data[2];
			// if data remain, flush it
			if (t9x_dlen > 0) {
				fp->Fputc(t9x_data);
			}
			//
			fp->Fseek(0, FILEIO_SEEK_SET);
			fp->Fwrite((void *)T9X_IDENTIFIER, sizeof(char), 32);
			data[0]=0;
			data[1]=9;	// TODO: unknown number?
			fp->Fwrite(data, sizeof(int), 2);
		}
		break;
	default:
		break;
	}
	return rc;
}

// ----------------------------------------------------------------------------
void CMT::register_my_event(int event_id)
{
	int period = 1;

	if(register_id[event_id] != -1) return;

	switch(event_id) {
	case 0:
		if (config.realmode_datarec) {
			if (is_bytedata) {
				// read/write 1byte data per one process
				period = (int)(CPU_CLOCKS / baud_rate * 11);	// 1startbit + 8databit + 2stopbit
			} else {
				// read/write 1bit data per one process
				period = (int)(CPU_CLOCKS / baud_rate);
			}
		} else {
			period = (int)(CPU_CLOCKS / 38400);
		}
		register_event_by_clock(this, event_id, period, true, &register_id[event_id]);
		break;
	case 1:
		period = 4 * ((file_type != 3) ? 1 : 11);
		register_event_by_clock(this, event_id, period, true, &register_id[event_id]);
		break;
	}
}

void CMT::cancel_my_event(int event_id)
{
	if(register_id[event_id] != -1) {
		cancel_event(this, register_id[event_id]);
	}
	register_id[event_id] = -1;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void CMT::event_callback(int event_id, int err)
{
	uint32_t data = 0;
	uint32_t mask = 0;

	switch(event_id) {
	case 0:
		if(remote) {
			// send to acia
			if (d_ctrl != NULL) {
				// send to rxclk
				d_ctrl->write_signal(SIG_ACIA_RXCLK, 1, 1);

				if (play && (send_ok || now_reset)) {
					// read data
					data = get_data();

					if (is_bytedata) {
						// .l3
						mask = 0xff;
					} else {
						mask = 1;
					}
					// send data to acia
					d_ctrl->write_signal(SIG_ACIA_RXDATA, data, mask);
				}
			}
			// receive from acia
			// send to txclk then acia send data to me
			if (d_ctrl != NULL) {
				if (is_bytedata) {
					// .l3
					mask = 0xff;
				} else {
					mask = 1;
				}
				d_ctrl->write_signal(SIG_ACIA_TXCLK, 1, mask);
			}
			if (rec) {
				// write data to file
				if (data_write_pos > 11) {
					save_image();
				}
			}
		}
		break;
	case 1:
		// f.f.
		get_data_fast();
		break;
	}
//	update_event();
}

// ----------------------------------------------------------------------------
void CMT::mix(int32_t* buffer, int cnt)
{
	int32_t *buffer_orig = buffer;

	// relay wave sound
	if ((relay_wav_enable & 1) != 0 && remote == true && remote_prev == false) {
		// relay on sound
		relay_play = 1;
		relay_play_pos = 0;
	} else if ((relay_wav_enable & 2) != 0 && remote == false && remote_prev == true) {
		// relay off sound
		relay_play = 2;
		relay_play_pos = 0;
	}
	remote_prev = remote;

	if (relay_play) {
		if (relay_volume) {
			for(int i=0; i<cnt; i++) {
				int32_t sample = ((int)relay_data[relay_play-1][relay_play_pos] - 128) * relay_volume / 128;
				*buffer++ += sample;
				*buffer++ += sample;
				relay_play_pos++;
				if (relay_play_pos >= relay_size[relay_play-1]) {
					break;
				}
			}
		}
		if (relay_play_pos >= relay_size[relay_play-1]) {
			relay_play = 0;
		}
	}

	// tape wave sound
	if (tape_wave_buf_wpos > tape_wave_buf_rpos) {
		buffer = buffer_orig;
		for(int i=0; i<cnt; i++) {
			*buffer++ += tape_wave_buf[tape_wave_buf_rpos];
			*buffer++ += tape_wave_buf[tape_wave_buf_rpos];
			tape_wave_buf_rpos++;
			if (tape_wave_buf_wpos <= tape_wave_buf_rpos) {
				break;
			}
		}
		// shift
		if (tape_wave_buf_wpos > (TAPEWAV_BUFFER_SIZE/2)) {
			for(int pos=tape_wave_buf_rpos; pos<tape_wave_buf_wpos; pos++) {
				tape_wave_buf[pos-tape_wave_buf_rpos] = tape_wave_buf[pos];
			}
			tape_wave_buf_wpos -= tape_wave_buf_rpos;
			tape_wave_buf_rpos = 0;
		}
//	} else {
//		surplus = 0;
	}
}

void CMT::set_volume(int decibel_relay, int decibel_tape, bool mute_relay, bool mute_tape)
{
	relay_volume = int(16384.0 * pow(10.0, decibel_relay / 40.0));
	if (mute_relay) relay_volume = 0;

	tape_volume = int(16384.0 * pow(10.0, decibel_tape / 40.0));
	if (mute_tape) tape_volume = 0;
}

void CMT::initialize_sound(int rate, int decibel)
{
	sample_rate_base = sample_rate = rate;
//	max_vol = volume;
	set_volume(decibel, decibel, false, false);
	relay_volume = 0;
	tape_volume = 0;

	// load relay wav file
	load_relay_wav();

	// create tape wave
	create_tape_wave();
}

// ----------------------------------------------------------------------------
void CMT::create_tape_wave()
{
	tape_wave_buf = new int16_t[TAPEWAV_BUFFER_SIZE + 1];

	devide = 0;
	surplus = 0;
	freq_prev = 2400;
	a_prev = 0;

	tape_wave_onoff = 1;
}

void CMT::set_tape_wave(uint32_t data)
{
	if (!is_bytedata) {
		// 1bit data
		if (data & 1) {
			// 2400Hz * 4
			set_tape_one_wave(tape_wave_nums << 1, 1);
		} else {
			// 1200Hz * 2
			set_tape_one_wave(tape_wave_nums, 2);
		}
	} else {
		// 1byte data
		// startbit
		// 1200Hz * 2
		set_tape_one_wave(tape_wave_nums, 2);
		// databit
		for(int b=0; b<8; b++) {
			if (data & (1 << b)) {
				// 2400Hz * 4
				set_tape_one_wave(tape_wave_nums << 1, 1);
			} else {
				// 1200Hz * 2
				set_tape_one_wave(tape_wave_nums, 2);
			}
		}
		// 2stopbit
		// 2400Hz * 4 * 2
		set_tape_one_wave(tape_wave_nums << 2, 1);
	}
}

/// @param cnt [in] wave nums (half wave)
/// @param samp [in] 0
void CMT::set_tape_one_wave(int cnt, int samp)
{
	for(int j=0; j<cnt; j++) {
		for (int n=0; n<samp; n++) {
			set_tape_one_amp(tape_wave_onoff);
		}
		tape_wave_onoff = (1 - tape_wave_onoff);
	}
}

void CMT::set_tape_amp(char *c_data)
{
	char *p = c_data;

	while(p != NULL && *p != '\0') {
		set_tape_one_amp(*p);
		p++;
	}
}

void CMT::set_tape_one_amp(int data)
{
	devide = (sample_rate + surplus) / TAPE_WAV_FREQ;
	surplus = (sample_rate + surplus) - (devide * TAPE_WAV_FREQ);

	int a = tape_volume;
	if ((data & 1) == 0) {
		a = -a;
	}
	if (devide > 0 && tape_wave_buf_wpos < TAPEWAV_BUFFER_SIZE) {
		if (abs(a + a_prev + a_prev) > abs(a)) a_prev = 0;

		tape_wave_buf[tape_wave_buf_wpos] = a + a_prev + a_prev;
		tape_wave_buf_wpos++;
	}
	for(int i=1; i<devide && tape_wave_buf_wpos < TAPEWAV_BUFFER_SIZE; i++) {
		tape_wave_buf[tape_wave_buf_wpos] = a;
		tape_wave_buf_wpos++;
	}
	a_prev = 0;
	if (surplus > 0) {
		a_prev = tape_volume * surplus / TAPE_WAV_FREQ;
		if ((data & 1) == 0) {
			a_prev = -a_prev;
		}
	}
//	emu->out_debugf(_T("CMT:one_amp: wpos:%6d de:%2d sr:%4d d:%2d a:%6d a_pre:%6d"), tape_wave_buf_wpos, devide, surplus, data, a, a_prev);
}

// ----------------------------------------------------------------------------
uint32_t CMT::get_cmt_mode()
{
	return (remote && play ? 0x01 : 0) | (remote && rec ? 0x02 : 0) | (play ? 0x04 : 0) | (rec ? 0x08 : 0) | (((data_total_pos / 800) % 12000) << 4);
}

// ----------------------------------------------------------------------------
void CMT::update_config()
{
	int cpu_power = (config.sync_irq ? config.cpu_power : 1);
	if (cpu_power >= 1) {
		sample_rate = sample_rate_base >> (cpu_power - 1);
	} else {
		sample_rate = sample_rate_base << (1 - cpu_power);
	}
}

// ----------------------------------------------------------------------------

void CMT::save_state(FILEIO *fp)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = Uint16_LE(3);
	vm_state_ident.size = Uint32_LE(sizeof(vm_state_ident) + sizeof(vm_state));

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.flags = (remote ? 1 : 0) | (play ? 2 : 0) | (rec ? 4 : 0);
	vm_state.flags |= (send_ok ? 0x10 : 0) | (recv_ok ? 0x20 : 0);
	vm_state.flags |= (config.realmode_datarec ? 0x80 : 0); // add version 2
	vm_state.baud_sel = baud_sel; // add version 3

	vm_state.register_id = Int32_LE(register_id[0]);

	fp->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fp->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void CMT::load_state(FILEIO *fp)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fp, vm_state_i, vm_state);

	// copy values
	remote = (vm_state.flags & 1) ? true : false;
//	play   = (vm_state.flags & 2) ? true : false;
//	rec    = false;	// ignore record mode
	baud_sel = (vm_state.flags & 8) ? 1 : 0;
	send_ok = (vm_state.flags & 0x10) ? true : false;
	recv_ok = (vm_state.flags & 0x20) ? true : false;
	if (Uint16_LE(vm_state_i.version) >= 2) {
		config.realmode_datarec = (vm_state.flags & 0x80) ? true : false;
	}
	if (Uint16_LE(vm_state_i.version) >= 3) {
		baud_sel = vm_state.baud_sel;
	}
	register_id[0] = Int32_LE(vm_state.register_id);
}

// ----------------------------------------------------------------------------

#ifdef USE_DEBUGGER
bool CMT::debug_write_reg(uint32_t reg_num, uint32_t data)
{
	switch(reg_num) {
	case 0:
		write_signal(SIG_CMT_REMOTE, data, 0x80);
		return true;
	case 1:
		write_signal(SIG_CMT_BAUD_SEL, data, 0x0f);
		return true;
	}
	return false;
}

void CMT::debug_regs_info(_TCHAR *buffer, size_t buffer_len)
{
	buffer[0] = _T('\0');
	_stprintf(&buffer[_tcslen(buffer)], _T(" %X(%s):%02X"), 0, _T("REMOTE"), remote ? 0x80 : 0);
	_stprintf(&buffer[_tcslen(buffer)], _T(" %X(%s):%02X"), 1, _T("BAUD_SEL"), baud_sel);
}
#endif