/** @file debugger.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2014.09.02-

	@brief [ debugger ]

	@note Modified by Sasaji at 2016.02.01 -
*/

#include "debugger.h"

#ifdef USE_DEBUGGER

DEBUGGER::DEBUGGER(VM* parent_vm, EMU* parent_emu, const char *identifier) : DEVICE(parent_vm, parent_emu, identifier)
{
	for(int i=0; i<BP_END; i++) {
		memset(&bp[i], 0, sizeof(bp[i]));
	}
	for(int i=0; i<TP_END; i++) {
		memset(&tp[i], 0, sizeof(tp[i]));
	}
	_tcscpy(file_path, _T("debug.bin"));
	now_going = 0;
	now_debugging = now_suspended = false;
	now_breakpoint = now_tracepoint = false;

	server = false;
	client_ch = -1;
	server_ch = -1;
	connect = false;

	memset(send_buff, 0, sizeof(send_buff));
	send_buff_w_pos = 0;
	send_buff_r_pos = 0;
	memset(recv_buff, 0, sizeof(recv_buff));
	recv_buff_w_pos = 0;
	recv_buff_r_pos = 0;

	init_lock();
}

DEBUGGER::~DEBUGGER()
{
	term_lock();
}

// ----------------------------------------------------------------------------

void DEBUGGER::check_mem_break_points(break_point_t *bp, uint32_t addr, int length, bool &judge, uint32_t type)
{
	for(int i = 0; i < MAX_BREAK_POINTS; i++) {
		break_point_table_t *bpt = &bp->table[i];
		if(bpt->status == 1 && bpt->mask == type) {
			if(addr >= bpt->addr
			&& addr < (bpt->addr + bpt->len + length)) {
				bp->hit = now_suspended = judge = true;
				bp->hit_addr = bpt->addr;
				break;
			}
		}
	}
}

void DEBUGGER::check_io_break_points(break_point_t *bp, uint32_t addr, bool &judge)
{
	for(int i = 0; i < MAX_BREAK_POINTS; i++) {
		break_point_table_t *bpt = &bp->table[i];
		if(bpt->status == 1) {
			if((addr & bpt->mask) == (bpt->addr & bpt->mask)) {
				bp->hit = now_suspended = judge = true;
				bp->hit_addr = addr;
				break;
			}
		}
	}
}

void DEBUGGER::check_intr_break_points(break_point_t *bp, uint32_t addr, bool &judge)
{
	for(int i = 0; i < MAX_BREAK_POINTS; i++) {
		break_point_table_t *bpt = &bp->table[i];
		if(bpt->status == 1) {
			if(addr == bpt->addr) {
				bp->hit = now_suspended = judge = true;
				bp->hit_addr = addr;
				break;
			}
		}
	}
}

// ----------------------------------------------------------------------------

void DEBUGGER::check_break_points(uint32_t addr)
{
#ifdef _MBS1
	check_mem_break_points(&bp[BP_FETCH_OP], addr, 1, now_breakpoint, 0);
	check_mem_break_points(&tp[TP_FETCH_OP], addr, 1, now_tracepoint, 0);
	uint32_t phy_addr = d_mem->debug_latch_address(addr);
	check_mem_break_points(&bp[BP_FETCH_OP], phy_addr, 1, now_breakpoint, 1);
	check_mem_break_points(&tp[TP_FETCH_OP], phy_addr, 1, now_tracepoint, 1);
#else
	check_mem_break_points(&bp[BP_FETCH_OP], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_FETCH_OP], addr, 1, now_tracepoint);
#endif
}

void DEBUGGER::check_intr_break_points(uint32_t addr)
{
	check_intr_break_points(&bp[BP_INTERRUPT], addr, now_breakpoint);
	check_intr_break_points(&tp[BP_INTERRUPT], addr, now_tracepoint);
}

void DEBUGGER::store_break_points()
{
	for(int i=0; i<BP_END; i++) {
		memcpy(bp[i].stored, bp[i].table, sizeof(bp[i].table));
		memset(bp[i].table, 0, sizeof(bp[i].table));
	}
	for(int i=0; i<TP_END; i++) {
		memcpy(tp[i].stored, tp[i].table, sizeof(tp[i].table));
		memset(tp[i].table, 0, sizeof(tp[i].table));
	}
}

void DEBUGGER::restore_break_points()
{
	for(int i=0; i<BP_END; i++) {
		memcpy(bp[i].table, bp[i].stored, sizeof(bp[i].table));
	}
	for(int i=0; i<TP_END; i++) {
		memcpy(tp[i].table, tp[i].stored, sizeof(tp[i].table));
	}
}

int DEBUGGER::hit_break_point()
{
	int i = -1;
	if (now_breakpoint) {
		i = BP_END - 1;
		for(; i>=0; i--) {
			if (bp[i].hit) break;
		}
	}
	return i;
}

int DEBUGGER::hit_trace_point()
{
	int i = TP_END - 1;
	if (now_tracepoint) {
		for(; i>=0; i--) {
			if (tp[i].hit) break;
		}
	}
	return i;
}

void DEBUGGER::set_break_point(bool val)
{
	for(int i=0; i<BP_END; i++) {
		bp[i].hit = val;
	}
	now_breakpoint = val;
}

void DEBUGGER::set_trace_point(bool val)
{
	for(int i=0; i<TP_END; i++) {
		tp[i].hit = val;
	}
	now_tracepoint = val;
}

// ----------------------------------------------------------------------------

void DEBUGGER::write_data8(uint32_t addr, uint32_t data)
{
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 1, now_tracepoint);
	d_mem->write_data8(addr, data);
}
uint32_t DEBUGGER::read_data8(uint32_t addr)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 1, now_tracepoint);
	return d_mem->read_data8(addr);
}
void DEBUGGER::write_data16(uint32_t addr, uint32_t data)
{
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 2, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 2, now_tracepoint);
	d_mem->write_data16(addr, data);
}
uint32_t DEBUGGER::read_data16(uint32_t addr)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 2, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 2, now_tracepoint);
	return d_mem->read_data16(addr);
}
void DEBUGGER::write_data32(uint32_t addr, uint32_t data)
{
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 4, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 4, now_tracepoint);
	d_mem->write_data32(addr, data);
}
uint32_t DEBUGGER::read_data32(uint32_t addr)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 4, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 4, now_tracepoint);
	return d_mem->read_data32(addr);
}
void DEBUGGER::write_data8w(uint32_t addr, uint32_t data, int* wait)
{
#ifdef _MBS1
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 1, now_breakpoint, 0);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 1, now_tracepoint, 0);
	uint32_t phy_addr = d_mem->debug_latch_address(addr);
	check_mem_break_points(&bp[BP_WRITE_MEMORY], phy_addr, 1, now_breakpoint, 1);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], phy_addr, 1, now_tracepoint, 1);
#else
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 1, now_tracepoint);
#endif
	d_mem->write_data8w(addr, data, wait);
}
uint32_t DEBUGGER::read_data8w(uint32_t addr, int* wait)
{
#ifdef _MBS1
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 1, now_breakpoint, 0);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 1, now_tracepoint, 0);
	uint32_t phy_addr = d_mem->debug_latch_address(addr);
	check_mem_break_points(&bp[BP_READ_MEMORY], phy_addr, 1, now_breakpoint, 1);
	check_mem_break_points(&tp[TP_READ_MEMORY], phy_addr, 1, now_tracepoint, 1);
#else
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 1, now_tracepoint);
#endif
	return d_mem->read_data8w(addr, wait);
}
void DEBUGGER::write_data16w(uint32_t addr, uint32_t data, int* wait)
{
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 2, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 2, now_tracepoint);
	d_mem->write_data16w(addr, data, wait);
}
uint32_t DEBUGGER::read_data16w(uint32_t addr, int* wait)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 2, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 2, now_tracepoint);
	return d_mem->read_data16w(addr, wait);
}
void DEBUGGER::write_data32w(uint32_t addr, uint32_t data, int* wait)
{
	check_mem_break_points(&bp[BP_WRITE_MEMORY], addr, 4, now_breakpoint);
	check_mem_break_points(&tp[TP_WRITE_MEMORY], addr, 4, now_tracepoint);
	d_mem->write_data32w(addr, data, wait);
}
uint32_t DEBUGGER::read_data32w(uint32_t addr, int* wait)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 4, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 4, now_tracepoint);
	return d_mem->read_data32w(addr, wait);
}
uint32_t DEBUGGER::fetch_op(uint32_t addr, int *wait)
{
	check_mem_break_points(&bp[BP_READ_MEMORY], addr, 1, now_breakpoint);
	check_mem_break_points(&tp[TP_READ_MEMORY], addr, 1, now_tracepoint);
	return d_mem->fetch_op(addr, wait);
}
void DEBUGGER::latch_address(uint32_t addr, int *wait)
{
	d_mem->latch_address(addr, wait);
}

void DEBUGGER::write_io8(uint32_t addr, uint32_t data)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io8(addr, data);
}
uint32_t DEBUGGER::read_io8(uint32_t addr)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io8(addr);
}
void DEBUGGER::write_io16(uint32_t addr, uint32_t data)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io16(addr, data);
}
uint32_t DEBUGGER::read_io16(uint32_t addr)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io16(addr);
}
void DEBUGGER::write_io32(uint32_t addr, uint32_t data)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io32(addr, data);
}
uint32_t DEBUGGER::read_io32(uint32_t addr)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io32(addr);
}
void DEBUGGER::write_io8w(uint32_t addr, uint32_t data, int* wait)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io8w(addr, data, wait);
}
uint32_t DEBUGGER::read_io8w(uint32_t addr, int* wait)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io8w(addr, wait);
}
void DEBUGGER::write_io16w(uint32_t addr, uint32_t data, int* wait)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io16w(addr, data, wait);
}
uint32_t DEBUGGER::read_io16w(uint32_t addr, int* wait)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io16w(addr, wait);
}
void DEBUGGER::write_io32w(uint32_t addr, uint32_t data, int* wait)
{
	check_io_break_points(&bp[BP_OUTPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_OUTPUT_IO], addr, now_tracepoint);
	d_io->write_io32w(addr, data, wait);
}
uint32_t DEBUGGER::read_io32w(uint32_t addr, int* wait)
{
	check_io_break_points(&bp[BP_INPUT_IO], addr, now_breakpoint);
	check_io_break_points(&tp[TP_INPUT_IO], addr, now_tracepoint);
	return d_io->read_io32w(addr, wait);
}

void DEBUGGER::write_signal(int id, uint32_t data, uint32_t mask)
{
	d_mem->write_signal(id, data, mask);
}
uint32_t DEBUGGER::read_signal(int id)
{
	return d_mem->read_signal(id);
}

// ----------------------------------------------------------------------------

void DEBUGGER::enable_server(bool enable)
{
	if (enable) {
		if (config.debugger_server_port == 0 || _tcslen(config.debugger_server_host) == 0) {
			enable = false;
		}
	}

	if (!server && enable) {
		// connection active as client
		if (client_ch != -1) {
			// error
			emu->out_log(LOG_ERROR, _T("Already connecting to another debugger server."));
			return;
		}
		// get socket channel
		server_ch = emu->get_socket_channel();
		if (server_ch < 0) {
			return;
		}
		// start server
		if (!emu->init_socket_tcp(server_ch, this, true)) {
			server_ch = -1;
			emu->out_log(LOG_ERROR, _T("Network socket initialize failed."));
			return;
		}
		if (!emu->connect_socket(server_ch, config.debugger_server_host, config.debugger_server_port, true)) {
			server_ch = -1;
			emu->out_log(LOG_ERROR, _T("Cannot start as debugger server."));
			return;
		}
//		emu->out_log(LOG_DEBUG, _T("Started debugger server."));
		server = true;
	} else {
		// stop server
		if (client_ch != -1) {
			// disconnect client
			emu->disconnect_socket(client_ch);
		}
		if (server_ch >= 0) {
			emu->disconnect_socket(server_ch);
		}
//		emu->out_log(LOG_DEBUG, _T("Stopped debugger server."));
		server = false;
	}
}

bool DEBUGGER::connect_socket()
{
	if (!server && client_ch == -1) {
		// get socket channel
		client_ch = emu->get_socket_channel();
		if (client_ch < 0) {
			return false;
		}
		// connect
		if (!emu->init_socket_tcp(client_ch, this)) {
			client_ch = -1;
			emu->out_log(LOG_ERROR, _T("Network socket initialize failed."));
			return false;
		}
		if (!emu->connect_socket(client_ch, config.debugger_server_host, config.debugger_server_port)) {
			client_ch = -1;
			emu->out_log(LOG_ERROR, _T("Cannot connect to debugger server."));
			return false;
		}
//		emu->out_log(LOG_DEBUG, _T("Connect debugger client."));
	}
	return true;
}

void DEBUGGER::disconnect_socket()
{
	if (client_ch != -1) {
		// disconnect
		emu->disconnect_socket(client_ch);
//		emu->out_log(LOG_DEBUG, _T("Disconnect debugger client."));
	}
}

void DEBUGGER::enable_connect()
{
	if (!connect) {
		connect_socket();
	} else {
		disconnect_socket();
	}
}


uint8_t* DEBUGGER::get_sendbuffer(int ch, int* size, int* flags)
{
	if (ch == client_ch) {
		*size = (send_buff_w_pos - send_buff_r_pos);
		return &send_buff[send_buff_r_pos];
	} else {
		*size = 0;
		return NULL;
	}
}
void DEBUGGER::inc_sendbuffer_ptr(int ch, int size)
{
	if (ch == client_ch) {
		send_buff_r_pos += size;
		if (send_buff_w_pos <= send_buff_r_pos) {
			// all data sent.
			send_buff_w_pos = 0;
			send_buff_r_pos = 0;
		}
		return;
	}
}
uint8_t* DEBUGGER::get_recvbuffer0(int ch, int* size0, int* size1, int* flags)
{
	if (ch == client_ch) {
		lock();
		*size0 = (DEBUGGER_MAX_BUFF - recv_buff_w_pos);
		*size1 = 0;
		return &recv_buff[recv_buff_w_pos];
	} else {
		*size0 = 0;
		*size1 = 0;
		return NULL;
	}
}

uint8_t* DEBUGGER::get_recvbuffer1(int ch)
{
	if (ch == client_ch) {
		return recv_buff;
	} else {
		return NULL;
	}
}

void DEBUGGER::inc_recvbuffer_ptr(int ch, int size)
{
	if (ch == client_ch) {
		recv_buff_w_pos += size;
		unlock();
	}
}

void DEBUGGER::network_connected(int ch)
{
	if (ch == client_ch) {
		connect = true;
	}
}
void DEBUGGER::network_disconnected(int ch)
{
	if (ch == client_ch) {
		connect = false;
		client_ch = -1;
	} else if (ch == server_ch) {
		server_ch = -1;
	}
}
void DEBUGGER::network_writeable(int ch)
{
}
void DEBUGGER::network_readable(int ch)
{
}
void DEBUGGER::network_accepted(int ch, int new_ch)
{
	if (ch == server_ch) {
		if (client_ch != -1) {
			// Another client is already connected on my server.
			emu->disconnect_socket(new_ch);
		}
		client_ch = new_ch;
		connect = true;

		emu->debugger_terminal_accepted(this);
	}
}


void DEBUGGER::write_data(const _TCHAR *data, int size)
{
#ifdef _UNICODE
	// TODO: convert wide to narrow
#endif
	send_buff_r_pos = 0;
	send_buff_w_pos = 0;
	// convert lf to crlf
	for(int i=0; i<size; i++) {
		if (data[i] == 0x0a) {
			send_buff[send_buff_w_pos++] = 0x0d;
		}
		send_buff[send_buff_w_pos++] = data[i];
	}
	if (client_ch >= 0 && send_buff_r_pos < send_buff_w_pos) {
		emu->send_data_tcp(client_ch);
	}
}

int DEBUGGER::read_data(_TCHAR *data, int size)
{
	int read_size = 0;
	lock();
	if (recv_buff_r_pos < recv_buff_w_pos) {
		read_size = recv_buff_w_pos - recv_buff_r_pos;
		if (size < read_size) read_size = size;
		memcpy(data, &recv_buff[recv_buff_r_pos], read_size);

		// shift
		memcpy(&recv_buff[0], &recv_buff[recv_buff_r_pos], read_size);
		recv_buff_w_pos -= read_size;
		recv_buff_r_pos = 0;
	}
	unlock();
	return read_size;
}

void DEBUGGER::init_lock()
{
	mutex = new CMutex();
}

void DEBUGGER::term_lock()
{
	delete mutex;
	mutex = NULL;
}

void DEBUGGER::lock()
{
	mutex->lock();
}

void DEBUGGER::unlock()
{
	mutex->unlock();
}

#endif
