/** @file debugger.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2014.09.02-

	@brief [ debugger ]

	@note Modified by Sasaji at 2016.02.01 -
*/

#ifndef _DEBUGGER_H_
#define _DEBUGGER_H_

#include "../emu.h"
#include "vm.h"

#ifdef USE_DEBUGGER

#include "device.h"

#define DEBUGGER_MAX_BUFF 1024

class DEBUGGER : public DEVICE
{
private:
	DEVICE *d_mem, *d_io;
	
	void check_mem_break_points(break_point_t *bp, uint32_t addr, int length, bool &judge, uint32_t type = 0);
	void check_io_break_points(break_point_t *bp, uint32_t addr, bool &judge);
	void check_intr_break_points(break_point_t *bp, uint32_t addr, bool &judge);
	
// -----

	bool server;
	int client_ch;
	int server_ch;
	bool connect;

//	bool now_receiving;
	uint8_t send_buff[DEBUGGER_MAX_BUFF];
	int send_buff_w_pos;
	int send_buff_r_pos;
	uint8_t recv_buff[DEBUGGER_MAX_BUFF];
	int recv_buff_w_pos;
	int recv_buff_r_pos;

	bool connect_socket();
	void disconnect_socket();

	CMutex *mutex;

	inline void init_lock();
	inline void term_lock();
	inline void lock();
	inline void unlock();

// ----------------------------------------------------------------------------

public:
	DEBUGGER(VM* parent_vm, EMU* parent_emu, const char *identifier);
	~DEBUGGER();

	enum en_break_point_type {
		BP_FETCH_OP = 0,
		BP_READ_MEMORY,
		BP_WRITE_MEMORY,
		BP_INPUT_IO,
		BP_OUTPUT_IO,
		BP_INTERRUPT,
		BP_END
	};
	enum en_trace_point_type {
		TP_FETCH_OP = 0,
		TP_READ_MEMORY,
		TP_WRITE_MEMORY,
		TP_INPUT_IO,
		TP_OUTPUT_IO,
		TP_INTERRUPT,
		TP_END
	};

	// common functions
	void write_data8(uint32_t addr, uint32_t data);
	uint32_t read_data8(uint32_t addr);
	void write_data16(uint32_t addr, uint32_t data);
	uint32_t read_data16(uint32_t addr);
	void write_data32(uint32_t addr, uint32_t data);
	uint32_t read_data32(uint32_t addr);
	void write_data8w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_data8w(uint32_t addr, int* wait);
	void write_data16w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_data16w(uint32_t addr, int* wait);
	void write_data32w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_data32w(uint32_t addr, int* wait);
	uint32_t fetch_op(uint32_t addr, int *wait);
	void latch_address(uint32_t addr, int *wait);

	void write_io8(uint32_t addr, uint32_t data);
	uint32_t read_io8(uint32_t addr);
	void write_io16(uint32_t addr, uint32_t data);
	uint32_t read_io16(uint32_t addr);
	void write_io32(uint32_t addr, uint32_t data);
	uint32_t read_io32(uint32_t addr);
	void write_io8w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_io8w(uint32_t addr, int* wait);
	void write_io16w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_io16w(uint32_t addr, int* wait);
	void write_io32w(uint32_t addr, uint32_t data, int* wait);
	uint32_t read_io32w(uint32_t addr, int* wait);

	void write_signal(int id, uint32_t data, uint32_t mask);
	uint32_t read_signal(int id);

	// unique functions
	void set_context_mem(DEVICE* device)
	{
		d_mem = device;
	}
	void set_context_io(DEVICE* device)
	{
		d_io = device;
	}

	void check_break_points(uint32_t addr);
	void check_intr_break_points(uint32_t addr);

	void store_break_points();
	void restore_break_points();

	int hit_break_point();
	int hit_trace_point();
	void set_break_point(bool val);
	void set_trace_point(bool val);

	break_point_t bp[BP_END];
	break_point_t tp[TP_END];
	_TCHAR file_path[_MAX_PATH];
	int now_going;
	bool now_debugging, now_suspended;
	bool now_breakpoint, now_tracepoint;

// -----

	void enable_server(bool enable);
	void enable_connect();
	void network_connected(int ch);
	void network_disconnected(int ch);
	void network_writeable(int ch);
	void network_readable(int ch);
	void network_accepted(int ch, int new_ch);
	uint8_t* get_sendbuffer(int ch, int* size, int* flags);
	void inc_sendbuffer_ptr(int ch, int size);
	uint8_t* get_recvbuffer0(int ch, int* size0, int* size1, int* flags);
	uint8_t* get_recvbuffer1(int ch);
	void inc_recvbuffer_ptr(int ch, int size);

	void write_data(const _TCHAR *data, int size);
	int  read_data(_TCHAR *data, int size);

};

#endif
#endif

