/** @file device.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ device base class ]
*/

#ifndef _DEVICE_H_
#define _DEVICE_H_

#include "vm.h"

class EMU;
class FILEIO;

#ifdef USE_DEBUGGER
class DebuggerConsole;
#endif

// common signal id
#define SIG_CPU_IRQ	101
#define SIG_CPU_FIRQ	102
#define SIG_CPU_NMI	103
#define SIG_CPU_BUSREQ	104
#define SIG_CPU_DEBUG	201
#ifdef USE_EMU_INHERENT_SPEC
#define SIG_CPU_BABS	196
#define SIG_CPU_ADD_DEAD_CYCLE	197
#define SIG_CPU_HALT	198
#define SIG_CPU_RESET	199
#endif

/**
	@brief device base class
*/
class DEVICE
{
protected:
	VM* vm;
	EMU* emu;
public:
	DEVICE(VM* parent_vm, EMU* parent_emu, const char *identifier);
	virtual ~DEVICE();

	virtual void initialize();
	virtual void release();

	virtual void update_config();
	virtual void save_state(FILEIO* fio);
	virtual void load_state(FILEIO* fio);
	virtual void set_class_name(const char *name);

	// control
	virtual void reset();
	virtual void special_reset();

	// memory bus
	virtual void    write_data8(uint32_t addr, uint32_t data);
	virtual uint32_t read_data8(uint32_t addr);
	virtual void    write_data16(uint32_t addr, uint32_t data);
	virtual uint32_t read_data16(uint32_t addr);
	virtual void    write_data32(uint32_t addr, uint32_t data);
	virtual uint32_t read_data32(uint32_t addr);
	virtual void    write_data8w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_data8w(uint32_t addr, int* wait);
	virtual void    write_data16w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_data16w(uint32_t addr, int* wait);
	virtual void    write_data32w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_data32w(uint32_t addr, int* wait);

	virtual uint32_t fetch_op(uint32_t addr, int *wait);
	virtual void    latch_address(uint32_t addr, int *wait);

	virtual void    write_dma_data8(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_data8(uint32_t addr);
	virtual void    write_dma_data16(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_data16(uint32_t addr);
	virtual void    write_dma_data32(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_data32(uint32_t addr);

	// i/o bus
	virtual void    write_io8(uint32_t addr, uint32_t data);
	virtual uint32_t read_io8(uint32_t addr);
	virtual void    write_io16(uint32_t addr, uint32_t data);
	virtual uint32_t read_io16(uint32_t addr);
	virtual void    write_io32(uint32_t addr, uint32_t data);
	virtual uint32_t read_io32(uint32_t addr);
	virtual void    write_io8w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_io8w(uint32_t addr, int* wait);
	virtual void    write_io16w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_io16w(uint32_t addr, int* wait);
	virtual void    write_io32w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_io32w(uint32_t addr, int* wait);

	virtual void    write_dma_io8(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_io8(uint32_t addr);
	virtual void    write_dma_io16(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_io16(uint32_t addr);
	virtual void    write_dma_io32(uint32_t addr, uint32_t data);
	virtual uint32_t read_dma_io32(uint32_t addr);

	// memory mapped i/o
	virtual void    write_memory_mapped_io8(uint32_t addr, uint32_t data);
	virtual uint32_t read_memory_mapped_io8(uint32_t addr);
	virtual void    write_memory_mapped_io16(uint32_t addr, uint32_t data);
	virtual uint32_t read_memory_mapped_io16(uint32_t addr);
	virtual void    write_memory_mapped_io32(uint32_t addr, uint32_t data);
	virtual uint32_t read_memory_mapped_io32(uint32_t addr);
	virtual void    write_memory_mapped_io8w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_memory_mapped_io8w(uint32_t addr, int* wait);
	virtual void    write_memory_mapped_io16w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_memory_mapped_io16w(uint32_t addr, int* wait);
	virtual void    write_memory_mapped_io32w(uint32_t addr, uint32_t data, int* wait);
	virtual uint32_t read_memory_mapped_io32w(uint32_t addr, int* wait);

	// device to device
	typedef struct {
		DEVICE *device;
		int id;
		uint32_t mask;
		int shift;
	} output_t;

	typedef struct {
		int count;
		output_t item[MAX_OUTPUT];
	} outputs_t;

	virtual void init_output_signals(outputs_t *items);
	virtual void register_output_signal(outputs_t *items, DEVICE *device, int id, uint32_t mask, int shift);
	virtual void register_output_signal(outputs_t *items, DEVICE *device, int id, uint32_t mask);
	virtual void write_signals(outputs_t *items, uint32_t data);
	virtual void write_signal(int id, uint32_t data, uint32_t mask);
	virtual uint32_t read_signal(int ch);

	// z80 daisy chain
	virtual void set_context_intr(DEVICE* device, uint32_t bit);
	virtual void set_context_child(DEVICE* device);

	// interrupt device to device
	virtual void set_intr_iei(bool val);

	// interrupt device to cpu
	virtual void set_intr_line(bool line, bool pending, uint32_t bit);

	// interrupt cpu to device
	virtual uint32_t intr_ack();
	virtual void intr_reti();
	virtual void intr_ei();

	// dma
	virtual void do_dma();

	// cpu
	virtual int run(int clock);
#ifdef USE_CPU_REAL_MACHINE_CYCLE
	virtual int run(int clock, int accum, int cycles);
#endif
	virtual uint32_t get_pc();
	virtual uint32_t get_next_pc();

	// bios
	virtual bool bios_call(uint32_t PC, uint16_t regs[], uint16_t sregs[], int32_t* ZeroFlag, int32_t* CarryFlag);
	virtual bool bios_int(int intnum, uint16_t regs[], uint16_t sregs[], int32_t* ZeroFlag, int32_t* CarryFlag);

	// event manager
	DEVICE* event_manager;

	virtual void set_context_event_manager(DEVICE* device);
	virtual int  event_manager_id();
	virtual void register_event(DEVICE* device, int event_id, double usec, bool loop, int* register_id, uint64_t *expire_clock = NULL);
	virtual void register_event_by_clock(DEVICE* device, int event_id, int clock, bool loop, int* register_id, uint64_t *expire_clock = NULL);
	virtual void modify_event_by_clock(int register_id, int clock);
	virtual void cancel_event(DEVICE* device, int register_id);
	virtual int  is_registerd_event(const char *class_name, const char *identifier, int register_id);
	virtual void register_frame_event(DEVICE* device);
	virtual	bool is_registerd_frame_event(const char *class_name, const char *identifier);
	virtual void register_vline_event(DEVICE* device);
	virtual	bool is_registerd_vline_event(const char *class_name, const char *identifier);
	virtual uint64_t get_current_clock();
	virtual uint64_t get_passed_clock(uint64_t prev);
	virtual double get_passed_usec(uint64_t prev);

	virtual uint32_t get_cpu_pc(int index);
	virtual void set_frames_per_sec(double frames);
	virtual void set_lines_per_frame(int lines);
	virtual void update_timing(int new_clocks, double new_frames_per_sec, int new_lines_per_frame);

	// event callback
	virtual void event_callback(int event_id, int err);
	virtual void event_pre_frame();	// this event is to update timing settings
	virtual void event_frame();
	virtual void event_vline(int v, int clock);
	virtual void event_hsync(int v, int h, int clock);

	// sound
	virtual void mix(int32_t* buffer, int cnt);
//	virtual void set_volume(int volume);
//	virtual void set_volume(int ch, int decibel_l, int decibel_r);

	// network
	virtual void network_connected(int ch);
	virtual void network_disconnected(int ch);
	virtual void network_writeable(int ch);
	virtual void network_readable(int ch);
	virtual void network_accepted(int ch, int new_ch);
	virtual uint8_t* get_sendbuffer(int ch, int* size, int* flags);
	virtual void inc_sendbuffer_ptr(int ch, int size);
	virtual uint8_t* get_recvbuffer0(int ch, int* size0, int* size1, int* flags);
	virtual uint8_t* get_recvbuffer1(int ch);
	virtual void inc_recvbuffer_ptr(int ch, int size);

#ifdef USE_DEBUGGER
	// debugger
	virtual void *get_debugger();
	virtual uint32_t debug_prog_addr_mask();
	virtual uint32_t debug_data_addr_mask();
	virtual uint32_t debug_physical_addr_mask(int type);
	virtual void debug_write_data8(int type, uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_data8(int type, uint32_t addr);
	virtual void debug_write_data16(int type, uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_data16(int type, uint32_t addr);
	virtual void debug_write_data32(int type, uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_data32(int type, uint32_t addr);

	virtual void debug_write_io8(uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_io8(uint32_t addr);
	virtual void debug_write_io16(uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_io16(uint32_t addr);
	virtual void debug_write_io32(uint32_t addr, uint32_t data);
	virtual uint32_t debug_read_io32(uint32_t addr);
	virtual uint32_t debug_latch_address(uint32_t addr);

	virtual void debug_event_frame();

	virtual uint32_t debug_read_bank(uint32_t addr);
	virtual bool debug_write_reg(const _TCHAR *reg, uint32_t data);
	virtual bool debug_write_reg(uint32_t reg_num, uint32_t data);

	virtual void debug_regs_info(_TCHAR *buffer, size_t buffer_len);
	virtual void debug_memory_map_info(DebuggerConsole *dc);
	virtual void debug_address_map_info(DebuggerConsole *dc, int index);
	virtual int debug_dasm(int type, uint32_t pc, _TCHAR *buffer, size_t buffer_len, int flags);
	virtual int debug_trace_back_regs(int index, _TCHAR *buffer, size_t buffer_len);
	virtual int debug_trace_back(int index, _TCHAR *buffer, size_t buffer_len);
	virtual uint32_t debug_address_mapping_rev(uint32_t addr);
	virtual bool reach_break_point();
	virtual bool now_suspend();

	virtual	uint32_t get_debug_pc(int type);
	virtual	uint32_t get_debug_next_pc(int type);

	virtual bool get_debug_signal_name_index(const _TCHAR *param, uint32_t *num, int *idx, const _TCHAR **name);
	virtual bool get_debug_signal_name_index(uint32_t num, int *idx, const _TCHAR **name);
	virtual void get_debug_signal_names_str(_TCHAR *buffer, size_t buffer_len);

	virtual bool get_debug_graphic_memory_size(int type, int *width, int *height);
	virtual bool debug_draw_graphic(int type, int width, int height, scrntype *buffer);

#endif /* USE_DEBUGGER */

	DEVICE* prev_device;
	DEVICE* next_device;
	int this_device_id;

	char this_class_name[13];
	char this_identifier[5];

#ifdef USE_EMU_INHERENT_SPEC
	// display
	virtual void update_vram(int v, int clock);

	bool now_reset;
#endif
	//for resume
#pragma pack(1)
	typedef struct vm_state_ident_st {
		char   class_name[12];
		char   identifier[4];
		uint32_t size;
		uint16_t version;
		int    device_id;
		char   reserved[6];
	} vm_state_ident_t;
	vm_state_ident_t vm_state_ident;
#pragma pack()

	bool find_state_chunk(FILEIO *fio, const char *class_name, const char *identifier, vm_state_ident_t *vm_state_i);
	bool find_state_chunk(FILEIO *fio, vm_state_ident_t *vm_state_i);

	void adjust_chunk_size(vm_state_ident_t *vm_state_i, uint32_t &chunk_size);

#define FIND_STATE_CHUNK(p_fio, st_i) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		p_fio->Fseek(-(long)sizeof(st_i), FILEIO_SEEK_CUR); \
	}

#define READ_STATE_CHUNK(p_fio, st_i, st_v) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		uint32_t i_size = Uint32_LE(st_i.size); \
		memset(&st_v, 0, sizeof(st_v)); \
		if (i_size >= (sizeof(st_v) + sizeof(st_i))) { \
			p_fio->Fread(&st_v, sizeof(st_v), 1); \
			p_fio->Fseek(i_size - sizeof(st_v) - sizeof(st_i), FILEIO_SEEK_CUR); \
		} else { \
			p_fio->Fread(&st_v, i_size - sizeof(st_i), 1); \
		} \
	}

#define READ_STATE_VARIABLE_CHUNK(p_fio, st_i, st_v) \
	{ \
		if (find_state_chunk(p_fio, &st_i) != true) { \
			return; \
		} \
		uint32_t i_size = Uint32_LE(st_i.size); \
		memset(&st_v, 0, sizeof(st_v)); \
		if (i_size >= (sizeof(st_v) + sizeof(st_i))) { \
			p_fio->Fread(&st_v, sizeof(st_v), 1); \
		} else { \
			p_fio->Fread(&st_v, i_size - sizeof(st_i), 1); \
		} \
	}

	// for debug
	virtual void dummyf(const void *format, ...);
};

#endif
