/** @file disk.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.09.16-

	@brief [ d88 image handler ]
*/

#ifndef _DISK_H_
#define _DISK_H_

#include "../common.h"
#include <stdlib.h>

// use crc calcrate function
#define USE_CALC_CRC

// d88 media type
#define MEDIA_TYPE_2D	0x00
#define MEDIA_TYPE_2DD	0x10
#define MEDIA_TYPE_2HD	0x20
#define MEDIA_TYPE_UNK	0xff

#define DRIVE_TYPE_2D	MEDIA_TYPE_2D
#define DRIVE_TYPE_2DD	MEDIA_TYPE_2DD
#define DRIVE_TYPE_2HD	MEDIA_TYPE_2HD
#define DRIVE_TYPE_UNK	MEDIA_TYPE_UNK


// d88 constant
#define DISK_BUFFER_SIZE	0x180000	// 1.5MB
#define TRACK_BUFFER_SIZE	0x40000		// 256KB

// teledisk decoder constant
#define STRING_BUFFER_SIZE	4096
#define LOOKAHEAD_BUFFER_SIZE	60
#define THRESHOLD		2
#define N_CHAR			(256 - THRESHOLD + LOOKAHEAD_BUFFER_SIZE)
#define TABLE_SIZE		(N_CHAR * 2 - 1)
#define ROOT_POSITION		(TABLE_SIZE - 1)
#define MAX_FREQ		0x8000

// track size
// preamble 146 + 16sectors 5936 + postamble (598)
#define TRACK_SIZE_5INCH_2D	0x1800
// preamble 146 + 26sectors 9646 + postamble (598)
#define TRACK_SIZE_8INCH_2D	0x2700

class FILEIO;

/**
	@brief d88 image handler
*/
class DISK
{
private:
	int drive_num;
	uint8_t buffer[DISK_BUFFER_SIZE];
	uint8_t tmp_track[TRACK_BUFFER_SIZE];
	int tmp_track_size;

	_TCHAR file_path[_MAX_PATH];
	int file_size, file_size_orig;
	int file_offset;
	uint32_t crc32;
	bool is_last_volume;
	bool filename_changed;

	bool check_media_type();
	bool check_d88_tracks();

	int get_track_offset(int trk, int side);
	uint8_t* get_track_buffer(int trk, int side);
	void set_sector_info(uint8_t *t);

	void replace_track(int trk, int side);
	void shrink_track_space(int start, int size);
	void expand_track_space(int start, int size);
	void recalc_track_data_table(int start, int offset);
	void trim_track_data_table();
	int calc_track_size(uint8_t *t);
	void set_file_size(int offset);

	void rename_file();
	bool write_file(bool new_file);

public:
	DISK(int drv);
	~DISK();

	bool open(const _TCHAR *path, int offset, uint32_t flags = 0);
	void close();
	bool flash(bool protect);
	void set_write_protect(bool val);
	bool get_track(int trk, int side);
	bool make_track(int trk, int side, int dden);
	bool parse_track(int trk, int side, int dden);
	bool parse_track2(int trk, int side, int dden);
	bool get_sector(int trk, int side, int index);
	bool get_sector(int trk, int side, int sect, bool cmp_side);

	uint16_t set_crc16(uint8_t data, uint16_t crc);
#ifdef USE_CALC_CRC
	uint16_t calc_crc16(uint8_t data, uint16_t crc);
#endif

	bool is_same_file(const _TCHAR *path, int offset);

	bool inserted;
	bool ejected;
	bool write_protected;
	bool changed;
	uint8_t media_type;
	uint8_t drive_type;

	//
	int num_of_side;

	// track
	uint8_t track[TRACK_BUFFER_SIZE];
	int track_size;
	int sector_num;

	// sector
	uint8_t* sector_id;
	uint8_t* sector_data;
	int sector_size;
	uint8_t id[6];
	uint8_t* density;
	uint8_t deleted;
	uint8_t status;
	uint8_t verify[128];
	uint8_t sector_pos[128];
};

/// parse and decode to d88 format from any disk image
class DISK_PARSER
{
private:
	FILEIO *fio;
	const _TCHAR *file_path;
	uint8_t *buffer;
	int *file_size_ptr;
	bool temporary;
	_TCHAR tmp_path[_MAX_PATH];

	struct td_hdr_t {
		char sig[3];
		uint8_t unknown;
		uint8_t ver;
		uint8_t dens;
		uint8_t type;
		uint8_t flag;
		uint8_t dos;
		uint8_t sides;
		uint16_t crc;
	};
	struct td_cmt_t {
		uint16_t crc;
		uint16_t len;
		uint8_t ymd[3];
		uint8_t hms[3];
	};
	struct td_trk_t {
		uint8_t nsec, trk, head;
		uint8_t crc;
	};
	struct td_sct_t {
		uint8_t c, h, r, n;
		uint8_t ctrl, crc;
	};
	struct imd_trk_t {
		uint8_t mode;
		uint8_t cyl;
		uint8_t head;
		uint8_t nsec;
		uint8_t size;
	};
	struct d88_hdr_t {
		char title[17];
		uint8_t rsrv[9];
		uint8_t protect;
		uint8_t type;
		uint32_t size;
		uint32_t trkptr[164];
	};
	struct d88_sct_t {
		uint8_t c, h, r, n;
		uint16_t nsec;
		uint8_t dens, del, stat;
		uint8_t rsrv[5];
		uint16_t size;
	};

	uint8_t text_buf[STRING_BUFFER_SIZE + LOOKAHEAD_BUFFER_SIZE - 1];
	uint16_t ptr;
	uint16_t bufcnt, bufndx, bufpos;
	uint16_t ibufcnt,ibufndx;
	uint8_t inbuf[512];
	uint16_t freq[TABLE_SIZE + 1];
	short prnt[TABLE_SIZE + N_CHAR];
	short son[TABLE_SIZE];
	uint16_t getbuf;
	uint8_t getlen;

	DISK_PARSER() {}

public:
	DISK_PARSER(FILEIO *, const _TCHAR *, uint8_t *, int *);
	~DISK_PARSER();

	// teledisk image decoder (td0)
	bool teledisk_to_d88();
	int next_word();
	int get_bit();
	int get_byte();
	void start_huff();
	void reconst();
	void update(int c);
	short decode_char();
	short decode_position();
	void init_decode();
	int decode(uint8_t *buf, int len);

	// imagedisk image decoder (imd)
	bool imagedisk_to_d88();

	// cpdread image decoder (dsk)
	bool cpdread_to_d88(int extended);

	// standard image decoder (fdi/tfd/2d/sf7)
	bool standard_to_d88(int type, int ncyl, int nside, int nsec, int size);

	void remove_tempfile();
};

#endif

