/** @file mc6809dasm.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.08.29 -

	@brief [ mc6809 disassembler ]
*/

#include "mc6809dasm.h"
#include "device.h"
#include "mc6809_consts.h"

#ifdef USE_DEBUGGER

static const opcode_t opcodes0[] =
{
{ 0x00, _T("NEG"), DIR },
{ 0x01, _T("NEG??"), DIR },
{ 0x02, _T("NGC??"), DIR },
{ 0x03, _T("COM"), DIR },
{ 0x04, _T("LSR"), DIR },
{ 0x05, _T("LSR??"), DIR },
{ 0x06, _T("ROR"), DIR },
{ 0x07, _T("ASR"), DIR },
{ 0x08, _T("ASL"), DIR },
{ 0x09, _T("ROL"), DIR },
{ 0x0A, _T("DEC"), DIR },
{ 0x0B, _T("DCC??"), DIR },
{ 0x0C, _T("INC"), DIR },
{ 0x0D, _T("TST"), DIR },
{ 0x0E, _T("JMP"), DIR },
{ 0x0F, _T("CLR"), DIR },

{ 0x10, _T("page1"), NONE },
{ 0x11, _T("page2"), NONE },
{ 0x12, _T("NOP"), INH },
{ 0x13, _T("SYNC"), INH },
{ 0x14, _T("HCF??"), INH },
{ 0x15, _T("HCF??"), INH },
{ 0x16, _T("LBRA"), REL2 },
{ 0x17, _T("LBSR"), REL2 },
{ 0x18, _T("SLCC?"), INH },
{ 0x19, _T("DAA"), INH },
{ 0x1A, _T("ORCC"), IMM1 },
{ 0x1B, _T("NOP??"), INH },
{ 0x1C, _T("ANDCC"), IMM1 },
{ 0x1D, _T("SEX"), INH },
{ 0x1E, _T("EXG"), TFR },
{ 0x1F, _T("TFR"), TFR },

{ 0x20, _T("BRA"), REL1 },
{ 0x21, _T("BRN"), REL1 },
{ 0x22, _T("BHI"), REL1 },
{ 0x23, _T("BLS"), REL1 },
{ 0x24, _T("BCC"), REL1 },
{ 0x25, _T("BCS"), REL1 },
{ 0x26, _T("BNE"), REL1 },
{ 0x27, _T("BEQ"), REL1 },
{ 0x28, _T("BVC"), REL1 },
{ 0x29, _T("BVS"), REL1 },
{ 0x2A, _T("BPL"), REL1 },
{ 0x2B, _T("BMI"), REL1 },
{ 0x2C, _T("BGE"), REL1 },
{ 0x2D, _T("BLT"), REL1 },
{ 0x2E, _T("BGT"), REL1 },
{ 0x2F, _T("BLE"), REL1 },

{ 0x30, _T("LEAX"), IDX },
{ 0x31, _T("LEAY"), IDX },
{ 0x32, _T("LEAS"), IDX },
{ 0x33, _T("LEAU"), IDX },
{ 0x34, _T("PSHS"), PSH },
{ 0x35, _T("PULS"), PSH },
{ 0x36, _T("PSHU"), PSH },
{ 0x37, _T("PULU"), PSH },
{ 0x38, _T("CWAI?"), IMM1 },
{ 0x39, _T("RTS"), INH },
{ 0x3A, _T("ABX"), INH },
{ 0x3B, _T("RTI"), INH },
{ 0x3C, _T("CWAI"), IMM1 },
{ 0x3D, _T("MUL"), INH },
{ 0x3E, _T("RES??"), INH },
{ 0x3F, _T("SWI"), INH },

{ 0x40, _T("NEGA"), INH },
{ 0x41, _T("NEGA?"), INH },
{ 0x42, _T("NGCA?"), INH },
{ 0x43, _T("COMA"), INH },
{ 0x44, _T("LSRA"), INH },
{ 0x45, _T("LSRA?"), INH },
{ 0x46, _T("RORA"), INH },
{ 0x47, _T("ASRA"), INH },
{ 0x48, _T("ASLA"), INH },
{ 0x49, _T("ROLA"), INH },
{ 0x4A, _T("DECA"), INH },
{ 0x4B, _T("DCCA?"), INH },
{ 0x4C, _T("INCA"), INH },
{ 0x4D, _T("TSTA"), INH },
{ 0x4E, _T("CLCA?"), INH },
{ 0x4F, _T("CLRA"), INH },

{ 0x50, _T("NEGB"), INH },
{ 0x51, _T("NEGB?"), INH },
{ 0x52, _T("NGCB?"), INH },
{ 0x53, _T("COMB"), INH },
{ 0x54, _T("LSRB"), INH },
{ 0x55, _T("LSRB?"), INH },
{ 0x56, _T("RORB"), INH },
{ 0x57, _T("ASRB"), INH },
{ 0x58, _T("ASLB"), INH },
{ 0x59, _T("ROLB"), INH },
{ 0x5A, _T("DECB"), INH },
{ 0x5B, _T("DCCB?"), INH },
{ 0x5C, _T("INCB"), INH },
{ 0x5D, _T("TSTB"), INH },
{ 0x5E, _T("CLCB?"), INH },
{ 0x5F, _T("CLRB"), INH },

{ 0x60, _T("NEG"), IDX },
{ 0x61, _T("NEG??"), IDX },
{ 0x62, _T("NGC??"), IDX },
{ 0x63, _T("COM"), IDX },
{ 0x64, _T("LSR"), IDX },
{ 0x65, _T("LSR??"), IDX },
{ 0x66, _T("ROR"), IDX },
{ 0x67, _T("ASR"), IDX },
{ 0x68, _T("ASL"), IDX },
{ 0x69, _T("ROL"), IDX },
{ 0x6A, _T("DEC"), IDX },
{ 0x6B, _T("DCC??"), IDX },
{ 0x6C, _T("INC"), IDX },
{ 0x6D, _T("TST"), IDX },
{ 0x6E, _T("JMP"), IDX },
{ 0x6F, _T("CLR"), IDX },

{ 0x70, _T("NEG"), EXT },
{ 0x71, _T("NEG??"), EXT },
{ 0x72, _T("NGC??"), EXT },
{ 0x73, _T("COM"), EXT },
{ 0x74, _T("LSR"), EXT },
{ 0x75, _T("LSR??"), EXT },
{ 0x76, _T("ROR"), EXT },
{ 0x77, _T("ASR"), EXT },
{ 0x78, _T("ASL"), EXT },
{ 0x79, _T("ROL"), EXT },
{ 0x7A, _T("DEC"), EXT },
{ 0x7B, _T("DCC??"), EXT },
{ 0x7C, _T("INC"), EXT },
{ 0x7D, _T("TST"), EXT },
{ 0x7E, _T("JMP"), JMP },
{ 0x7F, _T("CLR"), EXT },

{ 0x80, _T("SUBA"), IMM1 },
{ 0x81, _T("CMPA"), IMM1 },
{ 0x82, _T("SBCA"), IMM1 },
{ 0x83, _T("SUBD"), IMM2 },
{ 0x84, _T("ANDA"), IMM1 },
{ 0x85, _T("BITA"), IMM1 },
{ 0x86, _T("LDA"), IMM1 },
{ 0x87, _T("STA??"), IMM1 },
{ 0x88, _T("EORA"), IMM1 },
{ 0x89, _T("ADCA"), IMM1 },
{ 0x8A, _T("ORA"), IMM1 },
{ 0x8B, _T("ADDA"), IMM1 },
{ 0x8C, _T("CMPX"), IMM2 },
{ 0x8D, _T("BSR"), REL1 },
{ 0x8E, _T("LDX"), IMM2 },
{ 0x8F, _T("STX??"), IMM2 },

{ 0x90, _T("SUBA"), DIR },
{ 0x91, _T("CMPA"), DIR },
{ 0x92, _T("SBCA"), DIR },
{ 0x93, _T("SUBD"), DIR },
{ 0x94, _T("ANDA"), DIR },
{ 0x95, _T("BITA"), DIR },
{ 0x96, _T("LDA"), DIR },
{ 0x97, _T("STA"), DIR },
{ 0x98, _T("EORA"), DIR },
{ 0x99, _T("ADCA"), DIR },
{ 0x9A, _T("ORA"), DIR },
{ 0x9B, _T("ADDA"), DIR },
{ 0x9C, _T("CMPX"), DIR },
{ 0x9D, _T("JSR"), DIR },
{ 0x9E, _T("LDX"), DIR },
{ 0x9F, _T("STX"), DIR },

{ 0xA0, _T("SUBA"), IDX },
{ 0xA1, _T("CMPA"), IDX },
{ 0xA2, _T("SBCA"), IDX },
{ 0xA3, _T("SUBD"), IDX },
{ 0xA4, _T("ANDA"), IDX },
{ 0xA5, _T("BITA"), IDX },
{ 0xA6, _T("LDA"), IDX },
{ 0xA7, _T("STA"), IDX },
{ 0xA8, _T("EORA"), IDX },
{ 0xA9, _T("ADCA"), IDX },
{ 0xAA, _T("ORA"), IDX },
{ 0xAB, _T("ADDA"), IDX },
{ 0xAC, _T("CMPX"), IDX },
{ 0xAD, _T("JSR"), IDX },
{ 0xAE, _T("LDX"), IDX },
{ 0xAF, _T("STX"), IDX },

{ 0xB0, _T("SUBA"), EXT },
{ 0xB1, _T("CMPA"), EXT },
{ 0xB2, _T("SBCA"), EXT },
{ 0xB3, _T("SUBD"), EXT },
{ 0xB4, _T("ANDA"), EXT },
{ 0xB5, _T("BITA"), EXT },
{ 0xB6, _T("LDA"), EXT },
{ 0xB7, _T("STA"), EXT },
{ 0xB8, _T("EORA"), EXT },
{ 0xB9, _T("ADCA"), EXT },
{ 0xBA, _T("ORA"), EXT },
{ 0xBB, _T("ADDA"), EXT },
{ 0xBC, _T("CMPX"), EXT },
{ 0xBD, _T("JSR"), JMP },
{ 0xBE, _T("LDX"), EXT },
{ 0xBF, _T("STX"), EXT },

{ 0xC0, _T("SUBB"), IMM1 },
{ 0xC1, _T("CMPB"), IMM1 },
{ 0xC2, _T("SBCB"), IMM1 },
{ 0xC3, _T("ADDD"), IMM2 },
{ 0xC4, _T("ANDB"), IMM1 },
{ 0xC5, _T("BITB"), IMM1 },
{ 0xC6, _T("LDB"), IMM1 },
{ 0xC7, _T("STB??"), IMM1 },
{ 0xC8, _T("EORB"), IMM1 },
{ 0xC9, _T("ADCB"), IMM1 },
{ 0xCA, _T("ORB"), IMM1 },
{ 0xCB, _T("ADDB"), IMM1 },
{ 0xCC, _T("LDD"), IMM2 },
{ 0xCD, _T("HCF??"), INH },
{ 0xCE, _T("LDU"), IMM2 },
{ 0xCF, _T("STU??"), IMM2 },

{ 0xD0, _T("SUBB"), DIR },
{ 0xD1, _T("CMPB"), DIR },
{ 0xD2, _T("SBCB"), DIR },
{ 0xD3, _T("ADDD"), DIR },
{ 0xD4, _T("ANDB"), DIR },
{ 0xD5, _T("BITB"), DIR },
{ 0xD6, _T("LDB"), DIR },
{ 0xD7, _T("STB"), DIR },
{ 0xD8, _T("EORB"), DIR },
{ 0xD9, _T("ADCB"), DIR },
{ 0xDA, _T("ORB"), DIR },
{ 0xDB, _T("ADDB"), DIR },
{ 0xDC, _T("LDD"), DIR },
{ 0xDD, _T("STD"), DIR },
{ 0xDE, _T("LDU"), DIR },
{ 0xDF, _T("STU"), DIR },

{ 0xE0, _T("SUBB"), IDX },
{ 0xE1, _T("CMPB"), IDX },
{ 0xE2, _T("SBCB"), IDX },
{ 0xE3, _T("ADDD"), IDX },
{ 0xE4, _T("ANDB"), IDX },
{ 0xE5, _T("BITB"), IDX },
{ 0xE6, _T("LDB"), IDX },
{ 0xE7, _T("STB"), IDX },
{ 0xE8, _T("EORB"), IDX },
{ 0xE9, _T("ADCB"), IDX },
{ 0xEA, _T("ORB"), IDX },
{ 0xEB, _T("ADDB"), IDX },
{ 0xEC, _T("LDD"), IDX },
{ 0xED, _T("STD"), IDX },
{ 0xEE, _T("LDU"), IDX },
{ 0xEF, _T("STU"), IDX },

{ 0xF0, _T("SUBB"), EXT },
{ 0xF1, _T("CMPB"), EXT },
{ 0xF2, _T("SBCB"), EXT },
{ 0xF3, _T("ADDD"), EXT },
{ 0xF4, _T("ANDB"), EXT },
{ 0xF5, _T("BITB"), EXT },
{ 0xF6, _T("LDB"), EXT },
{ 0xF7, _T("STB"), EXT },
{ 0xF8, _T("EORB"), EXT },
{ 0xF9, _T("ADCB"), EXT },
{ 0xFA, _T("ORB"), EXT },
{ 0xFB, _T("ADDB"), EXT },
{ 0xFC, _T("LDD"), EXT },
{ 0xFD, _T("STD"), EXT },
{ 0xFE, _T("LDU"), EXT },
{ 0xFF, _T("STU"), EXT },
{ 0x00, NULL, NONE }
};

// ex1 opcodes (0x10 0x..)
static const opcode_t opcodes1[] =
{
{ 0x21, _T("LBRN"), REL2 },
{ 0x22, _T("LBHI"), REL2 },
{ 0x23, _T("LBLS"), REL2 },
{ 0x24, _T("LBCC"), REL2 },
{ 0x25, _T("LBCS"), REL2 },
{ 0x26, _T("LBNE"), REL2 },
{ 0x27, _T("LBEQ"), REL2 },
{ 0x28, _T("LBVC"), REL2 },
{ 0x29, _T("LBVS"), REL2 },
{ 0x2A, _T("LBPL"), REL2 },
{ 0x2B, _T("LBMI"), REL2 },
{ 0x2C, _T("LBGE"), REL2 },
{ 0x2D, _T("LBLT"), REL2 },
{ 0x2E, _T("LBGT"), REL2 },
{ 0x2F, _T("LBLE"), REL2 },
{ 0x3F, _T("SWI2"), SWI2 },
{ 0x83, _T("CMPD"), IMM2 },
{ 0x8C, _T("CMPY"), IMM2 },
{ 0x8E, _T("LDY"), IMM2 },
{ 0x8F, _T("STY??"), IMM2 },
{ 0x93, _T("CMPD"), DIR },
{ 0x9C, _T("CMPY"), DIR },
{ 0x9E, _T("LDY"), DIR },
{ 0x9F, _T("STY"), DIR },
{ 0xA3, _T("CMPD"), IDX },
{ 0xAC, _T("CMPY"), IDX },
{ 0xAE, _T("LDY"), IDX },
{ 0xAF, _T("STY"), IDX },
{ 0xB3, _T("CMPD"), EXT },
{ 0xBC, _T("CMPY"), EXT },
{ 0xBE, _T("LDY"), EXT },
{ 0xBF, _T("STY"), EXT },
{ 0xCE, _T("LDS"), IMM2 },
{ 0xCF, _T("STS??"), IMM2 },
{ 0xDE, _T("LDS"), DIR },
{ 0xDF, _T("STS"), DIR },
{ 0xEE, _T("LDS"), IDX },
{ 0xEF, _T("STS"), IDX },
{ 0xFE, _T("LDS"), EXT },
{ 0xFF, _T("STS"), EXT },
{ 0x00, NULL, NONE }
};

// ex2 opcodes (0x11 0x..)
static const opcode_t opcodes2[] =
{
{ 0x3F, _T("SWI3"), INH },
{ 0x83, _T("CMPU"), IMM2 },
{ 0x8C, _T("CMPS"), IMM2 },
{ 0x93, _T("CMPU"), DIR },
{ 0x9C, _T("CMPS"), DIR },
{ 0xA3, _T("CMPU"), IDX },
{ 0xAC, _T("CMPS"), IDX },
{ 0xB3, _T("CMPU"), EXT },
{ 0xBC, _T("CMPS"), EXT },
{ 0x00, NULL, NONE }
};

MC6809DASM::MC6809DASM()
{
	d_mem = NULL;

	// register stack
	memset(regs, 0, sizeof(regs));
	current_reg = &regs[0];
	current_idx = 0;
	codelen = 0;

	swi2_appear = ~0;
}

MC6809DASM::~MC6809DASM()
{
}

void MC6809DASM::ini_pc(uint16_t pc, uint8_t code)
{
	uint32_t inphyaddr = 0;
	if (d_mem) {
		inphyaddr = d_mem->debug_latch_address(pc);
	}
	push_stack_pc(inphyaddr, pc, code);
}

void MC6809DASM::push_stack_pc(uint32_t phyaddr, uint16_t pc, uint8_t code)
{
	uint16_t prev_state = current_reg->state;

	current_idx = (current_idx + 1) % PCSTACK_NUM;

	current_reg = &regs[current_idx];

	current_reg->phyaddr = phyaddr;
	current_reg->pc      = pc;
	current_reg->code[0] = code;
	current_reg->state   = prev_state;

	codelen = 1;

	current_reg->flags = 1;
}

int MC6809DASM::get_stack_pc(int index, regs_t *stack)
{
	if (index < 0) return -2;
	if (index >= PCSTACK_NUM) index = PCSTACK_NUM - 1;

	int reg_index = (current_idx - index + PCSTACK_NUM) % PCSTACK_NUM;
	bool exist = false;
	do {
		if (regs[reg_index].flags & 1) {
			exist = true;
			break;
		} 
		reg_index++;
		if (reg_index >= PCSTACK_NUM) reg_index = 0;
	} while(reg_index != current_idx);
	if (!exist) return -2;

	if (stack) *stack = regs[reg_index];

	index--;
	return index;
}

void MC6809DASM::set_mem(uint16_t addr, uint8_t data, bool write)
{
	if (d_mem) {
		current_reg->rw_phyaddr = d_mem->debug_latch_address(addr);
	}
	current_reg->rw_addr = addr;
	current_reg->rw_data = data;
	current_reg->flags &= 0xf1;
	current_reg->flags |= 2;
	if (write) current_reg->flags |= 4;
}

void MC6809DASM::set_mem(uint16_t addr, uint16_t data, bool write)
{
	if (d_mem) {
		current_reg->rw_phyaddr = d_mem->debug_latch_address(addr);
	}
	current_reg->rw_addr = addr;
	current_reg->rw_data = data;
	current_reg->flags &= 0xf1;
	current_reg->flags |= 0xa;
	if (write) current_reg->flags |= 4;
}

void MC6809DASM::set_phymem(uint32_t addr, uint8_t data, bool write)
{
	current_reg->rw_phyaddr = addr;
}

void MC6809DASM::set_code2(uint8_t code2)
{
	current_reg->code[codelen] = code2;
	codelen++;
}

void MC6809DASM::set_swi2_param(uint8_t cat, uint8_t idx)
{
	current_reg->code[codelen] = cat;
	codelen++;
	current_reg->code[codelen] = idx;
	codelen++;
}

void MC6809DASM::set_inh()
{
}

void MC6809DASM::set_swi()
{
}

void MC6809DASM::set_swi2()
{
}

void MC6809DASM::set_swi3()
{
}

void MC6809DASM::set_dir_addr(uint8_t dp, uint8_t addrl)
{
	current_reg->code[codelen] = addrl;
	codelen++;
	current_reg->dp = dp;
}

void MC6809DASM::set_dir(uint8_t dp, uint8_t addrl, uint8_t data, bool write)
{
	set_dir_addr(dp, addrl);
	set_mem((dp << 8) | addrl, data, write);
}

void MC6809DASM::set_dir(uint8_t dp, uint8_t addrl, uint16_t data, bool write)
{
	set_dir_addr(dp, addrl);
	set_mem((dp << 8) | addrl, data, write);
}

void MC6809DASM::set_imm(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_imm(uint16_t data)
{
	current_reg->code[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	current_reg->code[codelen] = (data & 0xff);
	codelen++;
}

void MC6809DASM::set_ext_addr(uint16_t addr)
{
	current_reg->code[codelen] = ((addr & 0xff00) >> 8);
	codelen++;
	current_reg->code[codelen] = (addr & 0xff);
	codelen++;
}

void MC6809DASM::set_ext(uint16_t addr, uint8_t data, bool write)
{
	set_ext_addr(addr);
	set_mem(addr, data, write);
}

void MC6809DASM::set_ext(uint16_t addr, uint16_t data, bool write)
{
	set_ext_addr(addr);
	set_mem(addr, data, write);
}

void MC6809DASM::set_idx1(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_idx2(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_idx2(uint16_t data)
{
	current_reg->code[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	current_reg->code[codelen] = (data & 0xff);
	codelen++;
}

void MC6809DASM::set_rel(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_rel(uint16_t data)
{
	current_reg->code[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	current_reg->code[codelen] = (data & 0xff);
	codelen++;
}

void MC6809DASM::set_jmp(uint16_t data)
{
	current_reg->code[codelen] = ((data & 0xff00) >> 8);
	codelen++;
	current_reg->code[codelen] = (data & 0xff);
	codelen++;
}

void MC6809DASM::set_tfr(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_psh(uint8_t data)
{
	current_reg->code[codelen] = data;
	codelen++;
}

void MC6809DASM::set_err()
{
}

void MC6809DASM::set_regs(int accum, int cycles, uint8_t cc, uint8_t dp, uint8_t a, uint8_t b, uint16_t x, uint16_t y, uint16_t s, uint16_t u)
{
//	start_accum = accum;
//	expended_cycles = cycles;
	current_reg->cycles = cycles;
	current_reg->cc = cc;
	current_reg->dp = dp;
	current_reg->a = a;
	current_reg->b = b;
	current_reg->x = x;
	current_reg->y = y;
	current_reg->s = s;
	current_reg->u = u;
}

/// @param [in] flags bit5: 1=calc rel addr using phy_addr  bit4: 1=not use dp
int MC6809DASM::print_dasm(uint32_t phy_addr, uint16_t pc, const uint8_t *ops, int opslen, int flags, uint8_t *dp)
{
	en_opmode mode = NONE;
	int opspos = set_cmd_str(phy_addr, ops, &mode);
	switch(mode) {
	case INH:
		break;
	case DIR:
		if (flags & 0x10) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" <$%02X")
				, ops[opspos]
			);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
				, (dp ? *dp : current_reg->dp), ops[opspos]
			);
		}
		opspos++;
		break;
	case EXT:
		_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
			, ops[opspos], ops[opspos+1]
		);
		opspos+=2;
		break;
	case IMM1:
		_stprintf(&cmd[_tcslen(cmd)], _T(" #$%02X")
			, ops[opspos]
		);
		opspos++;
		break;
	case IMM2:
		_stprintf(&cmd[_tcslen(cmd)], _T(" #$%02X%02X")
			, ops[opspos], ops[opspos+1]
		);
		opspos+=2;
		break;
	case REL1:
		{
			uint16_t data = ops[opspos];
			data = (flags & 0x20 ? phy_addr & 0xffff : pc) + opspos + 1 + ((data & 0x80) ? data - 256 : data);
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X")
				, data 
			);
		}
		opspos++;
		break;
	case REL2:
		{
			uint16_t data = (ops[opspos] << 8) + ops[opspos+1];
			data = (flags & 0x20 ? phy_addr & 0xffff : pc) + opspos + 2 + ((data & 0x8000) ? data - 65536 : data);
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X")
				, data
			);
		}
		opspos+=2;
		break;
	case JMP:
		_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
			, ops[opspos], ops[opspos+1]
		);
		opspos+=2;
		break;
	case IDX:
		opspos = set_idx_str(pc, ops, opspos);
		break;
	case TFR:
		opspos = set_tfr_str(ops, opspos);
		break;
	case PSH:
		opspos = set_psh_str(ops, opspos);
		break;
	case SWI2:
#ifdef _MBS1
		opspos = set_swi2_str(phy_addr, ops, opspos);
#endif
		break;
	case FDB:
		_stprintf(&cmd[_tcslen(cmd)], _T(" $%02X%02X")
			,ops[opspos]
			,ops[opspos+1]
		);
		opspos+=2;
		break;
	default:
		break;
	}

#ifdef _MBS1
	_stprintf(&line[_tcslen(line)], _T("%05X %04X"), phy_addr, pc);
#else
	_stprintf(&line[_tcslen(line)], _T("%04X"), pc);
#endif
	for(int i=0; i < opslen; i++) {
		if (i < opspos) {
			_stprintf(&line[_tcslen(line)], _T(" %02X"), ops[i]);
		} else {
			_tcscat(line, _T("   "));
		}
	}

	_stprintf(&line[_tcslen(line)], _T(" %-18s"), cmd);

	return opspos;
}

int MC6809DASM::print_dasm_preprocess(int type, uint32_t pc, int flags)
{
	// next opcode
	uint8_t ops[5];
	uint32_t phy_addr = 0;
#ifdef _MBS1
	if (type >= 0) {
		phy_addr = pc;
		if (type == 0) {
			pc = d_mem->debug_address_mapping_rev(pc);
		}
		for(int i = 0; i < 5; i++) {
			ops[i] = d_mem->debug_read_data8(type, phy_addr + i);
		}
	} else {
		phy_addr = d_mem->debug_latch_address(pc);
		for(int i = 0; i < 5; i++) {
			ops[i] = d_mem->debug_read_data8(-1, pc + i);
		}
	}
	if ((pc & 0xffff0000) == 0xffff0000) {
		// no mapping
		flags |= 0x20;
	}
#else
	phy_addr = pc;
	for(int i = 0; i < 5; i++) {
		ops[i] = d_mem->debug_read_data8(type, pc + i);
	}
#endif
	line[0]=0;
	int opspos = print_dasm(phy_addr, pc & 0xffff, ops, 5, flags, NULL);
	return opspos;
}

int MC6809DASM::print_dasm_processed(uint16_t pc)
{
	line[0]=0;
	int opspos = print_dasm(current_reg->phyaddr, current_reg->pc, current_reg->code, 5, 0, NULL);
	print_cycles(current_reg->cycles);
	print_memory_data(regs[current_idx]);
	return opspos;
}

int MC6809DASM::print_dasm_traceback(int index)
{
	regs_t stack;
	int next = get_stack_pc(index, &stack);
	if (next >= -1) {
		line[0]=0;
		print_dasm(stack.phyaddr, stack.pc & 0xffff, stack.code, 5, 0, &stack.dp);
		print_cycles(stack.cycles);
		print_memory_data(stack);
	}
	return next;
}

void MC6809DASM::print_cycles(int cycles)
{
	// expended machine cycles 
	_stprintf(&line[_tcslen(line)], _T(" (%2d)")
		,cycles
	);
}

void MC6809DASM::print_regs(uint8_t cc, uint8_t dp, uint8_t a, uint8_t b, uint16_t x, uint16_t y, uint16_t s, uint16_t u, uint32_t state)
{
	// register
	_stprintf(line,
		 _T(
/*		 "PC:%04X PPC:%04X" */
		 "DP:%02X A:%02X B:%02X X:%04X Y:%04X U:%04X S:%04X" /* EA:%04X" */
		 " CC:[%c%c%c%c%c%c%c%c]"
		 )
//		,PC
//		,PPC
		,dp
		,a, b
		,x, y, u, s
//		,EAD
		,(CC & CC_E) ? _T('E') : _T('-')
		,(CC & CC_IF) ? _T('F') : _T('-')
		,(CC & CC_H) ? _T('H') : _T('-')
		,(CC & CC_II) ? _T('I') : _T('-')
		,(CC & CC_N) ? _T('N') : _T('-')
		,(CC & CC_Z) ? _T('Z') : _T('-')
		,(CC & CC_V) ? _T('V') : _T('-')
		,(CC & CC_C) ? _T('C') : _T('-')
	 );
	if (state & MC6809_HALT_BIT)  _tcscat(line, _T(" HALT"));
	if (state & MC6809_RESET_BIT) _tcscat(line, _T(" RESET"));
	if (state & MC6809_NMI_BIT)   _tcscat(line, _T(" NMI"));
	if (state & MC6809_IRQ_BIT)   _tcscat(line, _T(" IRQ"));
	if (state & MC6809_FIRQ_BIT)  _tcscat(line, _T(" FIRQ"));
}

void MC6809DASM::print_regs_current()
{
	print_regs(current_reg->cc, current_reg->dp, current_reg->a, current_reg->b, current_reg->x, current_reg->y, current_reg->s, current_reg->u, current_reg->state);
}

int MC6809DASM::print_regs_traceback(int index)
{
	regs_t stack;
	int next = get_stack_pc(index, &stack);
	if (next >= -1) {
		print_regs(stack.cc, stack.dp, stack.a, stack.b, stack.x, stack.y, stack.s, stack.u, stack.state);
	}
	return next;
}

void MC6809DASM::print_memory_data(regs_t &stack)
{
	// memory data
	if ((stack.flags & 2) == 0) return;

	if ((stack.flags & 8) == 0) {
		_stprintf(&line[_tcslen(line)]
#ifdef _MBS1
			, _T(" (%c %05X %04X:%02X)")
			, (stack.flags & 4) ? _T('W') : _T('R')
			, stack.rw_phyaddr
#else
			, _T(" (%c %04X:%02X)")
			, (stack.flags & 4) ? _T('W') : _T('R')
#endif
			, stack.rw_addr, stack.rw_data & 0xff);
	} else {
		_stprintf(&line[_tcslen(line)]
#ifdef _MBS1
			, _T(" (%c %05X %04X:%04X)")
			, (stack.flags & 4) ? _T('W') : _T('R')
			, stack.rw_phyaddr
#else
			, _T(" (%c %04X:%04X)")
			, (stack.flags & 4) ? _T('W') : _T('R')
#endif
			, stack.rw_addr, stack.rw_data);
	}
}

void MC6809DASM::set_signal(int id, uint32_t state)
{
	current_reg->state = (uint16_t)state;
}

int MC6809DASM::set_cmd_str(uint32_t addr, const uint8_t *ops, en_opmode *mod)
{
	uint8_t ch = ops[0];
	uint8_t cl = ops[1];
	int oplen = 0;

	memset(cmd, 0, sizeof(cmd));

	if (swi2_appear == addr) {
		// system call for S1
		if ((ch & 0xf0) == 0x40) {
			_stprintf(cmd, _T("%-5s"), _T("FDB"));
			if (mod) *mod = FDB;
			return oplen;
		}
	}
	swi2_appear = ~0;

	if (ch == 0x10) {
		for(int i=0; opcodes1[i].s != NULL; i++) {
			if (cl == opcodes1[i].c) {
				_stprintf(cmd, _T("%-5s"), opcodes1[i].s);
				if (mod) *mod = opcodes1[i].m;
				oplen = 2;
				break;
			}
		}
		if (oplen != 2) {
			_stprintf(cmd, _T("%-5s"), _T("???"));
			oplen = 1;
		}
	} else if (ch == 0x11) {
		for(int i=0; opcodes2[i].s != NULL; i++) {
			if (cl == opcodes2[i].c) {
				_stprintf(cmd, _T("%-5s"), opcodes2[i].s);
				if (mod) *mod = opcodes2[i].m;
				oplen = 2;
				break;
			}
		}
		if (oplen != 2) {
			_stprintf(cmd, _T("%-5s"), _T("???"));
			oplen = 1;
		}
	} else {
		_stprintf(cmd, _T("%-5s"), opcodes0[ch].s);
		if (mod) *mod = opcodes0[ch].m;
		oplen = 1;
	}
	return oplen;
}

int MC6809DASM::set_idx_str(uint16_t pc, const uint8_t *ops, int opspos)
{
	_TCHAR reg;

	uint8_t postbyte = ops[opspos];
	uint8_t pdata1 = ops[opspos+1];
	uint16_t pdata2 = ((ops[opspos+1] << 8) | ops[opspos+2]);

	switch(postbyte & 0x60) {
	case 0x00: reg = _T('X'); break;
	case 0x20: reg = _T('Y'); break;
	case 0x40: reg = _T('U'); break;
	case 0x60: reg = _T('S'); break;
	}
	if ((postbyte & 0x80) == 0) {
		// 0x00 - 0x7f
		_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
		, ((int)postbyte & 0xf) - ((postbyte & 0x10) ? 16 : 0)
		, reg);
		opspos++;
	} else if ((postbyte & 0x8f) == 0x80) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c+")
			, reg);
		} else {
			// illegal
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,%c+] ???")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x81) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c++")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,%c++]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x82) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,-%c")
			, reg);
		} else {
			// illegal
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,-%c] ???")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x83) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,--%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,--%c]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x84) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" ,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [,%c]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x85) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" B,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [B,%c]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x86) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" A,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [A,%c]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x87) {
		// illegal
		_stprintf(&cmd[_tcslen(cmd)], _T(" ,???"));
		opspos++;
	} else if ((postbyte & 0x8f) == 0x88) {
		// 8bit offset
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
			, (int)pdata1 - ((pdata1 & 0x80) ? 256 : 0)
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [%d,%c]")
			, (int)pdata1 - ((pdata1 & 0x80) ? 256 : 0)
			, reg);
		}
		opspos+=2;
	} else if ((postbyte & 0x8f) == 0x89) {
		// 16bit offset
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" %d,%c")
			, (int)pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [%d,%c]")
			, (int)pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)
			, reg);
		}
		opspos+=3;
	} else if ((postbyte & 0x8f) == 0x8a) {
		// illegal
		_stprintf(&cmd[_tcslen(cmd)], _T(" ,???"));
		opspos++;
	} else if ((postbyte & 0x8f) == 0x8b) {
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" D,%c")
			, reg);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [D,%c]")
			, reg);
		}
		opspos++;
	} else if ((postbyte & 0x8f) == 0x8c) {
		// 8bit offset PCR
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X,PCR")
			, (opspos + 2 + pc + pdata1 - ((pdata1 & 0x80) ? 256 : 0)) & 0xffff
			);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X,PCR]")
			, (opspos + 2 + pc + pdata1 - ((pdata1 & 0x80) ? 256 : 0)) & 0xffff
			);
		}
		opspos+=2;
	} else if ((postbyte & 0x8f) == 0x8d) {
		// 16bit offset PCR
		if ((postbyte & 0x10) == 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X,PCR")
			, (opspos + 3  + pc + pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)) & 0xffff
			);
		} else {
			_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X,PCR]")
			, (opspos + 3  + pc + pdata2 - ((pdata2 & 0x8000) ? 65536 : 0)) & 0xffff
			);
		}
		opspos+=3;
	} else if ((postbyte & 0x8f) == 0x8e) {
		// illegal
		_stprintf(&cmd[_tcslen(cmd)], _T(" ,???"));
		opspos++;
	} else if ((postbyte & 0x8f) == 0x8f) {
		if ((postbyte & 0x10) == 0) {
			// illegal
			_stprintf(&cmd[_tcslen(cmd)], _T(" $%04X ???")
			, pdata2 & 0xffff
			);
		} else {
			// extended indirect
			_stprintf(&cmd[_tcslen(cmd)], _T(" [$%04X]")
			, pdata2 & 0xffff
			);
		}
		opspos+=3;
	}
	return opspos;
}

static const _TCHAR *tfr_regs[] = {
	_T("D"),
	_T("X"),
	_T("Y"),
	_T("U"),
	_T("S"),
	_T("PC"),
	_T("?"),
	_T("?"),

	_T("A"),
	_T("B"),
	_T("CC"),
	_T("DP"),
	_T("?"),
	_T("?"),
	_T("?"),
	_T("?"),
	NULL
};

int MC6809DASM::set_tfr_str(const uint8_t *ops, int opspos)
{
	uint8_t dh = ((ops[opspos] & 0xf0) >> 4);
	uint8_t dl = (ops[opspos] & 0xf);

	_stprintf(&cmd[_tcslen(cmd)], _T(" %s,%s")
		, tfr_regs[dh], tfr_regs[dl]
	);
	opspos++;
	return opspos;
}

static const _TCHAR *psh_regs[] = {
	_T("CC"),
	_T("A"),
	_T("B"),
	_T("DP"),
	_T("X"),
	_T("Y"),
	_T("S/U"),
	_T("PC"),
	NULL
};

int MC6809DASM::set_psh_str(const uint8_t *ops, int opspos)
{
	uint8_t data = ops[opspos];
	bool is_s = ((ops[0] & 0xf) < 0x06);
	bool is_psh = ((ops[0] & 1) == 0);
	bool one = true;
	int ist = (is_psh ? 7 : 0);
	int ied = (is_psh ? -1 : 8);
	int isp = (is_psh ? -1 : 1);
	for(int i=ist; i != ied; i+=isp) {
		if ((data & (1 << i)) != 0) {
			_stprintf(&cmd[_tcslen(cmd)], _T("%c%s")
				, one ? _T(' ') : _T(',')
				, (i == 6 ? (is_s ? _T("U") : _T("S")) : psh_regs[i])
			);
			one = false;
		}
	}
	opspos++;
	return opspos;
}

int MC6809DASM::set_swi2_str(uint32_t addr, const uint8_t *ops, int opspos)
{
	if ((ops[opspos] & 0xf0) == 0x40) {
		// system call for S1
		_stprintf(&cmd[_tcslen(cmd)], _T("  ($%02X $%02X)")
			,ops[opspos]
			,ops[opspos+1]
		);
		swi2_appear = addr + opspos;
	}
	return opspos;
}

#endif /* USE_DEBUGGER */
